/*
* Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008  K. M. Indlekofer
*
* m.indlekofer@gmx.de
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



#include "cmios9_err.h"
#include "cmios9.h"
#include "cmios9_bined.h"

#ifndef WIN32
/* getopt */
#include <unistd.h>
extern int getopt(int argc,char * const argv[],const char *optstring);
extern char *optarg;
extern int optind;
extern int opterr;
extern int optopt;
#endif /* !WIN32 */



static struct genfile_s bined_file;
static u_int bined_size;
static u_char *bined_buf;
static u_char *bined_bufmalloc;
static int bined_modified;



static int
bined_read_file(char *buf,u_int bufsiz)
{

	/* file size */
	bined_size=bined_file.size;
	/* XXX prevent int overflow */
	if (bined_size>0x7fffffff){
		printf("file too large\n");
		bined_size=0;
		return -1;
	}
	
	/* allocate memory for file */
	if (bined_bufmalloc!=NULL){
		free(bined_bufmalloc);
		bined_bufmalloc=NULL;
	}
	if ((buf==NULL)||(bufsiz<bined_size)){
		bined_bufmalloc=(u_char *)malloc(bined_size);
		if (bined_bufmalloc==NULL){
			printf("cannot allocate buffer for file\n");
			bined_size=0;
			bined_buf=NULL;
			bined_bufmalloc=NULL;
			return -1;
		}
		bined_buf=bined_bufmalloc;
	}else{
		bined_buf=(u_char *)buf;
		bined_bufmalloc=NULL;
	}
	
	/* read */
	printf("reading file (size %08x)...\n",bined_size);
	if (readfile4(&bined_file,(char *)bined_buf,0,bined_size)<0){
		printf("cannot read file\n");
		if (bined_bufmalloc!=NULL){
			free(bined_bufmalloc);
			bined_size=0;
			bined_buf=NULL;
			bined_bufmalloc=NULL;
		}
		return -1;
	}
	
	bined_modified=0;
	return 0;
}

static int
bined_hexdump(int pos,int len)
{
	int i,j,imax;
	u_char c;
	
	if (bined_size==0){
		return -1;
	}

#if 0
	printf("pos=%08x len=%08x\n",pos,len);
#endif
	
	if (pos<0){
		pos=0;
	}
	if (pos>=(int)bined_size){
		pos=((int)bined_size)-1;
	}
	if (len<0){
		len=1;
	}
	if (pos+len>(int)bined_size){
		len=((int)bined_size)-pos;
	}
	
	printf("addr    ");
	for (i=0;i<16;i++){
		printf("%1x  ",(pos+i)%16);
		if (i==7)
			printf(" ");
	}
	printf(" ascii\n");
	printf("--------------------------------------------------------------------------\n");
	
	imax=(len+15)/16;
	for (i=0;i<imax;i++){
		printf("%06x: ",pos+16*i); /* XXX display shift for >0xffffff */
		for (j=0;j<16;j++){
			if (16*i+j<len){
				c=bined_buf[pos+16*i+j];
				printf("%02x ",c);
			}else{
				printf("   ");
			}
			if (j==7){
				printf(" ");
			}
		}
		printf(" ");
		for (j=0;j<16;j++){
			if (16*i+j>=len){
				break; /* done */
			}
			c=bined_buf[pos+16*i+j];
			if (isprint(c)){
				printf("%c",c);
			}else{
				if (isprint(0x7f&c)){
					printf("%c",0x7f&c);
				}else{
					printf(".");
				}
			}
		}
		printf("\n");
	}
	
	return 0;
}

static int
bined_interpret(void)
{
#define CMDLEN 256
	char cmd[CMDLEN+1];
#define CMDTOKMAX 16
	char *(cmdtok[CMDTOKMAX]);
	int cmdtoknr;
	int i;
	char *p;

	/* position */
	int curpos;
	int curlen;
	
	/* defaults */
	curpos=0;
	curlen=256;
	
	err_reset();
	for (;;){
		err_output(stderr); /* print error message if necessary */
		err_reset();
interpret_parser_next:
		/* get command */
	    printf("%06x:",curpos);
		if ((curpos>=0)&&(curpos<(int)bined_size)){
			printf("%02x",bined_buf[curpos]);
		}else{
			printf("??");
		}
		if (bined_modified){
			printf(" *> ");
		}else{
			printf(" > ");
		}
		fflush(NULL);
		fgets(cmd,CMDLEN,stdin);
		if (strlen(cmd)==0){
			continue;
		}
		cmd[strlen(cmd)-1]='\0';
		if (strlen(cmd)==0){
			continue;
		}
		/* parse command line */
		for (i=0;i<CMDTOKMAX;i++){
			if (i==0){
				p=strtok(cmd," \t");
				if (p==NULL){
					/* no tokens at all */
					goto interpret_parser_next;
				}
				cmdtok[i]=p;
			}else{
				p=strtok(NULL," \t");
				if (p==NULL){
					/* done */
					break;
				}
				cmdtok[i]=p;
			}
		}
		cmdtoknr=i; /* number of tokens */
#ifdef DEBUGPARSER
		printf("parser: cmdtoknr: %i\n",cmdtoknr);
		for (i=0;i<cmdtoknr;i++){
			printf("%i: \"%s\"\n",i,cmdtok[i]);
		}
		printf("\n");
#endif
		
		/* parse command */
		if ((strcmp(cmdtok[0],"exit")==0)
			||(strcmp(cmdtok[0],"quit")==0)
			||(strcmp(cmdtok[0],"q")==0)
			||(strcmp(cmdtok[0],"bye")==0)){
			return 0;
		}else if ((strcmp(cmdtok[0],"help")==0)||(strcmp(cmdtok[0],"?")==0)){
			if (cmdtoknr!=1){
				fprintf(stderr,"usage: %s\n",cmdtok[0]);
				continue;
			}
			printf("\nbined commands:\n");
			printf("  pos <position> (=p)\n");
			printf("  len <length> (=l)\n");
			printf("  dump [<start> [<len>]] (=d)\n");
			printf("  nextdump (=n)\n");
			printf("  set <value> ... (=s)\n");
			printf("  string <string> (=str)\n");
			printf("  fill <value> <len>\n");
			printf("  zero <len>\n");
			printf("  copy <dst> <len>\n");
			printf("  import <tmppath> <offset> <len>\n");
			printf("  extract <start> <len>\n");
			printf("  size\n");
			printf("  write [<new-path>]\n");
			printf("  exit (=quit, bye)\n");
			printf("  help (=?)\n\n");
		}else if ((strcmp(cmdtok[0],"pos")==0)||(strcmp(cmdtok[0],"p")==0)){
			if (cmdtoknr!=2){
				fprintf(stderr,"usage: %s <position>\n",cmdtok[0]);
				continue;
			}
			sscanf(cmdtok[1],"%x",&curpos);
			if (curpos<0){
				curpos=((int)bined_size)-1; /* end */
			}
		}else if ((strcmp(cmdtok[0],"len")==0)||(strcmp(cmdtok[0],"l")==0)){
			if (cmdtoknr!=2){
				fprintf(stderr,"usage: %s <length>\n",cmdtok[0]);
				continue;
			}
			sscanf(cmdtok[1],"%x",&curlen);
			if (curlen<=0){ /* all */
				curlen=(int)bined_size;
			}
		}else if ((strcmp(cmdtok[0],"nextdump")==0)||(strcmp(cmdtok[0],"n")==0)){
			if (cmdtoknr!=1){
				fprintf(stderr,"usage: %s\n",cmdtok[0]);
				continue;
			}
			bined_hexdump(curpos,curlen);
			curpos+=curlen;
		}else if ((strcmp(cmdtok[0],"dump")==0)||(strcmp(cmdtok[0],"d")==0)){
			int tmppos,tmplen;
			
			if ((cmdtoknr<1)||(cmdtoknr>3)){
				fprintf(stderr,"usage: %s [<start> [<len>]]\n",cmdtok[0]);
				continue;
			}
			tmppos=curpos;
			tmplen=curlen;
			if (cmdtoknr>1){
				sscanf(cmdtok[1],"%x",&tmppos);
				if (tmppos<0){
					tmppos=((int)bined_size)-curlen; /* end */
				}
				if (tmppos<0){ /* problem? */
					tmppos=0;
				}
			}
			if (cmdtoknr>2){
				sscanf(cmdtok[2],"%x",&tmplen);
				if (tmplen<=0){ /* all */
					tmplen=(int)bined_size;
				}
			}
			bined_hexdump(tmppos,tmplen);
		}else if ((strcmp(cmdtok[0],"set")==0)||(strcmp(cmdtok[0],"s")==0)){
			u_int v; /* XXX must not be u_char for sscanf below!!! */
			int i;

			if (cmdtoknr<2){
				fprintf(stderr,"usage: %s <value> ...\n",cmdtok[0]);
				continue;
			}
			
			for (i=1;i<cmdtoknr;i++){ /* 1!!! */
				sscanf(cmdtok[i],"%x",&v);
				
				if (curpos<0){
					printf("not in buffer\n");
					break;
				}
				if (curpos>=(int)bined_size){
					printf("not in buffer\n");
					break;
				}
				bined_buf[curpos]=v;
				curpos++; /* auto increment */
				bined_modified=1;
			}
		}else if ((strcmp(cmdtok[0],"string")==0)||(strcmp(cmdtok[0],"str")==0)){
			int tmplen;
			int i;
			
			if (cmdtoknr!=2){
				fprintf(stderr,"usage: %s <string>\n",cmdtok[0]);
				continue;
			}
			tmplen=strlen(cmdtok[1]);
			
			for (i=0;i<tmplen;i++){
				if (curpos<0){
					printf("not in buffer\n");
					continue;
				}
				if (curpos>=(int)bined_size){
					printf("not in buffer\n");
					continue;
				}
				bined_buf[curpos]=cmdtok[1][i];
				curpos++; /* auto increment */
				bined_modified=1;
			}
		}else if (strcmp(cmdtok[0],"fill")==0){
			int i;
			int tmplen;
			u_int v; /* XXX must not be u_char for sscanf below!!! */

			if (cmdtoknr!=3){
				fprintf(stderr,"usage: %s <value> <len>\n",cmdtok[0]);
				continue;
			}
			sscanf(cmdtok[1],"%x",&v);
			sscanf(cmdtok[2],"%x",&tmplen);
			
			for (i=0;i<tmplen;i++){
				if (curpos<0){
					printf("not in buffer\n");
					break;
				}
				if (curpos>=(int)bined_size){
					printf("not in buffer\n");
					break;
				}
				bined_buf[curpos]=v;
				curpos++; /* auto increment */
				bined_modified=1;
			}
		}else if (strcmp(cmdtok[0],"zero")==0){
			int i;
			int tmplen;

			if (cmdtoknr!=2){
				fprintf(stderr,"usage: %s <len>\n",cmdtok[0]);
				continue;
			}
			sscanf(cmdtok[1],"%x",&tmplen);
			if (tmplen<0){
				continue;
			}
			
			for (i=0;i<tmplen;i++){
				if (curpos<0){
					printf("not in buffer\n");
					break;
				}
				if (curpos>=(int)bined_size){
					printf("not in buffer\n");
					break;
				}
				bined_buf[curpos]=0;
				curpos++; /* auto increment */
				bined_modified=1;
			}
		}else if (strcmp(cmdtok[0],"copy")==0){
			int i;
			int dst;
			int tmplen;

			if (cmdtoknr!=3){
				fprintf(stderr,"usage: %s <dst> <len>\n",cmdtok[0]);
				continue;
			}
			sscanf(cmdtok[1],"%x",&dst);
			if (dst==curpos){
				/* nothing to be done */
				continue;
			}
			sscanf(cmdtok[2],"%x",&tmplen);
			if (tmplen<0){
				continue;
			}
			
			/* Note: be careful with direction due to possible overlap! */
			if (dst>curpos){
				/* start with highest */
				dst+=tmplen-1;
				curpos+=tmplen-1; /* auto decrement */
			}
			for (i=0;i<tmplen;i++){
				if (dst<0){
					printf("not in buffer\n");
					break;
				}
				if (dst>=(int)bined_size){
					printf("not in buffer\n");
					break;
				}
				if (curpos<0){
					printf("not in buffer\n");
					break;
				}
				if (curpos>=(int)bined_size){
					printf("not in buffer\n");
					break;
				}
				bined_buf[dst]=bined_buf[curpos];
				if (dst>curpos){
					dst--;
					curpos--; /* auto decrement */
				}else{
					dst++;
					curpos++; /* auto increment */
				}
				bined_modified=1;
			}
			if (dst>curpos){
				curpos+=tmplen+1; /* auto increment */
			}
		}else if (strcmp(cmdtok[0],"import")==0){
			u_int tmpoff; /* u_int!!! */
			int tmplen;
			struct genfile_s tmpfile;

			if (cmdtoknr!=4){
				fprintf(stderr,"usage: %s <tmppath> <offset> <len>\n",cmdtok[0]);
				continue;
			}
			sscanf(cmdtok[2],"%x",&tmpoff); /* offset in file */
			sscanf(cmdtok[3],"%x",&tmplen);
			if (tmplen==0){
				continue;
			}
			if (tmplen<0){
				continue;
			}
			if ((curpos<0)||(curpos+tmplen>(int)bined_size)){
				printf("not in buffer\n");
				continue;
			}

			/* open file */
			if (readfiledes4(cmdtok[1],&tmpfile,0)<0){ /* 0: do not touch */
				printf("cannot open file\n");
				continue;
			}
			if (tmpoff+(u_int)tmplen>tmpfile.size){
				printf("range outside file\n");
				continue;
			}
			
			/* read */
			bined_modified=1; /* XXX anyway */
			printf("reading file (offset %08x, size %08x)...\n",tmpoff,tmplen);
			if (readfile4(&bined_file,(char *)(bined_buf+curpos),tmpoff,tmplen)<0){
				printf("cannot read file\n");
				continue;
			}
			curpos+=tmplen; /* auto increment */
		}else if (strcmp(cmdtok[0],"extract")==0){
			int start,len;

			if (cmdtoknr!=3){
				fprintf(stderr,"usage: %s <start> <len>\n",cmdtok[0]);
				continue;
			}
			sscanf(cmdtok[1],"%x",&start);
			if (start>=(int)bined_size){
				start=((int)bined_size)-1;
			}
			if (start<0){
				start=0;
			}
			sscanf(cmdtok[2],"%x",&len);
			if (len<0){
				len=((int)bined_size)-start;
			}
			if ((start+len)>(int)bined_size){
				len=((int)bined_size)-start;
			}

			/* move window: */
			bined_buf+=start;
			bined_size=len;

			/* Note: writing back to old file will not alter its length!!! */

			bined_modified=1;
		}else if (strcmp(cmdtok[0],"size")==0){
			if (cmdtoknr!=1){
				fprintf(stderr,"usage: %s\n",cmdtok[0]);
				continue;
			}
			printf("size %08x\n",bined_size);
		}else if (strcmp(cmdtok[0],"write")==0){
			if (cmdtoknr>2){
				fprintf(stderr,"usage: %s [<new-path>]\n",cmdtok[0]);
				continue;
			}

			if (bined_size==0){
				continue;
			}
	
			if (cmdtoknr>1){
				/* try to create new file if non-existent */
				creatfile3(cmdtok[1],bined_size,0,NULL); /* 0: non-dir */
				/* XXX ignore error */
				err_reset();
				/* open destination file */
				/* XXX allow for directory */
				if (readfiledes4(cmdtok[1],&bined_file,0)<0){ /* 0: do not touch */
					fprintf(stderr,"cannot open new file\n");
					continue;
				}
			}

			/* write file */
			printf("writing file...\n");
			if (writefile4(&bined_file,(char *)bined_buf,0,bined_size)<0){
				printf("cannot write file\n");
				continue;
			}

			/* touch */
			readfiledes3(NULL,&bined_file.fdes,NULL,NULL,
				&bined_file.startsec,&bined_file.part,
				NULL,0,NULL,NULL,1); /* 1: touch */
			/* XXX ignore errors */
			err_reset();

			bined_modified=0; /* written */
		}else{
			printf("unknown command, try \"help\"\n");
		}
	}

	return 0;
}



int
bined(char *path,char *buf,u_int bufsiz)
{
	int ret;

	if (path==NULL){
		return -1;
	}
	
	/* initialize variables */
	bined_size=0;
	bined_buf=NULL;
	bined_bufmalloc=NULL;
	bined_modified=0;
	
	/* open destination file */
	if (readfiledes4(path,&bined_file,0)<0){ /* 0: do not touch */
		fprintf(stderr,"cannot open file\n");
		return -1;
	}

	printf("entering bined\n");

	/* read file */
	if (bined_read_file(buf,bufsiz)<0){
		return -1;
	}

	printf("\ntry \"help\" for infos\n");
	
	/* command interpreter */
	ret=bined_interpret();
	printf("leaving bined\n\n");

	if (bined_bufmalloc!=NULL){
		free(bined_bufmalloc);
		bined_bufmalloc=NULL;
	}
	bined_size=0;
	return ret;
}



/* EOF */
