/*
* Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008  K. M. Indlekofer
*
* m.indlekofer@gmx.de
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



#include "cmios9_err.h"
#include "cmios9.h"
#include "cmios9_info.h"



#define FINFOFUNC_VOICE
#define FINFOFUNC_SUBVOICE
#define FINFOFUNC_SYSTEM
#define FINFOFUNC_INSTRUMENT
#define FINFOFUNC_MDRMT
#define FINFOFUNC_SQTK



typedef int (*finfofunc_t)(struct genfile_s *fp,char *id,int verbose);

struct finfo_s{
	char *id;
	int flag; /* 0: begin, 1: end, 2: complete */
	char *comment;
	finfofunc_t func;
};



#ifdef FINFOFUNC_VOICE
#include "cmios9_voice.h"

static int
finfofunc_voice(struct genfile_s *fp,char *id,int verbose)
{

	return printinfo_voice(fp,verbose);
}
#endif



#ifdef FINFOFUNC_VOICE
#include "cmios9_voice.h"

static int
finfofunc_subvoice(struct genfile_s *fp,char *id,int verbose)
{

	return printinfo_subvoice(fp,verbose);
}
#endif



#ifdef FINFOFUNC_SYSTEM
#include "cmios9_voice.h"

static int
finfofunc_system(struct genfile_s *fp,char *id,int verbose)
{

	/* XXX */
	if (fp->part->type==PARTTYPE_OS9MDR){
		/* assume CMI Series III system file */
		return printinfo_system(fp,verbose);
	}

	printf("CMI system file, QDOS system binary file\n");
	return 0;
}
#endif



#ifdef FINFOFUNC_SYSTEM
#include "cmios9_voice.h"

static int
finfofunc_instrument(struct genfile_s *fp,char *id,int verbose)
{

	return printinfo_instrument(fp,verbose);
}
#endif



#ifdef FINFOFUNC_MDRMT
#include "cmios9_mdr.h"

static int
finfofunc_mdrmt(struct genfile_s *fp,char *id,int verbose)
{

	return printinfo_mt(fp,verbose);
}
#endif



#ifdef FINFOFUNC_SQTK
#include "cmios9_sqtk.h"

static int
finfofunc_sqtk(struct genfile_s *fp,char *id,int verbose)
{

	return printinfo_sqtk(fp,id,verbose);
}
#endif



struct finfo_s finfo_list[]={
	"res",1,"resident subroutine module",NULL,
	"task",1,"6809 OS9 task",NULL,
	"68",1,"68k module",NULL,
	"_pg",1,"primary display page (Usr6 P Prog module)",NULL,
	"_dp",1,"display page (Usr6 P Prog module)",NULL,
	".DP",1,"Series I/II/IIx display page file",NULL,
	"_ov",1,"small display page (Usr6 P Prog module)",NULL,
	".OV",1,"Series I/II/IIx display overlay file, QDOS overlay file",NULL,
	"_xm",1,"ext. module (Usr6 P Prog module)",NULL,
	"_cm",1,"command module (Usr6 P Prog module)",NULL,
	"_pa",1,"MDR/GFX large display page (Usr6 P Prog module), associated _gfx file",NULL,
	"_pb",1,"MDR/GFX small display page (Usr6 P Prog module), associated _gfx file",NULL,
	"_gfx",1,"MDR/GFX display page (J-module), associated _pa/_pb file", NULL,
	"_dt",1,"DTM (Usr6 P Prog module)",NULL,
	"_ed",1,"CAPS (Usr6 P Prog module)",NULL,
	"_db",1,"Usr6 P Prog module",NULL,
	"_p1",1,"CPU P1 Subroutine module",NULL,
	"_cp",1,"CAPS Subroutine module",NULL,
	"_bgnd",1,"Subroutine/Program module",NULL,
	"_ovl",1,"WP Subroutine module",NULL,
	"_init",1,"Subroutine module",NULL,
	"_sr",1,"Subroutine module",NULL,
	"_fl",1,"FLANGER page file",NULL,
	"_ts",1,"Track Split page file",NULL,
	"_icon",1,"icon (Data module)",NULL,
	"_sy",1,"character bitmap",NULL,
	"_cfg",1,"configuration",NULL,
	"_px",1,"screendump bitmap",NULL,
	".jker",1,"JKERNEL data file",NULL,
	".mdr",1,"MDR config/data file",NULL,
	".mp",1,"MP data file",NULL,
	".wp",1,"WP data file",NULL,
	".os9",1,"OS9 file",NULL,
	".cmi",1,"CMI software file",NULL,
	".fft",1,"FFT page file",NULL,
	".pal",1,"PALRAM image (for type D graphics cards)",NULL,
	".ps",1,"PostScript file",NULL,
	".doc",1,"text file",NULL,
	".txt",1,"text file",NULL,
	".CM",1,"OS9 shell command file, QDOS command file",NULL,
	".a",1,"assembler source (rma)",NULL,
	".ma",1,"assembler source (c.mac, rma)",NULL,
	".ca",1,"C source (c.comp, rma)",NULL,
	".mca",1,"C source (c.prep, c.comp, rma)",NULL,
	".o",1,"assembler source (c.opt, rma)",NULL,
	".mo",1,"assembler source (c.mac, c.opt, rma)",NULL,
	".c",1,"C source (c.comp, c.opt, rma)",NULL,
	".mc",1,"C source (c.prep, c.comp, c.opt, rma)",NULL,
	".h",1,"C include file",NULL,
	".ha",1,"assembler include file",NULL,
	".d",1,"assembler include file",NULL,
	".r",1,"relocatable object file (rlink)",NULL,
	".l",1,"library (rlink)",NULL,
	".s",1,"binary image in Motorola hex format",NULL,
	".m",1,"macro definition file",NULL,
	".mh",1,"macro definition file",NULL,
	".u",1,"macro definition file",NULL,
	".i",1,"command include file",NULL,
	".p",1,"shell script, PASCAL souce",NULL,
	".asm",1,"assembler source (asm, OS-9 Level 1)",NULL,
	".b",1,"BASIC09 file",NULL,
	".bas",1,"BASIC09 file",NULL,
	"LBFILE.SY",2,"Series I/II/IIx library file",NULL,
#ifdef FINFOFUNC_SQTK
	".SY",1,NULL,finfofunc_system,
#else
	".SY",1,"CMI system file, QDOS system binary file",NULL,
#endif
	".RT",1,"router file",NULL,
#ifdef FINFOFUNC_INSTRUMENT
	".IN",1,NULL,finfofunc_instrument,
#else
	".IN",1,"CMI instrument file",NULL,
#endif
	".OP",1,"CAPS opus file",NULL,
	".RS",1,"Series I/II/IIx Real-time Composer file (page R), Series III Rhythm Sequencer file",NULL,
	".CL",1,"Cue List file (event module)",NULL,
#ifdef FINFOFUNC_SQTK
	".TK",1,NULL,finfofunc_sqtk,
#else
	".TK",1,"Cue List Take file",NULL,
#endif
#ifdef FINFOFUNC_VOICE
	".VC",1,NULL,finfofunc_voice,
#else
	".VC",1,"voice file, Series I/II/IIx voice file (page 3)",NULL,
#endif
#ifdef FINFOFUNC_SUBVOICE
	".SV",1,NULL,finfofunc_subvoice,
#else
	".SV",1,"subvoice file",NULL,
#endif
	".SX",1,"MIDI System Exclusive file",NULL,
	".TS",1,"Track Split file",NULL,
	".SL",1,"Series I/II/IIx scale file, Rev11+ scale file",NULL,
	".FN",1,"function curve file",NULL,
	".TP",1,"tape directory file",NULL,
	".PX",1,"screendump bitmap",NULL,
	".DR",1,"Disk Recorder tape file (<=Rev8)",NULL,
	".TR",1,"Disk Recorder track file (<=Rev8)",NULL,
#ifdef FINFOFUNC_MDRMT
	".MT",1,NULL,finfofunc_mdrmt,
#else
	".MT",1,"MDR Disk Recorder Multi-Track file (>=Rev9)",NULL,
#endif
	".MK",1,"MFX macro file",NULL,
	".DL",1,"Dubber Language file (MFX3 Rev>=13: for D.A.D. via MDL-command)",NULL,
	".EL",1,"???",NULL,
	".MX",1,"???",NULL,
	".TT",1,"Timecode Trigger file",NULL,
	".CO",1,"Series I/II/IIx control file (voice effect patches + function curves, page 7)",NULL,
#ifdef FINFOFUNC_SQTK
	".SQ",1,NULL,finfofunc_sqtk,
#else
	".SQ",1,"MIDI frame sequence file (MCL compiler output format), Series I/II/IIx keyboard sequence file (page 9), Series III sequence file",NULL,
#endif
	".PC",1,"Series I/II/IIx MCL Piece source file",NULL,
	".PT",1,"Series I/II/IIx MCL Part source file",NULL,
	".SS",1,"MCL Sequence source file (for SQ output)",NULL,
	".MC",1,"Series I/II/IIx MIDI configuration file (page I), Series III MCL source file",NULL,
	".AI",1,"Series I/II/IIx analog interface file (page A)",NULL,
	".HL",1,"Series I/II/IIx help file",NULL,
	".CF",1,"QDOS chain procedural file",NULL,
	".ED",1,"QDOS EDOS-converted file",NULL,
	".SA",1,"QDOS ASCII source file (optionally compressed spaces)",NULL,
	".AL",1,"QDOS assembler listing file",NULL,
	".LX",1,"QDOS loadable object in Motorola EXORciser (S-record) format",NULL,
	".LO",1,"QDOS loadable object in binary format (e.g. MIDI.LO, AIC09.LO)",NULL,
	".RO",1,"QDOS relocatable object file",NULL,
	"dsp96k.obj",0,"DSP96K object file",NULL,
	"mdr_data",2,"MDR config file",NULL,
	"mdr",0,"MDR file",NULL,
	NULL,0,NULL /* end of list */
};



/* general */
int
printinfo(char *path,int verbose)
{
	struct genfile_s fdes;
	int i;
	int l1,l2,l;
	int cmp;
	char *id;
	char *comment;
	finfofunc_t func;

	if (path==NULL){
		return -1;
	}

	/* open file */
	if (readfiledes4(path,&fdes,0)<0){ /* 0: do not touch */
		fprintf(stderr,"cannot open file\n");
		return -1;
	}

	if (fdes.dir){
		printf("directory\n");
		return -1;
	}

	/* find suffix in list */
	l1=strlen(path);
	func=NULL;
	for (i=0;;i++){
		id=finfo_list[i].id;
		if (id==NULL){
			/* end of list */
			printf("unknown file type\n");
			return -1;
		}
		l2=strlen(id);
		switch (finfo_list[i].flag){
		case 0: /* compare begin of path */
			l=0;
			cmp=(l1>=l2);
			break;
		case 1: /* compare end of path */
			l=l1-l2;
			cmp=(l1>=l2);
			break;
		default: /* compare complete path */
			l=0;
			cmp=(l1==l2);
		}
		/* compare */
		if (cmp&&(strncasecmp(id,path+l,l2)==0)){ /* case-insensitive */
			/* found */
			comment=finfo_list[i].comment;
			func=finfo_list[i].func;
			break;
		}
	}
	if (comment!=NULL){
		printf("%s\n",comment);
	}
	if (func!=NULL){
		return func(&fdes,id,verbose);
	}

	return 0;
}



/* EOF */
