/*
* Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2010  K. M. Indlekofer
*
* m.indlekofer@gmx.de
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



#include "cmios9_err.h"
#include "cmios9_io.h"



/* basic I/O */
/* device */
int readonly;



/* basic I/O */

/* general */
/* XXX no cache */
int
readblock(int devdes,OFF_T bytenr,void *buf,OFF_T bufsiz,OFF_T startbyte,OFF_T totsize,OFF_T dev_blksiz)
{
	OFF_T off;
	OFF_T bstart,bend,b1start,b1siz,b2start,b2siz,b3start,b3siz;
	static char *bbuf=NULL;
	static OFF_T bbufalloc=0;
	char *bp;
	
	if (buf==NULL){
		err_print("readblock: buf==NULL\n");
		return -1;
	}
	if (bufsiz<=0){
		return -1;
	}
	if (bytenr+bufsiz>totsize){
		err_print("readblock: bytenr out of range\n");
		return -1;
	}
	if (dev_blksiz<=0){
		return -1;
	}

	if (bbufalloc<dev_blksiz){
		if (bbuf!=NULL){
			free(bbuf);
			bbuf=NULL;
			bbufalloc=0;
		}
		bbuf=(char *)malloc((u_int)dev_blksiz);
		if (bbuf==NULL){
			bbufalloc=0;
			err_print("readblock: bbuf==NULL\n");
			return -1;
		}
		bbufalloc=dev_blksiz;
	}
	
	/* parameters for 3 chunks */
	bstart=startbyte+bytenr;
	bend=bstart+bufsiz;
	b1start=dev_blksiz*(bstart/dev_blksiz);
	b2start=dev_blksiz*((bstart+dev_blksiz-1)/dev_blksiz);
	b3start=dev_blksiz*(bend/dev_blksiz);
	b1siz=b2start-bstart; /* should be smaller than dev_blksiz now! */
	if (b1siz>=(OFF_T)bufsiz){
		/* just chunk 1 */
		b1siz=bufsiz;
		b2siz=0;
		b3siz=0;
	}else{
		/* chunk 1 is not enough */
		b2siz=b3start-b2start; /* should be multiple of dev_blksiz now! */
		if (b2siz<0){
			/* not enough for chunk 2 */
			b2siz=0;
		}
		b3siz=bend-b3start; /* should be smaller than dev_blksiz now! */
	}
	/* XXX off_t to prevent overflow of u_int!!! */
	off=b1start; /* lseek offset, multiple of dev_blksiz */
#ifdef DEBUGIO
	printf("---\n");
	printf("readblock: bytenr:  %08x\n",bytenr);
	printf("readblock: bufsiz:  %08x\n",bufsiz);
	printf("readblock: bstart:  %08x\n",bstart);
	printf("readblock: bend:    %08x\n",bend);
	printf("readblock: b1start: %08x\n",b1start);
	printf("readblock: b1siz:   %08x\n",b1siz);
	printf("readblock: b2start: %08x\n",b2start);
	printf("readblock: b2siz:   %08x\n",b2siz);
	printf("readblock: b3start: %08x\n",b3start);
	printf("readblock: b3siz:   %08x\n",b3siz);
	printf("---\n");
#endif
	bp=(char *)buf;
	
	if (LSEEK(devdes,off,SEEK_SET)!=off){
#ifdef DEBUGIO
		perror("readblock: lseek 1");
#endif
		return -1;
	}
	
	/* chunk 1 */
	if (b1siz>0){
		if (READ(devdes,bbuf,(size_t)dev_blksiz)!=(int)(dev_blksiz)){
#ifdef DEBUGIO
			perror("readblock: read 1");
#endif
			return -1;
		}
		/* extract part */
		bcopy(bbuf+(bstart-b1start),bp,(size_t)b1siz);
		bp+=b1siz;
	}
	
	/* chunk 2 */
	/* XXX b2siz must not exceed SSIZE_MAX! */
	if (b2siz>(OFF_T)(SSIZE_MAX)){
		err_print("readblock: bufsiz too large\n");
		return -1;
	}
	if (b2siz>0){
		if (READ(devdes,bp,(size_t)b2siz)!=(int)(b2siz)){
#ifdef DEBUGIO
			perror("readblock: read 2");
#endif
			return -1;
		}
		bp+=b2siz;
	}
	
	/* chunk 3 */
	if (b3siz>0){
		/* at least b3siz */
		if (READ(devdes,bbuf,(size_t)dev_blksiz)<(int)(b3siz)){
#ifdef DEBUGIO
			perror("readblock: read 3");
#endif
			return -1;
		}
		/* extract part */
		bcopy(bbuf,bp,(size_t)b3siz);
		bp+=b3siz;
	}
	
	return 0;
}

/* general */
/* XXX no cache */
int
writeblock(int devdes,OFF_T bytenr,void *buf,OFF_T bufsiz,OFF_T startbyte,OFF_T totsize,OFF_T dev_blksiz)
{
	OFF_T off;
	OFF_T bstart,bend,b1start,b1siz,b2start,b2siz,b3start,b3siz;
	static char *bbuf=NULL;
	static OFF_T bbufalloc=0;
	char *bp;
	
	if (buf==NULL){
		err_print("writeblock: buf==NULL\n");
		return -1;
	}
	if (bufsiz<=0){
		return -1;
	}
	if (bytenr+bufsiz>totsize){
		err_print("writeblock: bytenr out of range\n");
		return -1;
	}
	if (dev_blksiz<=0){
		return -1;
	}

	if (bbufalloc<dev_blksiz){
		if (bbuf!=NULL){
			free(bbuf);
			bbuf=NULL;
			bbufalloc=0;
		}
		bbuf=(char *)malloc((u_int)dev_blksiz);
		if (bbuf==NULL){
			bbufalloc=0;
			err_print("writeblock: bbuf==NULL\n");
			return -1;
		}
		bbufalloc=dev_blksiz;
	}
	
	/* parameters for 3 chunks */
	bstart=startbyte+bytenr;
	bend=bstart+bufsiz;
	b1start=dev_blksiz*(bstart/dev_blksiz);
	b2start=dev_blksiz*((bstart+dev_blksiz-1)/dev_blksiz);
	b3start=dev_blksiz*(bend/dev_blksiz);
	b1siz=b2start-bstart; /* should be smaller than dev_blksiz now! */
	if (b1siz>=(OFF_T)bufsiz){
		/* just chunk 1 */
		b1siz=bufsiz;
		b2siz=0;
		b3siz=0;
	}else{
		/* chunk 1 is not enough */
		b2siz=b3start-b2start; /* should be multiple of dev_blksiz now! */
		if (b2siz<0){
			/* not enough for chunk 2 */
			b2siz=0;
		}
		b3siz=bend-b3start; /* should be smaller than dev_blksiz now! */
	}
	/* XXX off_t to prevent overflow of u_int!!! */
	off=b1start; /* lseek offset, multiple of dev_blksiz */
#ifdef DEBUGIO
	printf("---\n");
	if (readonly)
		printf("writeblock: READONLY\n");
	printf("writeblock: bytenr:  %08x\n",bytenr);
	printf("writeblock: bufsiz:  %08x\n",bufsiz);
	printf("writeblock: bstart:  %08x\n",bstart);
	printf("writeblock: bend:    %08x\n",bend);
	printf("writeblock: b1start: %08x\n",b1start);
	printf("writeblock: b1siz:   %08x\n",b1siz);
	printf("writeblock: b2start: %08x\n",b2start);
	printf("writeblock: b2siz:   %08x\n",b2siz);
	printf("writeblock: b3start: %08x\n",b3start);
	printf("writeblock: b3siz:   %08x\n",b3siz);
	printf("---\n");
#endif
	bp=(char *)buf;
	
	if (LSEEK(devdes,off,SEEK_SET)!=off){
#ifdef DEBUGIO
		perror("writeblock: lseek 1");
#endif
		return -1;
	}
	
	/* chunk 1 */
	if (b1siz>0){
		/* read chunk */
		if (READ(devdes,bbuf,(size_t)dev_blksiz)!=(int)(dev_blksiz)){
#ifdef DEBUGIO
			perror("writeblock: read 1");
#endif
			return -1;
		}
		/* undo read offset */
		if (LSEEK(devdes,-(OFF_T)(dev_blksiz),SEEK_CUR)<0){
#ifdef DEBUGIO
			perror("writeblock: lseek undo 1");
#endif
			return -1;
		}
		/* modify */
		bcopy(bp,bbuf+(bstart-b1start),(size_t)b1siz);
		/* write chunk */
#ifndef READONLY
		if (!readonly){
			if (WRITE(devdes,bbuf,(size_t)dev_blksiz)!=(int)(dev_blksiz)){
#ifdef DEBUGIO
				perror("writeblock: write 1");
#endif
				return -1;
			}
		}
#endif
		bp+=b1siz;
	}
	
	/* chunk 2 */
	/* XXX b2siz must not exceed SSIZE_MAX! */
	if (b2siz>(OFF_T)(SSIZE_MAX)){
		err_print("writeblock: bufsiz too large\n");
		return -1;
	}
	if (b2siz>0){
#ifndef READONLY
		if (!readonly){
			if (WRITE(devdes,bp,(size_t)b2siz)!=(int)(b2siz)){
#ifdef DEBUGIO
				perror("writeblock: write 2");
#endif
				return -1;
			}
		}
#endif
		bp+=b2siz;
	}
	
	/* chunk 3 */
	/* XXX assume that we can access full dev_blksiz at the end */
	if (b3siz>0){
		int roff;

		/* read chunk */
		/* at least b3siz */
		if ((roff=READ(devdes,bbuf,(size_t)dev_blksiz))<(int)(b3siz)){
#ifdef DEBUGIO
			perror("writeblock: read 3");
#endif
			return -1;
		}
		/* undo read offset */
		if (LSEEK(devdes,-(OFF_T)(roff),SEEK_CUR)<0){ /* roff!!! */
#ifdef DEBUGIO
			perror("writeblock: lseek undo 3");
#endif
			return -1;
		}
		/* modify */
		bcopy(bp,bbuf,(size_t)b3siz);
		/* write chunk */
#ifndef READONLY
		if (!readonly){
			/* at least b3siz */
			if (WRITE(devdes,bbuf,(size_t)dev_blksiz)<(int)(b3siz)){
#ifdef DEBUGIO
				perror("writeblock: write 3");
#endif
				return -1;
			}
		}
#endif
		bp+=b3siz;
	}
	
	return 0;
}

OFF_T
getdevsize(int devdes,OFF_T dev_blksiz)
{
	OFF_T offmax,off,siz;
	static char *bbuf=NULL;
	static OFF_T bbufalloc=0;

	if (dev_blksiz<=0){
		return 0;
	}

	if (bbufalloc<dev_blksiz){
		if (bbuf!=NULL){
			free(bbuf);
			bbuf=NULL;
			bbufalloc=0;
		}
		bbuf=(char *)malloc((u_int)dev_blksiz);
		if (bbuf==NULL){
			bbufalloc=0;
			err_print("getdevsize: bbuf==NULL\n");
			return 0;
		}
		bbufalloc=dev_blksiz;
	}

#ifdef DEBUGIO
	fprintf(stderr,"getdevsize\n");
#endif

	/* max. device size, must be power of 2 */
	offmax=DEV_MAXSIZ/dev_blksiz;
	off=offmax;
	siz=off;
	do{
		siz>>=1;
		/* can read 1 block at off ? */
		if ((off>=offmax) /* beyond max. device limit? */
			||(readblock(devdes,off*dev_blksiz,bbuf,dev_blksiz,0,offmax*dev_blksiz,dev_blksiz)<0)){
			/* off is too high */
#ifdef DEBUGIO
			fprintf(stderr,"getdevsize: off=%08x siz=%08x high\n",(u_int)off,(u_int)siz);
#endif
			if (siz==0){ /* end of iteration? */
				off--; /* last block not readable */
			}else{
				off-=siz;
			}
		}else{
			/* off is OK or too low */
#ifdef DEBUGIO
			fprintf(stderr,"getdevsize: off=%08x siz=%08x low\n",(u_int)off,(u_int)siz);
#endif
			off+=siz;
		}
	}while ((siz>0)||(off==0)); /* if siz==0 and off==0: one last check for block 0!!! */
	/* now, off is last readable device block */

	/* return value in dev_blksize */
	off++; /* !!! */

#ifdef DEBUGIO
	fprintf(stderr,"getdevsize done: size=%08x\n",off);
#endif
	return off;
}

int
opendevice(char *devpath)
{
	int devdes;
	
	if (devpath==NULL){
		return -1;
	}

	/* open device/image file */
	/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
	if ((devdes=OPEN(devpath,(readonly?O_RDONLY:O_RDWR)|O_BINARY,0))<0){
		perror("opendevice");
		return -1;
	}

	return devdes;
}




/* EOF */
