/*
* Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2010  K. M. Indlekofer
*
* m.indlekofer@gmx.de
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



#include "cmios9_err.h"
#include "cmios9.h"
#include "cmios9_mdr.h"
#include "cmios9_wav.h"



#define MT2WAVALL
#define MT2WAVALL_NODUPLICATES



#define PRINT_HEAD_1(a,b) printf("%s: %02x\n",a,b)
#define PRINT_HEAD_2(a,b) printf("%s: %02x%02x\n",a,b[0],b[1])
#define PRINT_HEAD_4(a,b) printf("%s: %02x%02x%02x%02x\n",a,b[0],b[1],b[2],b[3])
#define GET_2(a) ((((u_int)a[0])<<8)+((u_int)a[1]))
#define GET_4(a) ((((u_int)a[0])<<24)+(((u_int)a[1])<<16)+(((u_int)a[2])<<8)+((u_int)a[3]))



static struct genfile_s *mdrmt_fp;
static u_int mdrmt_foff; /* current byte position */



static u_int natfrm;
static u_int natsub;
static u_int sblock;
static u_int nbytes;
static u_int ebase;
static u_int tbase;
static u_int samprate;
static u_int spacebase;
static u_int blksize;
static u_int markpos;
static u_int marksize;
static u_int newwave;
static struct mt_head_s mh;



static u_int
wnr2wbase(u_int wnr)
{
	u_int wperblk;
	u_int i;
	u_int skip,ssiz;
	int found;
	u_int wbase;

	/* waves per block (of nbytes) */
	wperblk=nbytes/sizeof(struct mt_wave_s);

	/* logical wave block */
	skip=wnr/wperblk;

	/* find physical wave block */
	found=0;
	for (i=0;i<MT_HEAD_WAVEDIRNR;i++){
		/* current segment */
		ssiz=(u_int)mh.wavedir[i].ssiz;
		if (ssiz==0){
			/* end of used segements */
			break;
		}

		/* check if this is wanted segment */
		if (skip<ssiz){
			found=1;
			break; /* done */
		}
		
		/* remaining offset */
		skip-=ssiz;
	}
	if (!found){
		return 0; /* invalid */
	}
	/* now, ssiz>0 */

	/* byte address of wave struct */
	wbase=sblock
		+(GET_2(mh.wavedir[i].off)+skip)*nbytes
		+(wnr*sizeof(struct mt_wave_s))%nbytes; /* wave within block */

	return wbase;
}



static void
set_bit(u_char *p,u_int bit,u_char val)
{
	u_int by,bi;
	u_char ma;

	if (p==NULL){
		return;
	}

	by=bit/8;
	bi=7-bit%8; /* upper Bit is first!!! */
	ma=1<<bi;

	if (val){
		/* set bit */
		p[by]|=ma;
	}else{
		/* clear bit */
		p[by]&=~ma;
	}
}

void
mdrmt_set_passwd(char *buf,u_char *pp)
{
	u_int by,bi;
	u_char ma;
	u_char c;

	if (pp==NULL){
		return;
	}

	/* clear passwd */
	for (bi=MT_HEAD_PASS_BIT0;bi<MT_HEAD_PASS_SZ*7;bi++){
		set_bit(pp,bi,0);
	}

	if (buf==NULL){
		return;
	}

	for (by=0;by<MT_HEAD_PASS_SZ;by++){
		if (buf[by]==0){
			break; /* end */
		}
		for (bi=0;bi<7;bi++){
			/* XXX ignore bit7 in buf */
			ma=1<<(6-bi); /* upper Bit is first */
			c=((u_char)buf[by])&ma;
			set_bit(pp,by*7+bi+MT_HEAD_PASS_BIT0,c);
		}
	}
	/* Note: rest of pp was cleared above! */
}

static u_int
get_bit(u_char *p,u_int bit)
{
	u_int by,bi;
	u_char ma;

	if (p==NULL){
		return 0; /* XXX */
	}

	by=bit/8;
	bi=7-bit%8; /* upper Bit is first!!! */
	ma=1<<bi;

	return p[by]&ma;
}

void
mdrmt_get_passwd(char *buf,u_char *pp)
{
	u_int by,bi;
	char c;

	if (buf==NULL){
		return;
	}
	if (pp==NULL){
		buf[0]=0;
		return;
	}

	for (by=0;by<MT_HEAD_PASS_SZ;by++){
		buf[by]=0;
		for (bi=0;bi<7;bi++){
			c=get_bit(pp,by*7+bi+MT_HEAD_PASS_BIT0)?1:0;
			buf[by]|=c<<(6-bi); /* upper Bit is first */
		}
		if (buf[by]==0){
			break; /* end */
		}
	}
	buf[by]=0;
}

void
mdrmt_print_passwd(char *sp,u_char *pp)
{
	u_int by;
	char buf[MT_HEAD_PASS_SZ+1];
	char val[MT_HEAD_PASS_SZ+1];

	if (pp==NULL){
		return;
	}

	mdrmt_get_passwd(buf,pp);

	for (by=0;by<MT_HEAD_PASS_SZ;by++){
		if (buf[by]==0){
			break; /* end */
		}
		/* check if valid character */
		if (isprint(buf[by])){
			val[by]='.';
		}else{
			val[by]='?';
			buf[by]='?'; /* printable replacement */
		}
	}
	val[by]=0;

	if (sp!=NULL){
		printf("%s",sp);
	}
	if (strlen(buf)!=0){
		printf("\"%s\"\n",buf);
		if (sp!=NULL){
			for (by=0;by<strlen(sp);by++){
				printf(" ");
			}
		}
		printf(" %s   (.==valid, ?==invalid)\n",val);
	}else{
		printf("(none)\n");
	}
}



static void
print_time(char *sp,u_int tu)
{
	u_int h,m,s,f,sf;

	if (sp!=NULL){
		printf("%s",sp);
	}

	h=tu/MT_TU_HOUR;
	tu-=h*MT_TU_HOUR;
	m=tu/MT_TU_MIN;
	tu-=m*MT_TU_MIN;
	s=tu/MT_TU_SEC;
	tu-=s*MT_TU_SEC;
	printf("%02d:%02d:%02d",h,m,s);
	if (natfrm!=0){
		f=tu/natfrm;
		tu-=f*natfrm;
		printf(":%02d",f);
		if (natsub!=0){
			sf=tu/natsub;
			tu-=sf*natsub;
			printf("+%02d",sf);
		}
	}
	printf("\n");
}



static float
get_float(u_int raw)
{
	static float f;

	*((u_int *)(&f))=raw;
	return f;
}



static void
print_wavedir(void)
{
	u_int i;
	u_int ssiz;

	for (i=0;i<MT_HEAD_WAVEDIRNR;i++){
		ssiz=(u_int)mh.wavedir[i].ssiz;
		if (ssiz==0){
			/* end */
			if (i==0){
				printf("(none)\n");
			}
			break;
		}
		printf("seg %04x: wbase %08x (%04x waves)\n",i,
			sblock+GET_2(mh.wavedir[i].off)*nbytes,
			ssiz*(nbytes/sizeof(struct mt_wave_s)));
	}
}

static int
print_wave(u_int wnr,int verbose)
{
	int i;
	struct mt_wave_s wbuf;
	u_int wbase;
	u_int off,cnt,cnttot;
	char fbuf[8]="CFL?EDOA";
	int fi;

	if (wnr>=newwave){
		printf("%04x: *** invalid wnr ***\n",wnr);
		return -1;
	}

	/* get from file */
	wbase=wnr2wbase(wnr);
	mdrmt_foff=wbase;
	if (mdrmt_foff==0){
		printf("%04x: *** invalid wnr ***\n",wnr);
		return -1;
	}
	if (readfile4(mdrmt_fp,(char *)&wbuf,mdrmt_foff,sizeof(struct mt_wave_s))<0){
		return -1;
	}
	mdrmt_foff+=sizeof(struct mt_wave_s);

	/* check if active */
	if ((wbuf.flags&MT_WAVE_FLAG_A)==0){
		/* not active */
		return 0;
	}

	printf("%04x:\n",wnr);
	if (verbose){
		printf("      wbase    %08x\n",wbase);
		printf("      nr       %04x\n",GET_2(wbuf.nr));
		printf("      flags    ");
#if 0
		printf("(%02x) ",wbuf.flags);
#endif
		for (fi=7;fi>=0;fi--){
			if (wbuf.flags&(1<<fi)){
				printf("%c",fbuf[fi]);
			}else{
				printf("-");
			}
		}
		printf("\n");
	}
	if (wbuf.file){
		printf("      library  %02x\n",wbuf.file);
	}else{
		printf("      library  (this file)\n");
	}
	if (verbose){
		printf("      used     %04x\n",GET_2(wbuf.used));
	}
	printf("      size     %08x\n",GET_4(wbuf.size));
	print_time("      birthday ",GET_4(wbuf.birthday));
	if (verbose){
		printf("      begin    %08x\n",GET_4(wbuf.begin));
#if 0
		printf("               (%08x %08x %08x %04x)\n",
			GET_4(wbuf.xxx4),
			GET_4(wbuf.xxx5),
			GET_4(wbuf.xxx6),
			GET_2(wbuf.xxx7));
#endif
	}

	if (verbose){
		/* scan segment-vector */
		cnttot=0;
		for (i=0;i<MT_WAVE_SEG_NR;i++){
			cnt=wbuf.seg[i].cnt;
			off=GET_2(wbuf.seg[i].off);
			if (cnt==0){
				break; /* end */
			}
			printf("      seg %02x @ blk %04x size %04x blks\n",
				i,off,cnt);
			cnttot+=cnt;
		}
		printf("                             %04x blks\n",
			cnttot);
	}

	return 0;
}


static int
print_clips(u_int cpos,u_int cbase,int verbose)
{
	struct mt_clip_s cbuf;
	u_int wnr,wnr2,flgs,flgs2;
	u_int cnt;
	u_int cnr;

	for (cnt=0;;cnt++){
		if (cnt>0){	
			/* next */
			cpos=GET_4(cbuf.fn);
			if (cpos==0){
				break; /* end of list */
			}
			cpos-=ebase;
			if (cpos>=nbytes){
				fprintf(stderr,"invalid cpos\n");
				return -1;
			}
		}

		/* get clip from file */
		mdrmt_foff=cbase+cpos;
		if (readfile4(mdrmt_fp,(char *)&cbuf,mdrmt_foff,sizeof(struct mt_clip_s))<0){
			return -1;
		}
		mdrmt_foff+=sizeof(struct mt_clip_s);
		
		/* get clip number */
		cnr=GET_2(cbuf.nr);
		if (cnr==0){
			/* not used */
			continue;
		}
		
		if (cnt==0){
			/* head */
			if (verbose){
				printf("    head @ %08x (%08x)\n",
					cbase+cpos,ebase+cpos);
			}
		}else{
			/* clip */
			if (verbose){
				printf("    clip @ %08x (%08x): %s\n",
					cbase+cpos,ebase+cpos,cbuf.name);
			}else{
				printf("    clip: %s\n",cbuf.name);
			}
		}
		if (verbose){
			printf("           fn %08x fp %08x rn %08x rp %08x\n",
				GET_4(cbuf.fn),
				GET_4(cbuf.fp),
				GET_4(cbuf.rn),
				GET_4(cbuf.rp));
		}
		
		if (cnt==0){
			/* head */
			if (verbose){
				printf("           nr       %04x (%04x)\n",
					cnr,
					GET_2(cbuf.xxx8));
				
				print_time("           head     ",GET_4(cbuf.head));
			}
		}else{
			/* clip */
			if (verbose){
				printf("           nr       %04x (%04x)\n",
					cnr,
					GET_2(cbuf.xxx8));
			}
			
			flgs=GET_2(cbuf.wave)&MT_CLIP_WFLAGS;
			flgs2=GET_2(cbuf.wave2)&MT_CLIP_WFLAGS2;
			if (verbose){
				printf("           flags    ");
#if 0
				printf("(%04x%04x) ",flgs,flgs2);
#endif
				if (flgs&MT_CLIP_FLAG_XXX4){
					printf("?");
				}else{
					printf("-");
				}
				if (flgs&MT_CLIP_FLAG_BEG){
					printf("B");
				}else{
					printf("-");
				}
				if (flgs&MT_CLIP_FLAG_STEREO){
					printf("S");
				}else{
					printf("-");
				}
				if (flgs&MT_CLIP_FLAG_LIB){
					printf("L");
				}else{
					printf("-");
				}
				if (flgs2&MT_CLIP_FLAG2_HEADFA){
					printf("H");
				}else{
					printf("-");
				}
				if (flgs2&MT_CLIP_FLAG2_TAILFA){
					printf("T");
				}else{
					printf("-");
				}
				if (flgs2&MT_CLIP_FLAG2_XXX2){
					printf("?");
				}else{
					printf("-");
				}
				if (flgs2&MT_CLIP_FLAG2_XXX1){
					printf("?");
				}else{
					printf("-");
				}
				printf("\n");
			}
			
			wnr=GET_2(cbuf.wave)&MT_CLIP_WAVE;
			wnr2=GET_2(cbuf.wave2)&MT_CLIP_WAVE2;
			printf("           wave     %04x\n",wnr);
			if (wnr>newwave){
				printf("                     *** invalid wnr ***\n");
			}
			if (flgs&MT_CLIP_FLAG_STEREO){
				printf("           wave2    %04x\n",wnr2);
				if (wnr2>newwave){
					printf("                     *** invalid wnr2 ***\n");
				}
			}
			/* XXX check if wave(s) active */
			
			print_time("           head     ",GET_4(cbuf.head));
			if (flgs2&MT_CLIP_FLAG2_HEADFA){
				print_time("           fadehead ",GET_4(cbuf.headfa));
			}
			if (flgs2&MT_CLIP_FLAG2_TAILFA){
				print_time("           fadetail ",GET_4(cbuf.tailfa));
			}
			print_time("           tail     ",GET_4(cbuf.tail));
			print_time("           sync     ",GET_4(cbuf.sync));
			printf("           begin    %f samples to sync\n",
				get_float(GET_4(cbuf.begin)));
			printf("           level    %i dB   (VCA %04x)\n",
				(int)(short)GET_2(cbuf.leveldb),
				GET_2(cbuf.levelvca));
		}
	}

	return 0;
}



static int
print_tracks(u_int cbase,int verbose)
{
	int i;
	struct mt_track_s tbuf;
	u_int cstart;

	for (i=0;i<MT_TRACK_NR;i++){
		/* get from file */
		mdrmt_foff=tbase+i*sizeof(struct mt_track_s);
		if (readfile4(mdrmt_fp,(char *)&tbuf,mdrmt_foff,sizeof(struct mt_track_s))<0){
			return -1;
		}
		mdrmt_foff+=sizeof(struct mt_track_s);

		if (verbose){
			printf("%02x  (%08x-%08x): %s\n",
				tbuf.nr,
				GET_4(tbuf.clisth),
				GET_4(tbuf.clistl),
				tbuf.name);
		}else{
			printf("%02x: %s\n",
				tbuf.nr,
				tbuf.name);
		}

		/* XXX no clips? */
		if (GET_4(tbuf.clistl)==0){
			continue; /* empty */
		}

#if 0
		printf("    (%02x %02x %02x)\n",
			tbuf.flags1,tbuf.xxx6,tbuf.flags2);
#endif
		print_time("    time   ",GET_4(tbuf.xxx2));

		cstart=GET_4(tbuf.clisth);
		if (cstart>0){
			cstart-=ebase;
			if (print_clips(cstart,cbase,verbose)<0){
				return -1;
			}
		}
	}

	return 0;
}



static int
print_libs(void)
{
	struct mt_lib_s ibuf;
	u_int i;
	u_int c;

	/* get from file */
	c=0;
	mdrmt_foff=MT_LIB_OFF;
	for (i=0;i<MT_LIB_NR;i++){
		if (readfile4(mdrmt_fp,(char *)&ibuf,mdrmt_foff,sizeof(struct mt_lib_s))<0){
			return -1;
		}
		mdrmt_foff+=sizeof(struct mt_lib_s);
		if (ibuf.flags&MT_LIB_FLAGS_USED){
			printf("%02x    name  : %s\n",ibuf.nr,ibuf.name);
			printf("      date  : %02i-%02i-%02i %02i:%02i\n",
				ibuf.year%100,
				ibuf.month,
				ibuf.day,
				ibuf.hour,
				ibuf.min);
			printf("      ID    : %08x\n",GET_4(ibuf.id));
			printf("      nbytes: %08x\n",GET_4(ibuf.nbytes));
			printf("      sblock: %08x\n",GET_4(ibuf.sblock));
			c++;
		}
	}
	if (c==0){
		printf("(none)\n");
	}

	return 0;
}



static int
print_marks9(void)
{
	u_int pos;
	struct mt_mark9_s mbuf;
	u_int nr;

	/* get from file */
	mdrmt_foff=markpos;
	for (pos=0;pos<marksize;pos+=sizeof(struct mt_mark9_s)){
		if (readfile4(mdrmt_fp,(char *)&mbuf,mdrmt_foff,sizeof(struct mt_mark9_s))<0){
			return -1;
		}
		mdrmt_foff+=sizeof(struct mt_mark9_s);
		nr=mbuf.nr;
		if (!(nr&MT_MARK9_FREE)){
			printf("%04x: %s\n",
				nr&MT_MARK11_NR,
				mbuf.name);
			print_time("      ",GET_4(mbuf.time));
#if 0
			printf("      (%08x %08x)\n",
				GET_4(mbuf.xxx1),
				GET_4(mbuf.xxx2));
#endif
		}
	}

	return 0;
}



static int
print_marks11(void)
{
	u_int pos;
	struct mt_mark11_s mbuf;
	u_int nr;

	/* get from file */
	mdrmt_foff=markpos;
	for (pos=0;pos<marksize;pos+=sizeof(struct mt_mark11_s)){
		if (readfile4(mdrmt_fp,(char *)&mbuf,mdrmt_foff,sizeof(struct mt_mark11_s))<0){
			return -1;
		}
		mdrmt_foff+=sizeof(struct mt_mark11_s);
		nr=GET_2(mbuf.nr);
		if (!(nr&MT_MARK11_FREE)){
			printf("%04x: %s\n",
				nr&MT_MARK11_NR,
				mbuf.name);
			print_time("      ",GET_4(mbuf.time));
#if 0
			printf("      (%08x %08x)\n",
				GET_4(mbuf.xxx1),
				GET_4(mbuf.xxx2));
#endif
		}
	}

	return 0;
}



static int
print_sat(void)
{
	u_int satpos,satsize;
	u_int remain,chunk;
	u_char buf[MDR_BLOCKSIZE];
	u_int i,j;
	u_int used,full;

	satpos=GET_4(mh.satpos);
	satsize=GET_4(mh.satsize);
	if ((satsize==0)||(satsize==MT_HEAD_ANCIENT_SATSIZE)
		||(GET_2(mh.ver)==MT_HEAD_VER_9_34)){
		/* XXX */
		satpos=MT_HEAD_ANCIENT_SATPOS;
		satsize=MT_HEAD_ANCIENT_SATSIZE;
	}

	mdrmt_foff=satpos;
	used=0;
	full=0;
	for (remain=satsize;remain>0;remain-=chunk){
		chunk=remain;
		if (chunk>MDR_BLOCKSIZE){
			chunk=MDR_BLOCKSIZE;
		}
		if (readfile4(mdrmt_fp,(char *)buf,mdrmt_foff,chunk)<0){
			printf("cannot read SAT\n");
			return -1;
		}
		mdrmt_foff+=chunk;
		for (i=0;i<chunk;i++){
			for (j=0;j<8;j++){
				if (((buf[i]>>j)&1)==0){
					used++;
				}
				full++;
			}
		}
	}

	printf("used %08x of %08x (%u%%)\n",
		used,full,
		(full==0)?100:(u_int)(100.0*((double)used)/((double)full)));

	return 0;
}



int
printinfo_mt(struct genfile_s *fp,int verbose)
{

	mdrmt_fp=fp;
	mdrmt_foff=0; /* begin of file */

	printf("MDR Disk Recorder Multi-Track file\n");

	if (readfile4(mdrmt_fp,(char *)&mh,mdrmt_foff,sizeof(struct mt_head_s))<0){
		return -1;
	}
	mdrmt_foff+=sizeof(struct mt_head_s);

	natfrm=GET_2(mh.natfrm);
	natsub=mh.natsub;

	spacebase=GET_4(mh.spacebase);
	if (spacebase==0){
		/* XXX */
		spacebase=MT_HEAD_ANCIENT_SPACEBASE;
	}
	sblock=GET_4(mh.sblock);
	if (sblock==0){
		/* XXX */
		sblock=MT_HEAD_ANCIENT_SBLOCK;
	}
	nbytes=GET_4(mh.nbytes);
	if (nbytes==0){
		printf("warning: invalid nbytes\n");
		nbytes=MT_HEAD_NBYTES11; /* XXX */
	}
	blksize=nbytes;
	ebase=GET_4(mh.ebase);
	tbase=MT_TRACK_OFF;

	newwave=GET_2(mh.newwave);

	printf("\n");
	printf("version     : %02x%02x (Rev. ",mh.ver[0],mh.ver[1]);
	switch (GET_2(mh.ver)){
	case MT_HEAD_VER_9_34:
		printf("9");
		break;
	case MT_HEAD_VER_11_39:
		printf("11");
		break;
	default:
		printf("???");
	}
	printf(")\n");
	PRINT_HEAD_1("nat fps     ",mh.natfps);
	if (verbose){
		PRINT_HEAD_2("nat frm     ",mh.natfrm);
		PRINT_HEAD_1("nat sub     ",mh.natsub);
		PRINT_HEAD_1("Q           ",mh.q);
		PRINT_HEAD_4("EBASE       ",mh.ebase);
		PRINT_HEAD_4("NBYTES      ",mh.nbytes);
		PRINT_HEAD_2("newwave     ",mh.newwave);
	}
	printf("fend        : %08x\n",sblock+GET_2(mh.fend)*nbytes);
	PRINT_HEAD_4("sample rate ",mh.samprate);
	if (verbose){
		PRINT_HEAD_4("SPACE BASE  ",mh.spacebase);
		PRINT_HEAD_4("SBLOCK      ",mh.sblock);
	}
	PRINT_HEAD_4("ID          ",mh.id);
	if (verbose){
		PRINT_HEAD_4("MARK SIZE   ",mh.marksize);
		PRINT_HEAD_2("marks       ",mh.marks);
		PRINT_HEAD_4("MARK POS    ",mh.markpos);
		PRINT_HEAD_4("SAT POS     ",mh.satpos);
		PRINT_HEAD_4("SAT SIZE    ",mh.satsize);
		mdrmt_print_passwd("password    : ",mh.permpasswd);
	}
	printf("public read : %s\n",
		(mh.permpasswd[0]&MT_HEAD_PERM_PREAD)?"no":"yes");
	printf("public write: %s\n",
		(mh.permpasswd[0]&MT_HEAD_PERM_PWRITE)?"no":"yes");
	printf("\n");

	if (verbose){
		/* wave segment directory */
		printf("wave segment directory:\n");
		print_wavedir();
		printf("\n");

		/* SAT */
		printf("SAT:\n");
		if (print_sat()<0){
			return -1;
		}
		printf("\n");
	}

	/* libraries */
	printf("libraries:\n");
	if (print_libs()<0){
		return -1;
	}
	printf("\n");

	/* waves */
	printf("waves:\n");
	{
		u_int i;

		for (i=0;i<newwave;i++){
			print_wave(i,verbose);
		}
	}
	printf("\n");

	/* tracks */
	printf("tracks:\n");
	if (print_tracks(sblock,verbose)<0){
		return -1;
	}
	printf("\n");

#if 0
	if (verbose){
		/* edit-list */
		printf("edit-list:\n");
		if (print_clips(0,sblock,verbose)<0){
			return -1;
		}
		printf("\n");
		printf("space edit-list:\n");
		if (print_clips(0,sblock+spacebase,verbose)<0){
			return -1;
		}
		printf("\n");
	}
#endif

	/* marks */
	markpos=GET_4(mh.markpos);
	marksize=GET_4(mh.marksize);
	if ((marksize==0)||(marksize==MT_HEAD_ANCIENT_MARKSIZE)
		||(GET_2(mh.ver)==MT_HEAD_VER_9_34)){
		/* XXX */
		markpos=MT_HEAD_ANCIENT_MARKPOS;
		marksize=MT_HEAD_ANCIENT_MARKSIZE;
		printf("marks (Rev. 9):\n");
		if (print_marks9()<0){
			return -1;
		}
	}else{
		if (marksize%sizeof(struct mt_mark11_s)!=0){
			fprintf(stderr,"invalid marksize\n");
			return -1;
		}else{
			printf("marks (Rev. 11):\n");
			if (print_marks11()<0){
				return -1;
			}
		}
	}
	printf("\n");

	return 0;
}



static void
print_progress(u_int blk,int first)
{
	static u_int blkend=0;  /* must be static!!! */
	static double xold=0.0; /* must be static!!! */
	
	if (first){
		if (blk==0){
			/* done */
			printf(".\ndone\n");
			return;
		}
		
		blkend=blk;
		xold=0.0;
		printf("0         20        40        60        80        100%%\n.");
		fflush(NULL);
		return;
	}
	
	if (blkend==0){
		/* not correctly initialized */
		return;
	}
	
	for (;((double)blk)/(double)blkend-xold>=0.02;xold+=0.02){
		printf(".");
		fflush(NULL);
	}
}




static int
lseek_wave(struct mt_wave_s *wp,u_int skip,u_int *retoff,u_int *retsiz)
{
	int i;
	u_int wstart;
	u_int remain;
	u_int ssiz;
	int found;
	
	if (wp==NULL){
		return -1;
	}
	
	/* remaining total waveform size */
	remain=GET_4(wp->size);
	if (skip>remain){
		return -1;
	}
	
	/* correct for actual waveform start */
	wstart=GET_4(wp->begin);
	remain+=wstart;
	skip+=wstart;
	
	/* scan segments */
	found=0;
	for (i=0;i<MT_WAVE_SEG_NR;i++){
		/* current segment */
		ssiz=nbytes*(wp->seg[i].cnt);
		if (remain<ssiz){
			/* segment not fully used */
			ssiz=remain;
		}
#ifdef DEBUG
		printf("skip %08x ssiz %08x remain %08x seg %02x @ blk %04x size %04x blks\n",
			skip,
			ssiz,
			remain,
			i,
			GET_2(wp->seg[i].off),
			wp->seg[i].cnt);
#endif
		if (ssiz==0){
			/* end of used segements */
			if (skip==0){
				/* ssiz==0 here! */
				/* now at end+1 of wave => 0 bytes for retsize */
				found=1;
			}
			break;
		}
		
		/* check if this is wanted segment */
		if (skip<ssiz){
			found=1;
			break; /* done */
		}
		
		/* remaining offset */
		skip-=ssiz;
		/* remaining waveform */
		remain-=ssiz;
	}
	if (!found){
		return -1;
	}
	
	if (retoff!=NULL){
		*retoff=sblock+nbytes*GET_2(wp->seg[i].off)+skip;
	}
	if (retsiz!=NULL){
		*retsiz=ssiz-skip;
	}
	
	return 0;
}



static int
convert_wave_mono(struct genfile_s *fp,int outdes,
				  struct mt_wave_s *wbuf,
				  u_int start,u_int size,
				  char *buf,u_int bufsiz,int verbose)
{
	u_char *blkbuf;
	int blkbufmalloc;
	u_int totsize;
	u_int remain,curpos;
	u_int soff,ssiz;
	u_int chunk,csiz,cpos;
	u_int foff;
	int ret;

	if ((fp==NULL)||(wbuf==NULL)){
		return -1;
	}
	
	totsize=GET_4(wbuf->size);
#ifdef DEBUG
	printf("size    : %08x\n",totsize);
#endif
	
	/* waveform block to get */
	if (start>=totsize){
		printf("invalid start\n");
		return -1;
	}
	if (size==0){
		/* take rest */
		size=totsize-start;
	}
	if ((start+size)>totsize){
		printf("warning: invalid size; using rest.\n");
		/* take rest */
		size=totsize-start;
	}
	remain=size;
	curpos=start;
	
	if ((buf==NULL)||(bufsiz<nbytes)){
		/* allocate block buffer */
		if ((blkbuf=(u_char *)malloc(nbytes))==NULL){
			printf("cannot allocate block buffer\n");
			return -1;
		}
		blkbufmalloc=1;
	}else{
		blkbuf=(u_char *)buf;
		blkbufmalloc=0;
	}
	
	ret=-1;

	/* create and write WAV-header */
	if (wav_write_head(outdes,remain,1,samprate,16)<0){ /* 1==mono */
		printf("cannot write WAV-header\n");
		goto convert_wave_exit;
	}
	
	/* convert */
	if (verbose){
		print_progress(remain,1);
	}
	for (;remain>0;){
#ifdef DEBUG
		printf("curpos %08x remain %08x\n",curpos,remain);
#endif
		/* goto curpos */
		if (lseek_wave(wbuf,curpos,&soff,&ssiz)<0){
			printf("error in lseek_wave\n");
			goto convert_wave_exit;
		}
		foff=soff;
#ifdef DEBUG
		printf("soff %08x ssiz %08x\n",soff,ssiz);
#endif
		
		/* split ssiz into buffer chunks */
		if (ssiz==0){
			printf("premature end of wave\n");
			goto convert_wave_exit;
		}
		if (remain<ssiz){
			ssiz=remain;
#ifdef DEBUG
			printf("truncate ssiz %08x\n",ssiz);
#endif
		}
		for (chunk=ssiz,cpos=0;chunk>0;){
			csiz=nbytes;
			if (chunk<csiz){
				csiz=chunk;
			}
#ifdef DEBUG
			printf("chunk %08x csiz %08x\n",
				chunk,csiz);
#endif
			
			/* read block */
			if (readfile4(fp,(char *)blkbuf,foff,csiz)<0){
				printf("cannot read file\n");
				goto convert_wave_exit;
			}
			foff+=csiz;
			
			/* swap Byte-order */
			/* XXX csiz should be multiples of 2! */
			{
				u_int k;
				u_char b;
				
				for (k=0;k<csiz;k+=2){
					b=blkbuf[k];
					blkbuf[k]=blkbuf[k+1];
					blkbuf[k+1]=b;
				}
			}
			
			/* write block */
			if (WRITE(outdes,blkbuf,csiz)!=(int)csiz){
				printf("cannot write file\n");
				goto convert_wave_exit;
			}
			chunk-=csiz;
			cpos+=csiz;
			if (verbose){
				print_progress(curpos+cpos,0);
			}
		}
		
		remain-=ssiz;
		curpos+=ssiz;
	}
	if (verbose){
		print_progress(0,1);
	}
	ret=0;
	
convert_wave_exit:
	if (blkbufmalloc){
		free(blkbuf);
	}
	return ret;
}



static int
convert_wave(struct genfile_s *fp,int outdes,
			 struct mt_wave_s *wbuf1,struct mt_wave_s *wbuf2,
			 u_int start,u_int size,
			 char *buf,u_int bufsiz,int verbose)
{
	u_char *blkbuf1;
	u_char *blkbuf2;
	u_char *blkbuf;
	int blkbufmalloc;
	u_int totsize;
	u_int totsize1;
	u_int totsize2;
	u_int remain,curpos;
	u_int soff1,ssiz1;
	u_int soff2,ssiz2;
	u_int ssiz;
	u_int chunk,csiz,cpos;
	u_int foff;
	int ret;

	if ((fp==NULL)||(wbuf1==NULL)){
		return -1;
	}

	if (wbuf2==NULL){
		return convert_wave_mono(fp,outdes,wbuf1,start,size,buf,bufsiz,verbose);
	}

	/* Note: now, stereo */
	
	totsize1=GET_4(wbuf1->size);
#ifdef DEBUG
	printf("size1   : %08x\n",totsize1);
#endif
	totsize2=GET_4(wbuf2->size);
#ifdef DEBUG
	printf("size2   : %08x\n",totsize2);
#endif
	
	/* waveform block to get */
	if (totsize1!=totsize2){
		printf("warning: waves have different size; using smallest.\n");
	}
	/* take smallest size */
	totsize=totsize1;
	if (totsize1>totsize2){
		totsize=totsize2;
	}
	if (start>=totsize){
		printf("invalid start\n");
		return -1;
	}
	if (size==0){
		/* take rest */
		size=totsize-start;
	}
	if ((start+size)>totsize){
		printf("warning: invalid size; using rest.\n");
		/* take rest */
		size=totsize-start;
	}
	remain=size;
	curpos=start;
	
	if ((buf==NULL)||(bufsiz<4*nbytes)){ /* 4* !!! */
		/* allocate block buffers */
		if ((blkbuf=(u_char *)malloc(4*nbytes))==NULL){
			printf("cannot allocate block buffer\n");
			return -1;
		}
		blkbufmalloc=1;
	}else{
		blkbuf=(u_char *)buf;
		blkbufmalloc=0;
	}
	blkbuf1=blkbuf+2*nbytes;
	blkbuf2=blkbuf1+nbytes;
	
	ret=-1;

	/* create and write WAV-header */
	if (wav_write_head(outdes,remain,2,samprate,16)<0){ /* 2==stereo */
		printf("cannot write WAV-header\n");
		goto convert_wave_stereo_exit;
	}
	
	/* convert */
	if (verbose){
		print_progress(remain,1);
	}
	for (;remain>0;){
#ifdef DEBUG
		printf("curpos %08x remain %08x\n",curpos,remain);
#endif
		/* goto curpos */
		/* wave side 1 */
#ifdef DEBUG
		printf("wave 1:\n");
#endif
		if (lseek_wave(wbuf1,curpos,&soff1,&ssiz1)<0){
			fprintf(stderr,"lseek_wave\n");
			goto convert_wave_stereo_exit;
		}
		/* wave side 2 */
#ifdef DEBUG
		printf("wave 2:\n");
#endif
		if (lseek_wave(wbuf2,curpos,&soff2,&ssiz2)<0){
			fprintf(stderr,"lseek_wave\n");
			goto convert_wave_stereo_exit;
		}
#ifdef DEBUG
		printf("soff1 %08x ssiz1 %08x soff2 %08x ssiz2 %08x\n",
			soff1,ssiz1,soff2,ssiz2);
#endif
		
		/* take smallest size */
		ssiz=ssiz1;
		if (ssiz1>ssiz2){
			ssiz=ssiz2;
		}
#ifdef DEBUG
		if (ssiz1!=ssiz2){
			printf("truncate ssiz %08x\n",ssiz);
		}
#endif
		
		/* split ssiz into buffer chunks */
		if (ssiz==0){
			fprintf(stderr,"premature end of wave\n");
			goto convert_wave_stereo_exit;
		}
		if (remain<ssiz){
			ssiz=remain;
#ifdef DEBUG
			printf("truncate ssiz %08x\n",ssiz);
#endif
		}
		for (chunk=ssiz,cpos=0;chunk>0;){
			csiz=nbytes;
			if (chunk<csiz){
				csiz=chunk;
			}
#ifdef DEBUG
			printf("chunk %08x csiz %08x\n",
				chunk,csiz);
#endif
			
			/* wave side 1 */
			foff=soff1+cpos;
			/* read block */
			if (readfile4(fp,(char *)blkbuf1,foff,csiz)<0){
				printf("cannot read file\n");
				goto convert_wave_stereo_exit;
			}
			foff+=csiz;
			
			/* wave side 2 */
			foff=soff2+cpos;
			/* read block */
			if (readfile4(fp,(char *)blkbuf2,foff,csiz)<0){
				printf("cannot read file\n");
				goto convert_wave_stereo_exit;
			}
			foff+=csiz;
			
			/* merge and swap Byte-order */
			/* XXX csiz should be multiples of 2! */
			{
				u_int k;
				
				for (k=0;k<csiz;k+=2){
					blkbuf[2*k+0]=blkbuf1[k+1];
					blkbuf[2*k+1]=blkbuf1[k+0];
					blkbuf[2*k+2]=blkbuf2[k+1];
					blkbuf[2*k+3]=blkbuf2[k+0];
				}
			}
			
			/* write block */
			if (WRITE(outdes,blkbuf,2*csiz)!=(int)(2*csiz)){
				printf("cannot write file\n");
				goto convert_wave_stereo_exit;
			}
			chunk-=csiz;
			cpos+=csiz;
			if (verbose){
				print_progress(curpos+cpos,0);
			}
		}
		
		remain-=ssiz;
		curpos+=ssiz;
	}
	if (verbose){
		print_progress(0,1);
	}
	ret=0;
	
convert_wave_stereo_exit:
	if (blkbufmalloc){
		free(blkbuf);
	}
	return ret;
}



#ifdef MT2WAVALL_NODUPLICATES
static char mt2wav_all_wdone[MT_HEAD_MAXWAVENR];
#endif

static int
clips2wav(char *wname,char *buf,u_int bufsiz,
		  u_int cpos,u_int cbase,
		  int tr,u_int wnr1,u_int wnr2)
{
	struct mt_clip_s cbuf;
	u_int flgs;
	u_int cnt;
	u_int cnr;
	int stereo;
	struct mt_wave_s ws1;
	struct mt_wave_s ws2;
	struct mt_wave_s *wbuf1;
	struct mt_wave_s *wbuf2;
	char *wwname;
	u_int wnamelen;
	int outdes;
	int ret;

	/* XXX no check for fp,wname,buf: done by mt2wav_all()!!! */
	wwname=buf;
	wnamelen=strlen(wname);

	for (cnt=0;;cnt++){
		if (tr>=0){
			if (cnt>0){	
				/* next */
				cpos=GET_4(cbuf.fn);
				if (cpos==0){
					break; /* end of list */
				}
				cpos-=ebase;
				if (cpos>=nbytes){
					fprintf(stderr,"invalid cpos\n");
					return -1;
				}
			}

			/* get clip from file */
			mdrmt_foff=cbase+cpos;
			if (readfile4(mdrmt_fp,(char *)&cbuf,mdrmt_foff,sizeof(struct mt_clip_s))<0){
				return -1;
			}
			mdrmt_foff+=sizeof(struct mt_clip_s);
			
			/* get clip number */
			cnr=GET_2(cbuf.nr);
			if (cnr==0){
				/* not used */
				continue;
			}
			if (cnt==0){
				/* head */
				continue;
			}

			/* get wave numbers */
			wnr1=GET_2(cbuf.wave)&MT_CLIP_WAVE;
			wbuf1=&ws1;
			flgs=GET_2(cbuf.wave)&MT_CLIP_WFLAGS;
			if (flgs&MT_CLIP_FLAG_STEREO){
				stereo=1;
				wnr2=GET_2(cbuf.wave2)&MT_CLIP_WAVE2;
				wbuf2=&ws2;
			}else{
				stereo=0;
				wnr2=wnr1;
				wbuf2=NULL;
			}
			
			if (stereo){
				printf("track %i, clip %u: converting stereo wave (%u,%u)\n",tr,cnr,wnr1,wnr2);
			}else{
				printf("track %i, clip %u: converting mono wave (%u)\n",tr,cnr,wnr1);
			}

#ifdef MT2WAVALL_NODUPLICATES
			/* already converted? */
			if (mt2wav_all_wdone[wnr1]){
				printf("already exported, skipping\n");
				continue; /* skip */
			}
			/* XXX no need to check wnr2 since wnr1 covers stereo */
			
			/* mark done */
			mt2wav_all_wdone[wnr1]=1;
			mt2wav_all_wdone[wnr2]=1; /* XXX actually not needed */
#endif
			
		}else{
			/* user-supplied wnr1,wnr2 */
			
			wbuf1=&ws1;
			if (wnr2==wnr1){
				/* mono */
				wbuf2=NULL;
			}else{
				/* stereo */
				wbuf2=&ws2;
			}
		}
		/* check wave numbers */
		if (wnr1>=newwave){ /* XXX */
			printf("invalid wave 1\n");
			return -1;
		}
		if (wnr2>=newwave){ /* XXX */
			printf("invalid wave 2\n");
			return -1;
		}
				
		/* get wave1 struct from file */
		mdrmt_foff=wnr2wbase(wnr1);
		if (mdrmt_foff==0){
			printf("invalid wave 1\n");
			return -1;
		}
		if (readfile4(mdrmt_fp,(char *)wbuf1,mdrmt_foff,sizeof(struct mt_wave_s))<0){
			printf("cannot read wave struct\n");
			if (tr>=0){
				/* XXX ignore error */
				continue;
			}else{
				/* user-supplied wnr1,wnr2 */
				return -1;
			}
		}
		mdrmt_foff+=sizeof(struct mt_wave_s);
		if (wbuf1->file){
			printf("borrowed wave, skipping\n");
			if (tr>=0){
				/* XXX ignore error */
				continue;
			}else{
				/* user-supplied wnr1,wnr2 */
				return -1;
			}
		}
		/* check if active */
		if ((wbuf1->flags&MT_WAVE_FLAG_A)==0){
			printf("inactive wave, skipping\n");
			if (tr>=0){
				/* XXX ignore error */
				continue;
			}else{
				/* user-supplied wnr1,wnr2 */
				return -1;
			}
		}
		if (wbuf2!=NULL){
			/* get wave2 struct from file */
			mdrmt_foff=wnr2wbase(wnr2);
			if (mdrmt_foff==0){
				printf("invalid wave 2\n");
				return -1;
			}
			if (readfile4(mdrmt_fp,(char *)wbuf2,mdrmt_foff,sizeof(struct mt_wave_s))<0){
				printf("cannot read wave 2 struct\n");
				if (tr>=0){
					/* XXX ignore error */
					continue;
				}else{
					/* user-supplied wnr1,wnr2 */
					return -1;
				}
			}
			mdrmt_foff+=sizeof(struct mt_wave_s);
			if (wbuf2->file){
				printf("borrowed wave 2, cannot read\n");
				if (tr>=0){
					/* XXX ignore error */
					continue;
				}else{
					/* user-supplied wnr1,wnr2 */
					return -1;
				}
			}
			/* check if active */
			if ((wbuf2->flags&MT_WAVE_FLAG_A)==0){
				printf("inactive wave 2, skipping\n");
				if (tr>=0){
					/* XXX ignore error */
					continue;
				}else{
					/* user-supplied wnr1,wnr2 */
					return -1;
				}
			}
		}
		
		/* name of output file */
		strcpy(wwname,wname);
		if (tr>=0){
			if ((wnamelen>=4)&&(strcasecmp(wname+wnamelen-4,".wav")==0)){
				if (stereo){
					sprintf(wwname+wnamelen-4,"_T%02iC%05u_W%05uS%05u%s",tr,cnr,wnr1,wnr2,wname+wnamelen-4);
				}else{
					sprintf(wwname+wnamelen-4,"_T%02iC%05u_W%05u%s",tr,cnr,wnr1,wname+wnamelen-4);
				}
			}else{
				if (stereo){
					sprintf(wwname+wnamelen,"_T%02iC%05u_W%05uS%05u.wav",tr,cnr,wnr1,wnr2);
				}else{
					sprintf(wwname+wnamelen,"_T%02iC%05u_W%05u.wav",tr,cnr,wnr1);
				}
			}
		}
		
		/* open output file */
		/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
		if ((outdes=OPEN(wwname,O_RDWR|O_CREAT|O_TRUNC|O_BINARY,0666))<0){
			perror("open");
			if (tr>=0){
				/* XXX ignore error */
				continue;
			}else{
				/* user-supplied wnr1,wnr2 */
				return -1;
			}
		}
		
		/* convert */
		ret=convert_wave(mdrmt_fp,outdes,wbuf1,wbuf2,0,0,buf,bufsiz,0);
		
		CLOSE(outdes);
		
		if (tr<0){
			/* user-supplied wnr1,wnr2 */
			return ret;
		}
		/* XXX ignore error in loop */
	}

	return 0;
}

static int
mt2wav_all(struct genfile_s *fp,char *wname,char *buf,u_int bufsiz)
{
	int i;
	struct mt_track_s tbuf;
	u_int cstart;
	u_int wnamelen;

	if ((fp==NULL)||(wname==NULL)||(buf==NULL)){
		return -1;
	}

	mdrmt_fp=fp;

	/* XXX use buf for name as well */
	wnamelen=strlen(wname);
	if (bufsiz<(wnamelen+27)){ /* need space for "_TxxCxxxxx_WxxxxxSxxxxx.wav" */
		fprintf(stderr,"mt2wav_all: bufsize too small\n");
		return -1;
	}

	/* read MT-file */
	mdrmt_foff=0; /* begin of file */
	if (readfile4(mdrmt_fp,(char *)&mh,mdrmt_foff,sizeof(struct mt_head_s))<0){
		fprintf(stderr,"cannot read file\n");
		return -1;
	}
	mdrmt_foff+=sizeof(struct mt_head_s);

	/* XXX check header for valid MT-file */
	
	natfrm=GET_2(mh.natfrm);
	natsub=mh.natsub;
	
	samprate=GET_4(mh.samprate);
	
	spacebase=GET_4(mh.spacebase);
	if (spacebase==0){
		/* XXX */
		spacebase=MT_HEAD_ANCIENT_SPACEBASE;
	}
	sblock=GET_4(mh.sblock);
	if (sblock==0){
		/* XXX */
		sblock=MT_HEAD_ANCIENT_SBLOCK;
	}
	nbytes=GET_4(mh.nbytes);
	if (nbytes==0){
		printf("warning: invalid nbytes\n");
		nbytes=MT_HEAD_NBYTES11; /* XXX */
	}
	blksize=nbytes;
	ebase=GET_4(mh.ebase);
	tbase=MT_TRACK_OFF;
	
	newwave=GET_2(mh.newwave);
	if (newwave==0){
		printf("no waves\n");
		return -1;
	}

#ifdef MT2WAVALL_NODUPLICATES
	/* mark not done yet */
	for (i=0;i<(int)newwave;i++){
		mt2wav_all_wdone[i]=0;
	}
#endif

	for (i=0;i<MT_TRACK_NR;i++){
		/* get from file */
		mdrmt_foff=tbase+i*sizeof(struct mt_track_s);
		if (readfile4(fp,(char *)&tbuf,mdrmt_foff,sizeof(struct mt_track_s))<0){
			fprintf(stderr,"cannot read track\n");
			return -1;
		}
		mdrmt_foff+=sizeof(struct mt_track_s);

		/* XXX no clips? */
		if (GET_4(tbuf.clistl)==0){
			continue; /* empty */
		}

		cstart=GET_4(tbuf.clisth);
		if (cstart>0){
			cstart-=ebase;
			if (clips2wav(wname,buf,bufsiz,cstart,sblock,i+1,0,0)<0){
				return -1;
			}
		}
	}

	return 0;
}

int
mt2wav(struct genfile_s *fp,char *wname,
	   u_int wnr1,u_int wnr2,
	   u_int start,u_int size,
	   char *buf,u_int bufsiz)
{
	int allflag;
	u_int wnrmin,wnrmax;
	char *wwname;
	u_int wnamelen;
	int ret;

	if ((fp==NULL)||(wname==NULL)||(buf==NULL)){
		return -1;
	}

	if (wnr1==(u_int)-1){
		allflag=1; /* all waves */
	}else{
		allflag=0; /* just wnr1 (and wnr2 if stereo) */
	}
	
#ifdef MT2WAVALL
	if (allflag){
		return mt2wav_all(fp,wname,buf,bufsiz);
	}
#endif

	mdrmt_fp=fp;

	/* XXX use buf for name as well */
	wnamelen=strlen(wname);
	if (bufsiz<(wnamelen+11)){ /* need space for "_Wxxxxx.wav" */
		fprintf(stderr,"mt2wav: bufsize too small\n");
		return -1;
	}

	/* make multiples of 2 */
	start=2*(start/2);
	size=2*(size/2);
	
	/* read MT-file */
	mdrmt_foff=0; /* begin of file */
	if (readfile4(fp,(char *)&mh,mdrmt_foff,sizeof(struct mt_head_s))<0){
	}
	mdrmt_foff+=sizeof(struct mt_head_s);

	/* XXX check header for valid MT-file */
	
	natfrm=GET_2(mh.natfrm);
	natsub=mh.natsub;
	
	samprate=GET_4(mh.samprate);
	
	spacebase=GET_4(mh.spacebase);
	if (spacebase==0){
		/* XXX */
		spacebase=MT_HEAD_ANCIENT_SPACEBASE;
	}
	sblock=GET_4(mh.sblock);
	if (sblock==0){
		/* XXX */
		sblock=MT_HEAD_ANCIENT_SBLOCK;
	}
	nbytes=GET_4(mh.nbytes);
	if (nbytes==0){
		printf("warning: invalid nbytes\n");
		nbytes=MT_HEAD_NBYTES11; /* XXX */
	}
	blksize=nbytes;
	ebase=GET_4(mh.ebase);
	tbase=MT_TRACK_OFF;
	
	newwave=GET_2(mh.newwave);
	if (newwave==0){
		printf("no waves\n");
		return -1;
	}

	if (allflag==0){
		/* check wave numbers */
		if (wnr1>=newwave){ /* XXX */
			printf("invalid wave 1\n");
			return -1;
		}
		if (wnr2>=newwave){ /* XXX */
			printf("invalid wave 2\n");
			return -1;
		}
	}

	/* scan waves */
	if (allflag){
		wnrmin=0;
		wnrmax=newwave-1;
		wwname=buf;
		start=0;
		size=0; /* means full size */
	}else{
		wnrmin=wnr1;
		wnrmax=wnr1;
		wwname=wname;
	}
	for (wnr1=wnrmin;wnr1<=wnrmax;wnr1++){
		if (allflag){
			/* name of output file */
			strcpy(wwname,wname);
			if ((wnamelen>=4)&&(strcasecmp(wname+wnamelen-4,".wav")==0)){
				sprintf(wwname+wnamelen-4,"_W%05u%s",wnr1,wname+wnamelen-4);
			}else{
				sprintf(wwname+wnamelen,"_W%05u.wav",wnr1);
			}
			wnr2=wnr1; /* XXX force to mono */
		}

		/* convert */
		if (wnr1==wnr2){
			printf("converting mono wave (%u)\n",wnr1);
		}else{
			printf("converting stereo wave (%u,%u)\n",wnr1,wnr2);
		}
		ret=clips2wav(wwname,buf,bufsiz,0,0,-1,wnr1,wnr2); /* -1: user-supplied wnr1,wnr2 */
		/* XXX ignore error in loop */
	}
	if (allflag){
		ret=0; /* XXX ignore error */
	}

	return ret;
}



/* EOF */
