#ifndef __CMIOS9_MDR_H
#define __CMIOS9_MDR_H
/*
* Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008  K. M. Indlekofer
*
* m.indlekofer@gmx.de
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



/* Note: time-unit (TU) is 1/48000 sec */
#define MT_TU_SEC  0x0000bb80 /* 1 sec in TU */
#define MT_TU_MIN  0x002bf200 /* 1 min in TU */
#define MT_TU_HOUR 0x0a4cb800 /* 1 hour in TU */
#define MT_TU_DAY  0xf7314000 /* 1 day in TU */



/* .MT-file header */
/* XXX must be 0x100 Bytes */
struct mt_head_s{
#define MT_HEAD_VER_9_34  0x000b /* Rev. 9.34 */
#define MT_HEAD_VER_11_39 0x0016 /* Rev. 11.39 */
	u_char ver[2];         /* version */
	u_char natfps;         /* nat fps */
	u_char xxx1;
	u_char natfrm[2];      /* nat frm in TU */
	u_char natsub;         /* nat sub in TU */
	u_char q;              /* Q */
	u_char ebase[4];       /* EBASE */
#define MT_HEAD_NBYTES11  0x20000 /* bytes per "block" */
	u_char nbytes[4];      /* NBYTES */
	u_char xxx2[2];
	u_char zero1[60];
#define MT_HEAD_MAXWAVENR 0x10000 /* XXX */
	u_char newwave[2];     /* next available wave number ??? */
	/* wave segment directory */
#define MT_HEAD_WAVEDIRNR 36
	struct{                /* segments for waves */
		u_char ssiz;       /* number of blocks in segment (end==0) */
		u_char off[2];     /* physical block offset (of nbytes) */
	}wavedir[MT_HEAD_WAVEDIRNR];
	u_char fend[2];        /* file end in blocks (of nbytes) rel. to sblock ??? */
	u_char samprate[4];    /* sample rate in Hz */
	u_char xxx7[4];
	u_char xxx8;
	u_char xxx9;
	u_char xxx10;
	u_char xxx11;
#define MT_HEAD_ANCIENT_SPACEBASE 0x00010000
	u_char spacebase[4];   /* SPACE BASE */
#define MT_HEAD_ANCIENT_SBLOCK    0x0000bf00
	u_char sblock[4];      /* SBLOCK */
	u_char id[4];          /* ID */
#define MT_HEAD_ANCIENT_MARKSIZE  0x00000d6a
	u_char marksize[4];    /* MARK SIZE */
	u_char marks[2];       /* marks??? */
#define MT_HEAD_ANCIENT_MARKPOS   0x00009700
	u_char markpos[4];     /* MARK POS */
#define MT_HEAD_ANCIENT_SATPOS    0x00006f00
	u_char satpos[4];      /* SAT POS */
#define MT_HEAD_ANCIENT_SATSIZE   0x00002800
	u_char satsize[4];     /* SAT SIZE */
	u_char zero3[6];
#define MT_HEAD_PERM_FLAGS  0xf0 /* mask */
#define MT_HEAD_PERM_PREAD  0x80 /* public read deny flag */
#define MT_HEAD_PERM_PWRITE 0x40 /* public write deny flag */
#define MT_HEAD_PERM_XXX2   0x20
#define MT_HEAD_PERM_XXX1   0x10
#define MT_HEAD_PASS_NBITS  0x8c /* number of Bits in password Bit-field */
#define MT_HEAD_PASS_BIT0   0x04 /* start of password Bit-field */
#define MT_HEAD_PASS_SZ     0x14 /* max. password size in Bytes */
	u_char permpasswd[18]; /* permission- and password-Bit-field */
};



/* libraries (import projects) */
/* should be 0x50 Bytes */
struct mt_lib_s{
#define MT_LIB_FLAGS_USED 0x80 /* entry used */
	u_char flags;     /* flags */
	u_char nr;        /* incremental number (0x01..0xff) */
	u_char xxx1;
	u_char year;      /* date: year */
	u_char month;     /*       month */
	u_char day;       /*       day */
	u_char hour;      /*       hour */
	u_char min;       /*       minute */
	char name[24];    /* last name of import project (\0-terminated) */
	u_char xxx2[28];
	u_char nbytes[4]; /* NBYTES of import project */
	u_char sblock[4]; /* SBLOCK of import project */
	u_char id[4];     /* ID of import project */
	u_char xxx3[8];
};
#define MT_LIB_OFF 0x00000100 /* offset to libraries-vector */
#define MT_LIB_NR  0xff       /* number of libraries */



/* track */
/* should be 0x100 Bytes */
struct mt_track_s{
#define MT_TRACK_FLAGS1_XXX 0x02 /* ??? */
	u_char flags1;    /* ??? flags */
	u_char nr;        /* number */
	char name[16];    /* name (\0-terminated) */
	u_char clisth[4]; /* pointer to clip-list head */
	u_char zero1[8];
	u_char xxx2[4];   /* ??? time in TU */
	u_char xxx3[4];   /* XXX must be 0xffffffff */
	u_char clistl[4]; /* pointer to last used clip in clip-list */
	u_char xxx4[4];
	u_char xxx5[4];
	u_char xxx6;      /* incremental number ??? */
#define MT_TRACK_FLAGS2_XXX1 0x01 /* ??? */
#define MT_TRACK_FLAGS2_XXX2 0x02 /* ??? */
#define MT_TRACK_FLAGS2_XXX3 0x04 /* ??? */
#define MT_TRACK_FLAGS2_XXX4 0x08 /* ??? */
	u_char flags2;    /* ??? flags */
	u_char xxx8[51][4];
};
#define MT_TRACK_OFF 0x00005100 /* offset to track-vector */
#define MT_TRACK_NR  24         /* number of tracks */



/* clip */
/* XXX should be 0x80 Bytes */
struct mt_clip_s{
	u_char fn[4];     /* forward-link next */
	u_char fp[4];     /* forward-link previous */
	u_char rn[4];     /* reverse-link next */
	u_char rp[4];     /* reverse-link previous */
#define MT_CLIP_WFLAGS      0xf000 /* wave flags */
#define MT_CLIP_FLAG_LIB    0x1000 /* uses library flag */
#define MT_CLIP_FLAG_STEREO 0x2000 /* stereo flag */
#define MT_CLIP_FLAG_BEG    0x4000 /* begin!=0 flag */
#define MT_CLIP_FLAG_XXX4   0x8000 /* ??? flag */
#define MT_CLIP_WAVE        0x0fff /* wave number mask */
	u_char wave[2];   /* wave number and flags */
	                  /* if Stereo: side2 wave number is in wave2 */
	                  /* Note: waves of stereo-clip need not have */
	                  /*       same segment-layout!!! */
	u_char head[4];   /* head absolute time in TU */
	u_char tail[4];   /* tail absolute time in TU */
	u_char sync[4];   /* sync absolute time in TU */
	                  /* Note: default: sync==head */
	u_char begin[4];  /* sync-time - waveform-begin-time in samples */
	                  /* Note: 68881 s-float format */
	                  /*       => sync must not be to far away from begin */
	                  /*          in order not to lose sync-accuracy! */
	                  /*       (see also mdrwsmain11 33ff2!!!) */
	                  /* Note: default: begin==0.0 */
	u_char headfa[4]; /* head fade-end absolute time in TU (if enabled) */
	u_char tailfa[4]; /* tail fade-start absolute time in TU (if enabled) */
	u_char xxx5[2];
	u_char nr[2];     /* incremental clip number in track */
	u_char xxx1[24];
#define MT_CLIP_WFLAGS2      0xf000 /* wave2 flags */
#define MT_CLIP_FLAG2_XXX1   0x1000 /* ??? flag */
#define MT_CLIP_FLAG2_XXX2   0x2000 /* ??? flag */
#define MT_CLIP_FLAG2_TAILFA 0x4000 /* tail fade enable */
#define MT_CLIP_FLAG2_HEADFA 0x8000 /* head fade enable */
#define MT_CLIP_WAVE2        0x0fff /* wave2 number mask */
	u_char wave2[2];  /* wave side 2 if Stereo and more flags */
	u_char xxx7[6];
	u_char xxx8[2];
	u_char levelvca[2];  /* level VCA-value (0x0000..0x0fff) */
	u_char leveldb[2];   /* level in dB */
	u_char xxx10[18];
	u_char xxx11;
	char name[25];    /* name (\0-terminated) */
};



/* wave */
/* XXX should be 0x200 Bytes */
struct mt_wave_s{
#define MT_WAVE_FLAG_A    0x80 /* Active */
#define MT_WAVE_FLAG_O    0x40
#define MT_WAVE_FLAG_D    0x20
#define MT_WAVE_FLAG_E    0x10
#define MT_WAVE_FLAG_XXX  0x08
#define MT_WAVE_FLAG_L    0x04
#define MT_WAVE_FLAG_F    0x02
#define MT_WAVE_FLAG_C    0x01
	u_char flags;       /* flags */
	u_char file;        /* library index if !=0 */
	u_char nr[2];       /* typically wave number (but can be different in older Rev.) ??? */
	u_char used[2];     /* used */
	u_char size[4];     /* size in Bytes */
	u_char birthday[4]; /* birthday time in TU */
	u_char begin[4];    /* start-offset in waveform-data */
	u_char xxx4[4];     /* offset ??? or -1 ??? */
	u_char xxx5[4];     /* size == totsize + xxx4 or 0 ??? */
	u_char xxx6[4];     /* size ??? */
	u_char xxx7[2];
	u_char zero3[2];
#define MT_WAVE_SEG_NR 159  /* max. number of segments */
	struct{
#define MT_WAVE_SEG_ANCIENT_MAXCNT 0x80 /* max. count ??? */
		u_char cnt;    /* block-count of segment */
		u_char off[2]; /* block-offset to start of segment */
	}seg[159];          /* waveform-data segment-vector */
	u_char zero4;
};
/* waveform-data block size: nbytes */
/* base for waveform-data offsets: sblock */



/* mark Rev. 9 */
/* XXX should be 0x22 Bytes */
struct mt_mark9_s{
	u_char xxx1[4];
	u_char xxx2[4];
	u_char time[4];  /* mark time in TU */
#define MT_MARK9_FREE 0x80 /* free-flag */
#define MT_MARK9_NR   0x7f /* number */
#define MT_MARK9_XXX  0xff /* first mark??? */
	u_char nr;       /* free-flag and number */
	char name[21];   /* name (\0-terminated) */
};
/* see MT_HEAD_ANCIENT_MARKPOS and MT_HEAD_ANCIENT_MARKSIZE !!! */



/* mark Rev. 11 */
/* XXX should be 0x24 Bytes */
struct mt_mark11_s{
	u_char xxx1[4];
	u_char xxx2[4];
	u_char time[4]; /* mark time in TU */
#define MT_MARK11_FREE 0x8000 /* free-flag */
#define MT_MARK11_NR   0x7fff /* number */
#define MT_MARK11_XXX  0xffff /* first mark??? */
	u_char nr[2];   /* free-flag and number */
	u_char zero1;
	char name[21];  /* name (\0-terminated) */
};



extern void mdrmt_set_passwd(char *buf,u_char *pp);
extern void mdrmt_get_passwd(char *buf,u_char *pp);
extern void mdrmt_print_passwd(char *sp,u_char *pp);

extern int printinfo_mt(struct genfile_s *fp,int verbose);

extern int mt2wav(struct genfile_s *fp,char *wname,
				  u_int wnr1,u_int wnr2,
				  u_int start,u_int size,
				  char *buf,u_int bufsiz);



#endif /* !__CMIOS9_MDR_H */
