/*
* Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2011  K. M. Indlekofer
*
* m.indlekofer@gmx.de
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



#include "cmios9_err.h"
#include "cmios9.h"
#include "cmios9_sqtk.h"



/* number of data bytes: indexed by high nibble of channel voice message */
int midi_chanvclen[16]={-1,-1,-1,-1,-1,-1,-1,-1, 2, 2, 2, 2, 1, 1, 2,-1};
/*                                               8  9  a  b  c  d  e     */



#define SEQ_MAXNUM 16 /* XXX max. number of input files */

static u_int sq2tk_seq_num;

static struct smf_head_s sq2tk_smfheader;
static u_char sq2tk_header[TK_HEADERSIZE]; /* same size for all! */

static struct tframe_s sq2tk_tfbuf_noop={0x00,0x00,FRAME_TYPE_NOOP,0x00,0x00,0x00,0x00,0x00};

static struct tframe_s sq2tk_tfbuf_in[SEQ_MAXNUM];
static struct tframe_s sq2tk_tfbuf_out;

static struct genfile_s sq2tk_inf[SEQ_MAXNUM];
static u_int sq2tk_infoff[SEQ_MAXNUM];
static u_int sq2tk_inrev[SEQ_MAXNUM]; /* input file type */

static int sq2tk_smfoutfdes;
static u_int sq2tk_smfsize; /* byte counter for fsize in SMF header */
static struct genfile_s sq2tk_outf;
static u_int sq2tk_outfoff;

static long sq2tk_tdiff[SEQ_MAXNUM]; /* in CUE-LIST TK microbeats (==TU) */
static u_long sq2tk_tdiff1[SEQ_MAXNUM]; /* in CUE-LIST TK microbeats/tempo */
static u_long sq2tk_deftempo[SEQ_MAXNUM]; /* default tempo */
static u_long sq2tk_tempo[SEQ_MAXNUM]; /* tempo */
static int sq2tk_pending[SEQ_MAXNUM]; /* input frame pending */
static int sq2tk_stopped[SEQ_MAXNUM]; /* input end */

static u_long sq2tk_ttot[SEQ_MAXNUM];  /* in CUE-LIST TK microbeats (==TU) */
static u_long sq2tk_ttotout;  /* in CUE-LIST TK microbeats (==TU) */



static void
sq2tk_timeacc(u_long dt,u_int seq)
{
	volatile u_long t,t1; /* must not optimize division below! */

	if (sq2tk_stopped[seq])
		return;

	/* dt in input microbeats! */

	if (sq2tk_inrev[seq]==TK_REV){
		/* Note: input microbeats per second==tempo */
		/* XXX no check for overflow! */
		dt*=TK_UBEAT_SEC;
	}else if ((sq2tk_inrev[seq]==SQ_REV)||(sq2tk_inrev[seq]==RSTK_REV)){
		/* Note: input microbeats per second==tempo/2.5 */
		/* XXX no check for overflow! */
		dt*=TK_UBEAT_SEC+TK_UBEAT_SEC+(TK_UBEAT_SEC/2);
	}else{
		/* XXX nothing */
		return;
	}

	/* convert dt to CUE-LIST TK-microbeats */
	/* division must come at last! */
	t=dt/sq2tk_tempo[seq];
	t1=dt%sq2tk_tempo[seq];
	/* the problem is that tempo might change and the meaning of t1 as well! */
	/* -> tdiff1 must be rescaled in gather() with every tempo change! */

	/* account for sub-microbeat errors */
#ifndef NOSUBUB
	sq2tk_tdiff1[seq]+=t1; /* sub-TK-microbeat accumulator */
#endif
	t+=sq2tk_tdiff1[seq]/sq2tk_tempo[seq]; /* correct integer TK-microbeats */
	sq2tk_tdiff1[seq]%=sq2tk_tempo[seq]; /* new remainder */

	/* integer TK-microbeat accumulator */
	/* XXX no check for tdiff overflow */
	sq2tk_tdiff[seq]+=t; /* integer TK-microbeat due-time accumulator */
	sq2tk_ttot[seq]+=t; /* total time */
}

static void
sq2tk_gather(u_int seq)
{
	volatile u_long oldtempo; /* must not optimize division below! */
	int ret;

	while ((!sq2tk_stopped[seq])&&(!sq2tk_pending[seq])){
		/* read input */
		ret=readfile4(&sq2tk_inf[seq],(char *)&sq2tk_tfbuf_in[seq],sq2tk_infoff[seq],sizeof(sq2tk_tfbuf_in[0]));
		sq2tk_infoff[seq]+=sizeof(sq2tk_tfbuf_in[0]);
		if (ret<0){
			/* end */
			sq2tk_stopped[seq]=1;
			return;
		}

		/* accumulate time */
		sq2tk_timeacc((((u_long)(sq2tk_tfbuf_in[seq].dtime[0]))<<8)
			+((u_long)(sq2tk_tfbuf_in[seq].dtime[1])),seq); /* in input microbeats! */

		/* CMI frame type switch */
		switch (sq2tk_tfbuf_in[seq].type){
		case FRAME_TYPE_NOOP:
			/* nothing to do */
			break;
		case FRAME_TYPE_MIDI:
			sq2tk_pending[seq]=1; /* new MIDI frame to issue */
			break;
		case FRAME_TYPE_TEMPO:
			/* get new tempo */
			oldtempo=sq2tk_tempo[seq];
			sq2tk_tempo[seq]=(((u_long)((struct frame_tempo_s *)(sq2tk_tfbuf_in[seq].frame))->tempo[0])<<8)
				+((u_long)((struct frame_tempo_s *)(sq2tk_tfbuf_in[seq].frame))->tempo[1]);
			if (sq2tk_tempo[seq]==0){
				sq2tk_tempo[seq]=sq2tk_deftempo[seq];
			}

			/* must rescale tdiff1 for timeacc() if tempo change! */
			if (oldtempo!=sq2tk_tempo[seq]){
				/* division must come at last! */
				/* XXX no overflow check */
				sq2tk_tdiff1[seq]*=sq2tk_tempo[seq];
				sq2tk_tdiff1[seq]/=oldtempo;
				/* this is not 100% exact in integer arithmetic, but good enough... */

				/* account for sub-microbeat errors */
				/* XXX actually not necessary here since timeacc will be called next */
				sq2tk_tdiff[seq]+=sq2tk_tdiff1[seq]/sq2tk_tempo[seq]; /* correct integer TK-microbeats */
				sq2tk_tdiff1[seq]%=sq2tk_tempo[seq]; /* new remainder */
			}
			break;
		default:
#ifdef PASS_ALL
			sq2tk_pending[seq]=1; /* new input frame to issue */
#endif
			break;
		}
	}

	if ((!sq2tk_stopped[seq])&&(!sq2tk_pending[seq])){
		fprintf(stderr,"internal error in gather.\n");
	}
}

static int
sq2tk_issue(u_long dt,u_int seq)
{
	int ret;

	/* dt in CUE-LIST TK microbeats! */

	if (sq2tk_stopped[seq])
		return 0;

	if (!sq2tk_pending[seq]){
		fprintf(stderr,"warning: input[%u] due but no output\n",seq);
		return 0;
	}

	if (sq2tk_smfoutfdes>=0){
		/* SMF-file output */

		u_char *tfp;
		u_char buf[32]; /* XXX enough for timestamp and max. MIDI packet */
		int count,len,i;

		count=0;

		/* encode timestamp */
		len=smf_encode32(dt,buf+count);
		count+=len;

		/* encode MIDI channel voice message */
		tfp=&sq2tk_tfbuf_in[seq].frame[0]; /* frame data */
		len=midi_chanvclen[0x0f&(tfp[1]>>4)]; /* number of MIDI data bytes */
		if ((len<0)||(len>2)){
			/* not channel voice message */
			fprintf(stderr,"invalid MIDI message (%02x), ignored.\n",tfp[1]);
			len=0;
		}else{
#if 1
			/* set CMI instrument as MIDI channel */
			tfp[1]=(0xf0&tfp[1])|(0x0f&(tfp[0]-1));
#endif
			len++; /* for status byte */
		}
		/* copy status byte and data bytes */
		for (i=0;i<len;i++){
			buf[count++]=tfp[1+i];
		}

		/* issue frame */
		if (WRITE(sq2tk_smfoutfdes,buf,count)!=count){
			printf("cannot write\n");
			return -1;
		}
		sq2tk_smfsize+=count; /* byte counter */
	}else{
		/* CUE-LIST TK-file output */

		/* burn time if necessary */
		while (dt>TK_DTIME_MAX){
			/* issue noop frame */
			sq2tk_tfbuf_noop.dtime[0]=(u_char)(TK_DTIME_MAX>>8);
			sq2tk_tfbuf_noop.dtime[1]=(u_char)(TK_DTIME_MAX);
			/* extend and write */
			if (extendfile4(&sq2tk_outf,sizeof(sq2tk_tfbuf_noop))<0){
				fprintf(stderr,"cannot extend\n");
				return -1;
			}
			ret=writefile4(&sq2tk_outf,(char *)&sq2tk_tfbuf_noop,sq2tk_outfoff,sizeof(sq2tk_tfbuf_noop));
			sq2tk_outfoff+=sizeof(sq2tk_tfbuf_noop);
			if (ret<0){
				fprintf(stderr,"cannot write(noop)\n");
				return -1;
			}

			/* TK_DTIME_MAX done */
			dt-=TK_DTIME_MAX;
		}

		/* issue frame */
		bcopy(&sq2tk_tfbuf_in[seq],&sq2tk_tfbuf_out,sizeof(sq2tk_tfbuf_out));
		sq2tk_tfbuf_out.dtime[0]=(u_char)(dt>>8);
		sq2tk_tfbuf_out.dtime[1]=(u_char)(dt);
		/* extend and write */
		if (extendfile4(&sq2tk_outf,sizeof(sq2tk_tfbuf_out))<0){
			fprintf(stderr,"cannot extend\n");
			return -1;
		}
		ret=writefile4(&sq2tk_outf,(char *)&sq2tk_tfbuf_out,sq2tk_outfoff,sizeof(sq2tk_tfbuf_out));
		sq2tk_outfoff+=sizeof(sq2tk_tfbuf_out);
		if (ret<0){
			fprintf(stderr,"cannot write\n");
			return -1;
		}
	}

	sq2tk_pending[seq]=0; /* must gather next input frame */
	return 0;
}

static void
sq2tk_propagate(u_long dt,u_int seq)
{

	/* dt in CUE-LIST TK microbeats! */

	if (sq2tk_stopped[seq])
		return;

	/* propagate in time */
	sq2tk_tdiff[seq]-=dt;

	/* Note: even if dt was too large so that tdiff is negative now, */
	/*       this time won't be lost, because next events in gather() will */
	/*       occur earlier exacly by this amount, because tdiff accumulates! */
}

static int
sq2tk_findseq(u_long *dtp)
{
	int seq,seq0;
	u_long dt0;

	/* find first seq which is/will be due */
	dt0=0xffffffffUL; /* XXX should be enough*/
	seq0=-1; /* none yet */
	for (seq=0;seq<(int)sq2tk_seq_num;seq++){
		if (!sq2tk_stopped[seq]){
			if (sq2tk_tdiff[seq]<0){ /* one seq over-due? */
				seq0=seq;
				dt0=0; /* must stay where we are! */
				break; /* must do it now */
			}else if (((u_long)sq2tk_tdiff[seq])<dt0){
				seq0=seq;
				dt0=sq2tk_tdiff[seq];
			}
		}
	}

	if (dtp!=0){
		*dtp=dt0;
	}

	return seq0;
}

static int
sq2tk_runseq(void)
{
	int seq,seq0;
	u_long dt0;

	/* firstly, fill input buffer for all seq */
	for (seq=0;seq<(int)sq2tk_seq_num;seq++){
		if (!sq2tk_stopped[seq]){
			sq2tk_gather(seq);
		}
	}

	/* propagte through input files in parallel... */
	while (1){
		/* find next seq which is/will be due */
		seq0=sq2tk_findseq(&dt0);
		if (seq0<0){ /* none found? */
			return 0; /* end */
		}

		/* issue seq0 */
		if (sq2tk_issue(dt0,seq0)<0){
			return -1;
		}

		/* propagate all seq by dt0 */
		for (seq=0;seq<(int)sq2tk_seq_num;seq++){
			if (!sq2tk_stopped[seq]){
				sq2tk_propagate(dt0,seq);
			}
		}
		sq2tk_ttotout+=dt0; /* output total time */

		/* re-load seq0 if necessary */
		sq2tk_gather(seq0);
	}
	/* XXX never reached */
	return -1;
}

int
sq2tk(int argc,char **argv,int smfoutflag)
{
	u_int seq;
	int ret;
	int l;
	int sqflag;

	if ((argc<2)||((argc-2)>SEQ_MAXNUM)){
		return -1;
	}
	sq2tk_seq_num=argc-2;

	/* read input file header */
	for (seq=0;seq<sq2tk_seq_num;seq++){
		fprintf(stderr,"input[%u] `%s':\t\t ",seq,argv[seq+1]);

		/* open input file */
		if (readfiledes4(argv[seq+1],&sq2tk_inf[seq],0)<0){ /* 0: do not touch */
			fprintf(stderr,"cannot open %s\n",argv[seq+1]);
			return -1;
		}
		sq2tk_infoff[seq]=0;

		/* read header */
		ret=readfile4(&sq2tk_inf[seq],(char *)sq2tk_header,sq2tk_infoff[seq],sizeof(sq2tk_header));
		sq2tk_infoff[seq]+=sizeof(sq2tk_header);
		if (ret<0){
			fprintf(stderr,"cannot read %s\n",argv[seq+1]);
			return -1;
		}

		/* check file name suffix */
		l=strlen(argv[seq+1]);
		if ((l>=3)&&(strncasecmp(".SQ",argv[seq+1]+l-3,3)==0)){ /* case-insensitive */
			sqflag=1;
		}else{
			sqflag=0;
		}
		/* check input file header */
		sq2tk_inrev[seq]=(((u_long)(sq2tk_header[0]))<<8)+((u_long)(sq2tk_header[1]));
		if ((sqflag==0)&&(sq2tk_inrev[seq]==TK_REV)){
			fprintf(stderr,"CUE-LIST TK-file.\n");
			sq2tk_deftempo[seq]=(((u_long)(sq2tk_header[2]))<<8)+((u_long)(sq2tk_header[3]));
			if (sq2tk_deftempo[seq]==0){
				sq2tk_deftempo[seq]=TK_DEFTEMPO;
			}
			sq2tk_stopped[seq]=0; /* may run */
		}else if ((sqflag==0)&&(sq2tk_inrev[seq]==RSTK_REV)){
			fprintf(stderr,"RS TK-file.\n");
			sq2tk_deftempo[seq]=RSTK_DEFTEMPO; /* default tempo */
			sq2tk_stopped[seq]=0; /* may run */
		}else if ((sqflag==1)&&(sq2tk_inrev[seq]==SQ_REV)){
			fprintf(stderr,"MCL SQ-file.\n");
			sq2tk_deftempo[seq]=SQ_DEFTEMPO; /* default tempo */
			sq2tk_stopped[seq]=0; /* may run */
		}else{
			fprintf(stderr,"unsupported file type, skipped.\n");
			sq2tk_stopped[seq]=1; /* must not run */
		}
	}

	if (smfoutflag){
		/* SMF-file output */

		fprintf(stderr,"\nconverting/merging to SMF (MIDI) file...\n\n");

		/* open output file */
		/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
		if ((sq2tk_smfoutfdes=OPEN(argv[seq+1],O_RDWR|O_CREAT|O_TRUNC|O_BINARY,0666))<0){
			fprintf(stderr,"cannot open destination file\n");
			return -1;
		}

		/* write SMF dummy header since we don't know final sq2tk_smfsize yet */
		bzero(&sq2tk_smfheader,sizeof(struct smf_head_s));
		if (WRITE(sq2tk_smfoutfdes,&sq2tk_smfheader,sizeof(struct smf_head_s))!=sizeof(struct smf_head_s)){
			printf("cannot write SMF header\n");
			CLOSE(sq2tk_smfoutfdes);
			return -1;
		}
		sq2tk_smfsize=0; /* byte counter */
	}else{
		/* CUE-LIST TK-file output */

		fprintf(stderr,"\nconverting/merging to CUE-LIST TK file...\n\n");

		/* open output file */
		/* check if destination file already exists */
		/* see findfile3: return value 0 is virtual directory -> count as non-existing here!!! */
		ret=findfile3(argv[seq+1],NULL,NULL);
		err_reset(); /* clear findfile error! */
		if (ret>0){
			/* exists */
			/* delete to overwrite old file */
			if (delfile3(argv[seq+1],0)<0){
				fprintf(stderr,"cannot overwrite destination file\n");
				return -1;
			}
		}
		/* create output file */
		if (creatfile3(argv[seq+1],0,0,NULL)<0){ /* 0: non-dir, size 0 so far */
			fprintf(stderr,"cannot create destination file\n");
			return -1;
		}
		/* open output file */
		if (readfiledes4(argv[seq+1],&sq2tk_outf,0)<0){ /* 0: do not touch */
			fprintf(stderr,"cannot access destination file\n");
			return -1;
		}
		sq2tk_outfoff=0;

		/* set TK-file header */
		bzero(sq2tk_header,sizeof(sq2tk_header));
		sq2tk_header[0]=(u_char)(TK_REV>>8);
		sq2tk_header[1]=(u_char)(TK_REV);
		sq2tk_header[2]=(u_char)(TK_UBEAT_SEC>>8);
		sq2tk_header[3]=(u_char)(TK_UBEAT_SEC);
		/* extend and write TK-file header */
		if (extendfile4(&sq2tk_outf,sizeof(sq2tk_header))<0){
			fprintf(stderr,"cannot extend\n");
			return -1;
		}
		ret=writefile4(&sq2tk_outf,(char *)sq2tk_header,sq2tk_outfoff,sizeof(sq2tk_header));
		sq2tk_outfoff+=sizeof(sq2tk_header);
		if (ret<0){
			fprintf(stderr,"cannot write(header)\n");
			return -1;
		}

		sq2tk_smfoutfdes=-1; /* not SMF */
	}

	/* initialize */
	for (seq=0;seq<sq2tk_seq_num;seq++){
		sq2tk_tempo[seq]=sq2tk_deftempo[seq];
		sq2tk_tdiff[seq]=0; /* reset time accumulator */
		sq2tk_tdiff1[seq]=0; /* reset sub-TK-microbeat accumulator */
		sq2tk_ttot[seq]=0; /* reset total time */
		sq2tk_pending[seq]=0; /* nothing yet */
	}
	sq2tk_ttotout=0; /* nothing yet */

	/* propagate through input files in parallel... */
	if (sq2tk_runseq()<0){
		if (sq2tk_smfoutfdes>=0){
			CLOSE(sq2tk_smfoutfdes);
		}
		return -1;
	}

	for (seq=0;seq<sq2tk_seq_num;seq++){
		fprintf(stderr,"input[%u]:\t\t %02luh %02lum %02lus %03.1lfms\n",
			seq,
			sq2tk_ttot[seq]/(3600*TK_UBEAT_SEC),
			(sq2tk_ttot[seq]/(60*TK_UBEAT_SEC))%60,
			(sq2tk_ttot[seq]/TK_UBEAT_SEC)%60,
			1000.0*((double)(sq2tk_ttot[seq]%TK_UBEAT_SEC))/((double)TK_UBEAT_SEC));
	}
	fprintf(stderr,"output:   \t\t %02luh %02lum %02lus %03.1lfms\n",
		sq2tk_ttotout/(3600*TK_UBEAT_SEC),
		(sq2tk_ttotout/(60*TK_UBEAT_SEC))%60,
		(sq2tk_ttotout/TK_UBEAT_SEC)%60,
		1000.0*((double)(sq2tk_ttotout%TK_UBEAT_SEC))/((double)TK_UBEAT_SEC));
	/* Note: not necessarily max. of ttot since TFRAMEs at the end */
	/*       might not be accepted for output! */

	if (smfoutflag){
#if 1
		/* issue SMF end of track */
		{
			u_char endoftrk[]={0x00,SMF_META,SMF_META_ENDOFTRK,0x00}; /* delta time=0 */

			if (WRITE(sq2tk_smfoutfdes,endoftrk,4)!=4){
				printf("cannot write SMF-header\n");
				CLOSE(sq2tk_smfoutfdes);
				return -1;
			}
			sq2tk_smfsize+=4;
		}
#endif
		/* now that we know sq2tk_smfsize, we can set and write the SMF-header */
		/* set SMF-header */
		smf_head(&sq2tk_smfheader,SMF_HEAD_FORMAT1,SMF_HEAD_NUMTRKS1,SMF_HEAD_DIVISIONTK,sq2tk_smfsize);
		/* goto begin of file */
		if (LSEEK(sq2tk_smfoutfdes,0,SEEK_SET)!=0){
			printf("cannot goto SMF-header\n");
			CLOSE(sq2tk_smfoutfdes);
			return -1;
		}
		/* write header */
		if (WRITE(sq2tk_smfoutfdes,&sq2tk_smfheader,sizeof(struct smf_head_s))!=sizeof(struct smf_head_s)){
			printf("cannot write SMF-header\n");
			CLOSE(sq2tk_smfoutfdes);
			return -1;
		}
	}

	if (sq2tk_smfoutfdes>=0){
		CLOSE(sq2tk_smfoutfdes);
	}
	return 0;
}



int
printinfo_sqtk(struct genfile_s *fp,char *id,int verbose)
{
	u_char header[TK_HEADERSIZE]; /* same size for all! */
	u_int inrev;
	int sqflag;
	u_int deftempo;

	u_int foff;

	if (fp==NULL){
		return -1;
	}
	if (id==NULL){
		return -1;
	}

	/* read header */
	foff=0; /* begin of file */
	if (readfile4(fp,(char *)header,foff,sizeof(header))<0){
		printf("cannot read file\n");
		return -1;
	}
	foff+=sizeof(header);

	if (strcasecmp(id,".SQ")==0){
		sqflag=1;
	}else{
		sqflag=0;
	}

	/* check input file header */
	inrev=(((u_long)(header[0]))<<8)+((u_long)(header[1]));
	if ((sqflag==0)&&(inrev==TK_REV)){
		printf("CMI Series III CUE-LIST take file\n");
		deftempo=(((u_long)(header[2]))<<8)+((u_long)(header[3]));
		if (deftempo==0){
			deftempo=TK_DEFTEMPO;
		}
		printf("deftempo: %u\n",deftempo);
	}else if ((sqflag==0)&&(inrev==RSTK_REV)){
		printf("CMI Series III RS capsgen TK-file\n");
		deftempo=RSTK_DEFTEMPO;
		printf("deftempo: %u\n",deftempo);
	}else if ((sqflag==1)&&(inrev==SQ_REV)){
		printf("CMI Series III MCL SQ-file\n");
		deftempo=SQ_DEFTEMPO;
		printf("deftempo: %u\n",deftempo);
	}else if (sqflag==1){
		/* XXX */
		printf("Series I/II/IIx keyboard sequence file (page 9)\n");
	}else{
		printf("unsupported file type\n");
	}

	return 0;
}



void
smf_head(struct smf_head_s *smfheadp,u_int format,u_int numtrks,u_int division,u_int fsize)
{

	if (smfheadp==NULL){
		return;
	}

	/* create SMF-header */
	bzero(smfheadp,sizeof(struct smf_head_s));
	bcopy(SMF_HEAD_MTHDSTR,smfheadp->mthdstr,strlen(SMF_HEAD_MTHDSTR));
	{
		u_int csize;

		csize=SMF_HEAD_CSIZE;
		smfheadp->csize[3]=0xff&csize;
		smfheadp->csize[2]=0xff&(csize>>8);
		smfheadp->csize[1]=0xff&(csize>>16);
		smfheadp->csize[0]=0xff&(csize>>24);
	}
	smfheadp->format[1]=0xff&format;
	smfheadp->format[0]=0xff&(format>>8);
	smfheadp->numtrks[1]=0xff&numtrks;
	smfheadp->numtrks[0]=0xff&(numtrks>>8);
	smfheadp->division[1]=0xff&division;
	smfheadp->division[0]=0xff&(division>>8);
	bcopy(SMF_HEAD_MTRKSTR,smfheadp->mtrkstr,strlen(SMF_HEAD_MTRKSTR));
	smfheadp->fsize[3]=0xff&fsize;
	smfheadp->fsize[2]=0xff&(fsize>>8);
	smfheadp->fsize[1]=0xff&(fsize>>16);
	smfheadp->fsize[0]=0xff&(fsize>>24);
}



/* encodes 32bit integer, returns number of bytes */
int
smf_encode32(u_int dt,u_char *out)
{
	u_char buf[5];
	int l,i;

	if (out==NULL){
		return 0;
	}

	l=1; /* even if dt==0 */

	buf[0]=0x7f&dt;
	if (buf[0]!=0){
		l=1;
	}
	/* only LSB has Bit7=0 */

	dt>>=7;
	buf[1]=0x7f&dt;
	if (buf[1]!=0){
		l=2;
	}
	buf[1]|=0x80;

	dt>>=7;
	buf[2]=0x7f&dt;
	if (buf[2]!=0){
		l=3;
	}
	buf[2]|=0x80;

	dt>>=7;
	buf[3]=0x7f&dt;
	if (buf[3]!=0){
		l=4;
	}
	buf[3]|=0x80;

	dt>>=7;
	buf[4]=0x7f&dt;
	if (buf[4]!=0){
		l=5;
	}
	buf[4]|=0x80;

	for (i=0;i<l;i++){
		out[l-i-1]=buf[i];
	}

	return l;
}



/* EOF */
