#ifndef __CMIOS9_SQTK_H
#define __CMIOS9_SQTK_H
/*
* Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2011  K. M. Indlekofer
*
* m.indlekofer@gmx.de
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



/* CMI Series III timed-frame files: .SQ, .TK, CUE-LIST .TK */ 



/* CUE-LIST TK-file specific */
#define TK_HEADERSIZE 0x0100
#define TK_REV        0x0001
#define TK_UBEAT_SEC  0x0000bb80UL /* microbeats per second (same as TU) */
/* Note: TK microbeats per second==tempo */
#define TK_DEFTEMPO   TK_UBEAT_SEC
#define TK_DTIME_MAX (TK_UBEAT_SEC/4) /* max. dtime in tframe, is integer!!! */



/* MCL SQ-file specific */
#define SQ_HEADERSIZE    0x0100
#define SQ_REV           0xff01
#define SQ_UBEAT_QUARTER 0x000001e0UL /* microbeats per quarter */
#define SQ_UBEAT_16MCLK  0x00000140UL /* microbeats per 16 MIDI clocks */
/* Note: SQ microbeats per second==tempo/2.5 */
#define SQ_DEFTEMPO      0x00000960UL /* 120bpm */



/* RS TK-file ("capsgen") specific */
#define RSTK_HEADERSIZE    0x0100
#define RSTK_REV           0x0000
#define RSTK_UBEAT_QUARTER 0x000000c0UL /* microbeats per quarter */
#define RSTK_UBEAT_16MCLK  0x000000a0UL /* microbeats per 16 MIDI clocks */
/* Note: RSTK microbeats per second==tempo/2.5 */
#define RSTK_DEFTEMPO      0x000004b0UL /* 120bpm */



/* timed frame */
struct tframe_s{
  u_char dtime[2];    /* delay time in microbeats until this frame */
  u_char type;        /* frame type (FRAME_TYPE_...) */
  u_char typemod;     /* frame type modifier */
  u_char frame[4];    /* frame data (struct frame_..._s) */
};



/* NOOP frame (e.g. to burn time) */
/* frame data all 0! */

#define FRAME_TYPE_NOOP 0



/* MIDI frame */
struct frame_midi_s{
  u_char inst;       /* CMI instrument number */
  u_char mevent[3];  /* MIDI event */
};

#define FRAME_TYPE_MIDI 2



/* tempo frame */
struct frame_tempo_s{
  u_char dummy[2];
  u_char tempo[2]; /* tempo */
};

#define FRAME_TYPE_TEMPO 5



/* MIDI channel voice messages, lower nibble is channel-1 */
#define MIDI_VOICE_NOTEOFF		0x80
#define MIDI_VOICE_NOTEON		0x90
#define MIDI_VOICE_POLYKPRESS	0xa0
#define MIDI_VOICE_CONTRLCHG	0xb0
#define MIDI_VOICE_PROGCHG		0xc0
#define MIDI_VOICE_CHANPRESS	0xd0
#define MIDI_VOICE_PITCHBEND	0xe0
/* number of data bytes: indexed by high nibble of channel voice message */
extern int midi_chanvclen[16];
/* Note: data byte is <= 0x7f */

/* MIDI realtime messages: 1 byte, >0xf7  */
#define MIDI_RT_CLOCK			0xf8
#define MIDI_RT_START			0xfa
#define MIDI_RT_CONTINUE		0xfb
#define MIDI_RT_STOP			0xfc
#define MIDI_RT_ACTSENS			0xfe
#define MIDI_RT_SYSRESET		0xff

/* MIDI system common messages */
#define MIDI_SYSCOMMON_SONGPOS	0xf2
#define MIDI_SYSCOMMON_SONGSEL	0xf3
#define MIDI_SYSCOMMON_TUNEREQ	0xf6

/* MIDI system exclusive message (variable length, EOX) */
#define MIDI_SYSEX				0xf0
#define MIDI_SYSEX_EOX			0xf7



/* SFM (Standard MIDI File) format header */
/* Note: big-endian!!! */
struct smf_head_s{
/* begin MThd chunk */
#define SMF_HEAD_MTHDSTR "MThd"
	char mthdstr[4]; /* must be "MThd" */
#define SMF_HEAD_CSIZE 6
	u_char csize[4];   /* following chunk size in bytes, must be 6 */
#define SMF_HEAD_FORMAT1 1
	u_char format[2]; /* format */
#define SMF_HEAD_NUMTRKS1 1
	u_char numtrks[2]; /* number of tracks */
#define SMF_DEFTEMPO 120 /* default tempo == 120bpm */
#define SMF_HEAD_DIVISIONTK (TK_UBEAT_SEC/(SMF_DEFTEMPO/60)) /* compatible with CUE-LIST TK, is integer!!! */
	u_char division[2]; /* number of microbeats per quarter (PPQN) */
/* end MThd chunk */
/* begin MTrk chunk */
#define SMF_HEAD_MTRKSTR "MTrk"
	char mtrkstr[4]; /* must be "MTrk" */
#define SMF_HEAD_FSIZE 6
	u_char fsize[4];   /* following file size in bytes */
};

/* non-MIDI events (meta events) */
#define SMF_META				0xff
/* meta event types */
#define SMF_META_SEQNR			0x00
#define SMF_META_TEXT			0x01
#define SMF_META_COPYRIGHT		0x02
#define SMF_META_SEQTRKNAME		0x03
#define SMF_META_INSTR			0x04
#define SMF_META_LYRICS			0x05
#define SMF_META_MARKER			0x06
#define SMF_META_CUEPT			0x07
#define SMF_META_PROGNAME		0x08
#define SMF_META_DEVNAME		0x09
#define SMF_META_CHPREFIX		0x20
#define SMF_META_ENDOFTRK		0x2f
#define SMF_META_TEMPO			0x51
#define SMF_META_SMPTEOFF		0x54
#define SMF_META_TIMESIGN		0x58
#define SMF_META_KEYSIGN		0x59
#define SMF_META_PROPR			0x7f



extern int sq2tk(int argc,char **argv,int smfoutflag);
extern int printinfo_sqtk(struct genfile_s *fp,char *id,int verbose);
extern void smf_head(struct smf_head_s *smfheadp,u_int format,u_int numtrks,u_int division,u_int fsize);
extern int smf_encode32(u_int dt,u_char *out);



#endif /* !__CMIOS9_SQTK_H */
