/*
* Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2010, 2011  K. M. Indlekofer
*
* m.indlekofer@gmx.de
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* XXX make buffers and large structs static (if non-recursive) to prevent from stack overflow!?! */



#include "cmios9_err.h"
#include "cmios9.h"
#include "cmios9_tree.h"
#include "cmios9_voice.h"
#include "cmios9_mdr.h"



/* OS9 */
int
listdirtree(int (*recfct)(char *name,int rec,int arg),int arg)
{
	/* recursive: variables must not be static!!! */
	struct filedes_s curdirdes;
	u_int curdirsize;
	struct dirent_s *curdirp;
	u_int blknr;
	u_int i,j;
	char namebuf[FILENAMELEN+1];
	u_int fdessec;
	struct filedes_s fdes;
	int ret;
	u_int fsize,asize,corrupt;
	int dir;
	
	static u_int reclevel=0; /* static recursion level */
	static u_int maxreclevel=0; /* static recursion level */
	static u_int totfiles; /* static for recursion! */
	static u_int totdirs; /* static for recursion! */
	static u_int totsize; /* static for recursion! */
	static u_int totalloc; /* static for recursion! */
	static u_int corruptcount; /* static for recursion! */
	
	if (curpart->type!=PARTTYPE_OS9MDR){
		return -1;
	}
	if (curpart->mdr){
		return -1;
	}
	
	/* must be recursive */
	if (recfct==NULL){
		return -1;
	}
	
	if (reclevel==0){
		/* reset */
		totfiles=0;
		totdirs=0;
		totsize=0;
		totalloc=0;
		corruptcount=0;
	}
	
	/* XXX read whole directory file */

	if (getdir(curdir.dirsec,&curdirdes,&curdirsize)<0){
		return -1;
	}
	if (curdirsize==0){
#if 0
		err_print("listdirtree: directory empty\n");
		return -1;
#else
		corruptcount++;
		printf("*** empty directory ***\n");
		return 0; /* XXX non-fatal */
#endif
	}
	
	/* read directory file */
	/* XXX 32-Bit overflow protection */
	blknr=(u_int)((((OFF_T)curdirsize)+((OFF_T)curpart->blksiz-1))
		/((OFF_T)curpart->blksiz));
	
	/* allocate memory for directory file */
	curdirp=(struct dirent_s *)malloc(blknr*curpart->blksiz);
	if (curdirp==NULL){
		perror("listdirtree: malloc");
		return -1;
	}
	
	/* read directory file */
	if (fileread((union genfdes_u *)&curdirdes,(void *)curdirp,0,blknr)!=(int)blknr){
#if 0
		err_print("listdirtree: fileread(curdirdes)\n");
		free((void *)curdirp);
		return -1;
#else
		corruptcount++;
		printf("*** cannot read directory ***\n");
		free((void *)curdirp);
		return 0; /* XXX non-fatal */
#endif
	}
	
	/* scan directory entries */
	for (i=0;i<curdirsize/sizeof(struct dirent_s);i++){
		if (curdirp[i].dir_name[0]==0){
			continue; /* skip entry */
		}
		translatename(curdirp[i].dir_name,namebuf);
		/* skip "." and ".." */
		if ((strcmp(namebuf,".")==0)||(strcmp(namebuf,"..")==0)){
			continue; /* skip entry */
		}
		
		/* get file-descriptor */
		fdessec=(curdirp[i].dir_addr[0]<<16)
				+(curdirp[i].dir_addr[1]<<8)
				+curdirp[i].dir_addr[2];
		if (READBLOCK_CURPART(fdessec,&fdes,1)<0){
#if 0
			err_print("listdirtree: read file descriptor\n");
			free((void *)curdirp);
			return -1;
#else
			corruptcount++;
			printf(" *** cannot access file/directory: %s ***\n",namebuf);
			continue; /* skip entry */
#endif
		}
		dir=fdes.fd_att&ATT_D; /* directory flag */
		
		/* file size */
		fsize=(fdes.fd_fsize[0]<<24)
			+(fdes.fd_fsize[1]<<16)
			+(fdes.fd_fsize[2]<<8)
			+fdes.fd_fsize[3];
		/* allocated */
		asize=countseg(&(fdes.fd_seg[0]),FILEDES_SEGNR,NULL);
		/* check if corrupt */
		corrupt=(fsize>asize*curpart->blksiz)||(asize>curpart->filesyssize);
		if (corrupt){
			corruptcount++;
		}else{
			totsize+=fsize;
			totalloc+=asize;
		}
		
		/* print entry */
		for (j=0;j<reclevel;j++){
			printf("    ");
		}
		printf("%s",namebuf);
		if (corrupt){
			printf(" *** invalid size ***");
		}
		if (dir){ /* directory? */
			printf("/");
		}
		printf("\n");
		
		/* directory? */
		if (dir){
			if (!corrupt){
				totdirs++;
				
				reclevel++;
				if (reclevel>maxreclevel){
					maxreclevel=reclevel;
				}
#define MAX_RECLEVEL 8 /* XXX set VC++ linker option /STACK:0x1000000 */
				if (reclevel<=MAX_RECLEVEL){
					ret=recfct(namebuf,1,arg); /* 1: recursive */
				}else{
					ret=0;
					printf("*** MAX_RECLEVEL ***\n");
				}
				reclevel--;
				/* check for error */
				if (ret<0){
					printf("*** ERROR ***\n");
					free((void *)curdirp);
					return -1;
				}
			}
		}else{
			totfiles++;
		}
	}
	
	if (reclevel==0){
		printf("\ntotal\n----------------------------------------------\n");
		printf("files:       %11u\n",totfiles);
		printf("directories: %11u\n",totdirs);
		printf("maxreclevel: %11u\n",maxreclevel);
		printf("damaged:     %11u\n",corruptcount);
		printf("data size:   %11u bytes\n",totsize);
		printf(" allocated:  %11u bytes (%06x blocks)\n\n",
			totalloc*curpart->blksiz,totalloc);
	}
	
	free((void *)curdirp);
	
	return 0;
}

/* MDR-DOS */
int
mdr_listdir(void)
{
	static struct mdr_dirent_s dirent;
	u_int fnr;
	u_int fsize;
	u_int cnr;
	
	if (curpart->type!=PARTTYPE_OS9MDR){
		return -1;
	}
	if (!curpart->mdr){
		return -1;
	}
	
	printf("fnr  creat/bk modif          size      name\n");
	printf("-------------------------------------------------------------------------------\n");
	
	/* scan directory */
	for (fnr=0;fnr<curpart->mdr_filenr;fnr++){ /* XXX first mdr_filenr entries */
		/* read directory entry */
		if (READBLOCK_CURPART(MDR_DIRSTART+fnr*MDR_DIRENTRYSIZE,
			(void *)&dirent,MDR_DIRENTRYSIZE)<0){
			err_print("mdr_listdir: READBLOCK_CURPART\n");
			return -1;
		}
		
		/* check if entry not used */
		if (dirent.name[0]==0)
			continue; /* next */
		
		/* check if correct type */
		if ((curdir.mdr_type[0]!='\0')&&(strncasecmp(dirent.type,curdir.mdr_type,4)!=0))
			continue; /* next */

		/* file size in clusters */
		fsize=(dirent.size[0]<<24)
			+(dirent.size[1]<<16)
			+(dirent.size[2]<<8)
			+dirent.size[3];
		
		/* print entry */
		printf("%4i ",fnr);
		printf("%02i-%02i-%02i ",
			dirent.back[0]%100,dirent.back[1]%100,dirent.back[2]%100);
		/* XXX don't print backup times */
		printf("%02i-%02i-%02i %02i:%02i ",
			dirent.modif[0]%100,dirent.modif[1]%100,dirent.modif[2]%100,
			dirent.modif[3]%100,dirent.modif[4]%100);
		/* XXX don't print modif seconds */
		/* count used clusters */
		cnr=mdr_countclu(&dirent,MDR_FILECLNR,NULL);
		if (fsize>cnr*curpart->mdr_clustersize*curpart->blksiz){
			printf("???????? ");
		}else{
			printf("%08x  ",fsize);
		}
		/* print type only in root directory */
		if ((curdir.mdr_type[0]=='\0')&&(dirent.type[0]!='\0')){
			int i;

			for (i=0;i<4;i++){
				if (dirent.type[i]=='\0'){
					break;
				}
				printf("%c",dirent.type[i]);
			}
			printf("/");
		}
		printf("%s\n",dirent.name); /* XXX no length check */
	}
	printf("-------------------------------------------------------------------------------\n\n");
	
	return 0;
}

/* QDOS */
int
qdos_listdir(void)
{
	u_int fnr;
	u_char dirbuf[QDOS_DIRSIZE*QDOS_BLOCKSIZE];
	struct qdos_dir_s *qdp;
	u_int ribblk;
	struct qdos_rib_s qr;
	u_int cnr;
	u_int fsize;

	if (curpart->type!=PARTTYPE_QDOS){
		return -1;
	}

	printf("fnr  rib  ov  lb sl   la   sa   size    att   fmt      name\n");
	printf("---------------------------------------------------------------------\n");
	
	/* read directory */
	if (READBLOCK_CURPART(QDOS_DIRSTART,(void *)dirbuf,QDOS_DIRSIZE)<0){
		err_print("qdos_listdir: READBLOCK_CURPART\n");
		return -1;
	}
	qdp=(struct qdos_dir_s *)dirbuf;

	/* scan directory */
	for (fnr=0;fnr<QDOS_DIRENTRYNR;fnr++){
		/* check if entry not used */
		if (qdp[fnr].dir_nm[0]==QDOS_DIR_VACANT){
			/* vacant directory entry */
			continue; /* next */
		}
		if (qdp[fnr].dir_nm[0]==QDOS_DIR_DELETED){
			/* unused directory entry */
			continue; /* next */
		}

		/* print entry */
		printf("%3i  ",fnr);

		/* read RIB */
		ribblk=(qdp[fnr].dir_rb[0]<<8)+qdp[fnr].dir_rb[1];
		if (ribblk==0){
			printf("????     ?? ???? ???? ???? ?????? ??????  ");
		}else{
			struct qdos_did_s *qp;
			u_int ov;

			printf("%04x ",ribblk);
			/* check if overlay */
			qp=QDOS_LSN0_P(curpart->lsn0);
			for (ov=0;ov<QDOS_DID_RBNR;ov++){
				if ((qp->did_rb[ov][0]==qdp[fnr].dir_rb[0])
					&&(qp->did_rb[ov][1]==qdp[fnr].dir_rb[1])){
					break; /* found */
				}
			}
			if (ov<QDOS_DID_RBNR){
				printf("%02x ",ov);
			}else{
				printf("   ");
			}

			if (READBLOCK_CURPART(ribblk,(void *)&qr,1)<0){
				printf(" ?? ???? ???? ???? ??????  ");
			}else{
				/* print RIB */
				printf(" %02x",qr.rib_lb);
				printf(" %02x%02x",qr.rib_sl[0],qr.rib_sl[1]);
				printf(" %02x%02x",qr.rib_la[0],qr.rib_la[1]);
				printf(" %02x%02x",qr.rib_sa[0],qr.rib_sa[1]);
				fsize=qdos_fsize(&qr);
				cnr=qdos_countseg(&qr,QDOS_RIB_SDWNR,NULL); /* number of allocated clusters */
				/* Note: first block of file allocation is used for RIB!!! */
				if (cnr<1){
					printf(" ??????  ");
				}else{
					if (fsize>(cnr*QDOS_CSIZE*curpart->blksiz-curpart->blksiz)){
						printf(" ??????  ");
					}else{
						printf(" %06x  ",fsize);
					}
				}
			}
		}

		/* print entry (continued)*/
		qdos_printatt0(qdp[fnr].dir_at[0]);
#if 0
		printf(" (%02x%02x)  ",qdp[fnr].dir_at[0],qdp[fnr].dir_at[1]);
#endif
		printf("  ");
		printnstr(qdp[fnr].dir_nm,8);
		printf(".");
		printnstr(qdp[fnr].dir_sx,2);
		printf("\n");
	}
	printf("---------------------------------------------------------------------\n\n");

	return 0;
}

/* general */
int
listdir(int (*recfct)(char *name,int rec,int arg),int arg)
{
	/* recursive: variables must not be static!!! */
	struct filedes_s curdirdes;
	u_int curdirsize;
	struct dirent_s *curdirp;
	u_int blknr;
	u_int i;
	char namebuf[FILENAMELEN+1];
	struct filedes_s fdes;
	u_int fdessec;
	u_int fsize,totsize;
	int dir;
	int ret;
	
	if (curpart->type==PARTTYPE_QDOS){
		return qdos_listdir(); /* non-recursive, end here! */
	}else if (curpart->type!=PARTTYPE_OS9MDR){
		return -1;
	}

	/* now OS9 or MDR-DOS */
	if (curpart->mdr){
		return mdr_listdir(); /* non-recursive, end here! */
	}
	
	/* now OS9 */
	/* XXX read whole directory file */
	
	if (getdir(curdir.dirsec,&curdirdes,&curdirsize)<0){
		return -1;
	}
	if (curdirsize==0){
		if (recfct==NULL){
			err_print("listdir: directory empty\n");
			return -1;
		}else{
			printf("*** empty directory ***\n");
			return 0; /* XXX non-fatal */
		}
	}
	
	/* read directory file */
	/* XXX 32-Bit overflow protection */
	blknr=(u_int)((((OFF_T)curdirsize)+((OFF_T)curpart->blksiz-1))
		/((OFF_T)curpart->blksiz));
	
	/* allocate memory for directory file */
	curdirp=(struct dirent_s *)malloc(blknr*curpart->blksiz);
	if (curdirp==NULL){
		perror("listdir: malloc");
		return -1;
	}
	
	/* read directory file */
	if (fileread((union genfdes_u *)&curdirdes,(void *)curdirp,0,blknr)!=(int)blknr){
		if (recfct==NULL){
			err_print("listdir: fileread(curdirdes)\n");
			free((void *)curdirp);
			return -1;
		}else{
			printf("*** cannot read directory ***\n");
			free((void *)curdirp);
			return 0; /* XXX non-fatal */
		}
	}

	/* recursive? */
	if (recfct!=NULL){
		/* scan directory entries for sub-directories */
		for (i=0;i<curdirsize/sizeof(struct dirent_s);i++){
			if (curdirp[i].dir_name[0]==0){
				continue; /* skip entry */
			}
			translatename(curdirp[i].dir_name,namebuf);
			/* skip "." and ".." */
			if ((strcmp(namebuf,".")==0)||(strcmp(namebuf,"..")==0)){
				continue; /* skip entry */
			}
			
			/* get file-descriptor */
			fdessec=(curdirp[i].dir_addr[0]<<16)
				+(curdirp[i].dir_addr[1]<<8)
				+curdirp[i].dir_addr[2];
			if (READBLOCK_CURPART(fdessec,&fdes,1)<0){
#if 0
				err_print("listdir: read file descriptor\n");
				free((void *)curdirp);
				return -1;
#else
				printf("*** cannot access file/directory: %s ***\n",namebuf);
				continue; /* skip entry */
#endif
			}
			dir=fdes.fd_att&ATT_D; /* directory flag */

			/* skip non-directory */
			if (!dir){
				continue; /* skip entry */
			}
			
			printf("entering %s\n\n",namebuf);
			ret=recfct(namebuf,1,arg); /* 1: recursive */
			printf("leaving %s\n\n",namebuf);
			/* check for error */
			if (ret<0){
				printf("*** ERROR ***\n");
				free((void *)curdirp);
				return -1;
			}
		}
	}
	
	/* print path */
	/* printdirlist();*/ /* XXX not working yet */
	printf("curdirsec: %06x curdirsize: %08x (%u bytes)\n",
		curdir.dirsec,curdirsize,curdirsize);
	printf("own  att      creat    modif         lnk size     fdes    name\n");
	printf("-------------------------------------------------------------------------------\n");
	
	/* scan directory entries */
	totsize=0;
	for (i=0;i<curdirsize/sizeof(struct dirent_s);i++){
		if (curdirp[i].dir_name[0]==0){
			continue; /* skip entry */
		}
		translatename(curdirp[i].dir_name,namebuf);
		
		/* get file-descriptor */
		fdessec=(curdirp[i].dir_addr[0]<<16)
				+(curdirp[i].dir_addr[1]<<8)
				+curdirp[i].dir_addr[2];
		if (READBLOCK_CURPART(fdessec,&fdes,1)<0){
#if 0
			err_print("listdir: read file descriptor\n");
			free((void *)curdirp);
			return -1;
#else
			printf("***                                                       %s ***\n",namebuf);
			continue; /* skip entry */
#endif
		}
		
		/* print entry */
		{
			fsize=printfiledes(curpart,&fdes,0);
			printf(" %02x%02x%02x ",
				curdirp[i].dir_addr[0],
				curdirp[i].dir_addr[1],
				curdirp[i].dir_addr[2]);
			printf(" %s",namebuf);
			if (fsize==-7){ /* corrupt? */
				printf(" ***");
				/* don't count this file */
			}else{
				totsize+=fsize;
			}
			printf("\n");
		}
	}
	printf("-------------------------------------------------------------------------------\n");
	printf("total:                                   %08x (%u bytes)\n\n",totsize,totsize);
	
	free((void *)curdirp);
	
	return 0;
}

/* general */
int
printdir3(char *path,int recursive,int tree)
{
	/* recursive via listdir: variables must not be static! */
	struct part_s *savecurpart;
	struct gendir_s savecurdir;
	int ret;
	
	if (path==NULL){
		return -1;
	}
	
	/* XXX not needed for "." */
	/* save current state */
	savecurpart=curpart;
	savecurdir=curdir;
	
	ret=changedir3(path,0); /* 0: no dirlist update */
	
	if (ret<0){
		err_print("printdir3: cannot change directory\n");
		return -1;
	}
	
	if (curpart->type==PARTTYPE_QDOS){
		/* QDOS has no directories except root! */
		recursive=0;
		tree=0;
	}else if (curpart->type==PARTTYPE_OS9MDR){
		/* now OS9 or MDR-DOS */
		if (curpart->mdr){
			/* MDR-DOS has no directories except root! */
			recursive=0;
			tree=0;
		}
	}else{
		return -1;
	}
	
	if (tree){
		ret=listdirtree(&printdir3,tree);
	}else{
		ret=listdir(recursive?(&printdir3):NULL,tree);
	}
	
	/* restore state */
	curpart=savecurpart;
	curdir=savecurdir;
	
	return ret; /* ret: if error... */
}



/* OS9 */
int
treecall(int (*callback)(char *name,struct genfile_s *fp,int flag),int recursive)
{
	/* recursive: variables must not be static!!! */
	struct filedes_s curdirdes;
	u_int curdirsize;
	struct dirent_s *curdirp;
	u_int blknr;
	u_int i;
	char namebuf[FILENAMELEN+1];
	struct genfile_s fdes;
	int ret;
	
	static u_int reclevel=0; /* static recursion level */

	if (curpart->type!=PARTTYPE_OS9MDR){
		return -1;
	}
	if (curpart->mdr){
		return -1;
	}

	/* now OS9 */
	/* XXX read whole directory file */
	
	if (getdir(curdir.dirsec,&curdirdes,&curdirsize)<0){
		return -1;
	}
	if (curdirsize==0){
		return 1; /* directory empty */
	}
	
	/* read directory file */
	/* XXX 32-Bit overflow protection */
	blknr=(u_int)((((OFF_T)curdirsize)+((OFF_T)curpart->blksiz-1))
		/((OFF_T)curpart->blksiz));
	
	/* allocate memory for directory file */
	curdirp=(struct dirent_s *)malloc(blknr*curpart->blksiz);
	if (curdirp==NULL){
		perror("listdir: malloc");
		return -1;
	}
	
	/* read directory file */
	if (fileread((union genfdes_u *)&curdirdes,(void *)curdirp,0,blknr)!=(int)blknr){
		free((void *)curdirp);
		return 2; /* cannot read directory */
	}
	
	/* scan directory entries and call handler function */
	for (i=0;i<curdirsize/sizeof(struct dirent_s);i++){
		if (curdirp[i].dir_name[0]==0){
			continue; /* skip entry */
		}
		
		translatename(curdirp[i].dir_name,namebuf);
		/* skip "." and ".." */
		if ((strcmp(namebuf,".")==0)||(strcmp(namebuf,"..")==0)){
			continue; /* skip entry */
		}
		translatename(curdirp[i].dir_name,namebuf);
		if (readfiledes4(namebuf,&fdes,0)<0){
			char b[128];
			sprintf(b,"treecall: cannot access file: %64s.\n",namebuf);
			err_print(b);
#if 0
			ret=-1;
			goto treecall_exit;
#else
			continue; /* skip entry */
#endif
		}

		/* Note: callback first: important if directory has to be created which will be entered below! */
		if (!fdes.dir){
			/* regular file */
			if (callback!=NULL){
				ret=callback(namebuf,&fdes,0); /* 0: regular file */
				/* check for fatal error */
				if (ret<0){
					goto treecall_exit;
				}
			}
		}else{
			/* directory */
			if (recursive){
				if (callback!=NULL){
					ret=callback(namebuf,&fdes,1); /* 1: before entering directory */
					/* check for fatal error */
					if (ret<0){
						goto treecall_exit;
					}
				}
				reclevel++;
#define MAX_TREERECLEVEL 8 /* XXX set VC++ linker option /STACK:0x1000000 */
				if (reclevel<=MAX_TREERECLEVEL){
					/* recursion */
					ret=treecall3(namebuf,callback,recursive);
				}else{
					ret=0;
					err_print("treecall: MAX_TREERECLEVEL\n");
				}
				reclevel--;
				/* check for fatal error */
				if (ret<0){
					goto treecall_exit;
				}
				if (callback!=NULL){
					ret=callback(namebuf,&fdes,2); /* 2: after leaving directory */
					/* check for fatal error */
					if (ret<0){
						goto treecall_exit;
					}
				}
			}else{
				if (callback!=NULL){
					ret=callback(namebuf,&fdes,3); /* 3: directory non-recursive */
					/* check for fatal error */
					if (ret<0){
						goto treecall_exit;
					}
				}
			}
		}
	}

treecall_exit:
	free((void *)curdirp);
	return ret;
}

/* MDR-DOS */
int
mdr_treecall(int (*callback)(char *name,struct genfile_s *fp,int flag),int recursive)
{
	static struct mdr_dirent_s dirent;
	u_int fnr;
	struct genfile_s fdes;
	int ret;
	char dirnamebuf[5]; /* place for type + '\0' */
	
	if (curpart->type!=PARTTYPE_OS9MDR){
		return -1;
	}
	if (!curpart->mdr){
		return -1;
	}

	/* recursive only for root directory */
	if ((recursive)&&(curdir.mdr_type[0]!='\0')){
		/* not root */
		recursive=0;
	}
	
	/* scan directory */
	for (fnr=0;fnr<curpart->mdr_filenr;fnr++){ /* XXX first mdr_filenr entries */
		/* read directory entry */
		if (READBLOCK_CURPART(MDR_DIRSTART+fnr*MDR_DIRENTRYSIZE,
			(void *)&dirent,MDR_DIRENTRYSIZE)<0){
#if 0
			err_print("mdr_treecall: READBLOCK_CURPART\n");
			return -1;
#else
			continue; /* skip entry */
#endif
		}
		
		/* check if entry not used */
		if (dirent.name[0]==0)
			continue; /* next */

		if (recursive){
			bcopy(dirent.type,dirnamebuf,4);
			dirnamebuf[4]='\0';
			if (dirent.type[0]!='\0'){
				/* enter directory: */
				mdr_changedir(MDR_TYPEROOT);
				mdr_changedir(dirent.type);
			}
		}else{
			/* check if correct type */
			if (strncasecmp(dirent.type,curdir.mdr_type,4)!=0)
				continue; /* next */
		}

		/* XXX readfiledes4 searches and reads dirent again */
		if (readfiledes4(dirent.name,&fdes,0)<0){
			char b[128];
			sprintf(b,"mdr_treecall: cannot access file: %64s.\n",dirent.name);
			err_print(b);
#if 0
			return -1;
#else
			continue; /* skip entry */
#endif
		}

		if ((recursive)&&(dirent.type[0]!='\0')){
			if (callback!=NULL){
				ret=callback(dirnamebuf,&fdes,1); /* 1: before entering directory */
				/* check for fatal error */
				if (ret<0){
					return -1;
				}
			}
		}

		if (callback!=NULL){
			ret=callback(dirent.name,&fdes,0); /* 0: regular file */
			/* check for fatal error */
			if (ret<0){
				return -1;
			}
		}

		if ((recursive)&&(dirent.type[0]!='\0')){
			if (callback!=NULL){
				ret=callback(dirnamebuf,&fdes,2); /* 2: after leaving directory */
				/* check for fatal error */
				if (ret<0){
					return -1;
				}
			}
		}
	}
	
	return 0;
}

/* QDOS */
int
qdos_treecall(int (*callback)(char *name,struct genfile_s *fp,int flag))
{
	u_int fnr;
	u_char dirbuf[QDOS_DIRSIZE*QDOS_BLOCKSIZE];
	struct qdos_dir_s *qdp;
	u_int ribblk;
	struct qdos_rib_s qr;
	char namebuf[12]; /* NNNNNNNN.SS with terminating '\0' */
	struct genfile_s fdes;
	int ret;

	if (curpart->type!=PARTTYPE_QDOS){
		return -1;
	}

	/* read directory */
	if (READBLOCK_CURPART(QDOS_DIRSTART,(void *)dirbuf,QDOS_DIRSIZE)<0){
		err_print("qdos_treecall: READBLOCK_CURPART\n");
		return -1;
	}
	qdp=(struct qdos_dir_s *)dirbuf;

	/* scan directory */
	for (fnr=0;fnr<QDOS_DIRENTRYNR;fnr++){
		/* check if entry not used */
		if (qdp[fnr].dir_nm[0]==QDOS_DIR_VACANT){
			/* vacant directory entry */
			continue; /* next */
		}
		if (qdp[fnr].dir_nm[0]==QDOS_DIR_DELETED){
			/* unused directory entry */
			continue; /* next */
		}

		/* read RIB */
		ribblk=(qdp[fnr].dir_rb[0]<<8)+qdp[fnr].dir_rb[1];
		if (ribblk==0){
			continue; /* skip entry */
		}
		if (READBLOCK_CURPART(ribblk,(void *)&qr,1)<0){
			continue; /* skip entry */
		}

		/* create name without ' ' */
		/* Note: dir_sx[2] directly follows dir_nm[8] in struct qdos_dir_s!!! */
		qdos_getfname(qdp[fnr].dir_nm,namebuf);

		/* XXX readfiledes4 searches and reads RIB again */
		if (readfiledes4(namebuf,&fdes,0)<0){
			char b[128];
			sprintf(b,"qdos_treecall: cannot access file: %64s.\n",namebuf);
			err_print(b);
#if 0
			return -1;
#else
			continue; /* skip entry */
#endif
		}

		if (callback!=NULL){
			ret=callback(namebuf,&fdes,0); /* 0: regular file */
			/* check for fatal error */
			if (ret<0){
				return -1;
			}
		}
	}

	return 0;
}

/* general */
int
treecall3(char *path,int (*callback)(char *name,struct genfile_s *fp,int flag),int recursive)
{
	/* recursive: variables must not be static! */
	struct part_s *savecurpart;
	struct gendir_s savecurdir;
	int ret;
	
	if (path==NULL){
		return -1;
	}
	
	/* XXX not needed for "." */
	/* save current state */
	savecurpart=curpart;
	savecurdir=curdir;
	
	ret=changedir3(path,0); /* 0: no dirlist update */
	if (ret<0){
		err_print("treecall3: cannot change directory\n");
		return -1;
	}

	ret=-1; /* default: error */
	if (curpart->type==PARTTYPE_QDOS){
		/* QDOS */
		/* Note: non-recursive, QDOS has no directories except root! */
		ret=qdos_treecall(callback);
	}else if (curpart->type==PARTTYPE_OS9MDR){
		/* now OS9 or MDR-DOS */
		if (curpart->mdr){
			/* MDR-DOS */
			/* Note: non-recursive, but MDR-DOS has virtual directories! */
			ret=mdr_treecall(callback,recursive);
		}else{
			/* OS9 */
			ret=treecall(callback,recursive);
			/* error messages */
			if (ret==1){
				char b[128];
				sprintf(b,"treecall3: directory empty: %64s.\n",path);
				err_print(b);
			}else if (ret==2){
				char b[128];
				sprintf(b,"treecall3: cannot read directory: %64s.\n",path);
				err_print(b);
			}
		}
	}
	
	/* restore state */
	curpart=savecurpart;
	curdir=savecurdir;
	
	return ret; /* ret: if error... */
}



/* general */
int
treecallback_fdes(char *name,struct genfile_s *fp,int flag)
{

	if (name==NULL){
		return -1;
	}
	if (fp==NULL){
		return -1;
	}

	if (flag==0){
		/* regular file */
		printf("calling %s\n",name);
		if (fp->part->type==PARTTYPE_OS9MDR){
			if (fp->part->mdr){
				/* MDR-DOS */
				printf("fnr:   %08x\n",fp->startsec-1);
				mdr_printdirent(&fp->fdes.mdrdos);
			}else{
				/* OS9 */
				printfiledes(fp->part,&fp->fdes.os9,1);
			}
		}else if (fp->part->type==PARTTYPE_QDOS){
			/* QDOS */
			printf("fnr:    %08x\n",fp->startsec-1);
			qdos_printfdes(&fp->fdes.qdos.dir,&fp->fdes.qdos.rib);
		}
	}else if ((flag==1)||(flag==3)){
		/* before entering directory or directory non-recursive */
		if (flag==1){
			printf("entering %s\n",name);
		}
		if (fp->part->type==PARTTYPE_OS9MDR){
			if (!fp->part->mdr){
				/* OS9 */
				printfiledes(fp->part,&fp->fdes.os9,1);
			}
		}
	}else if (flag==2){
		/* after leaving directory */
		printf("leaving %s\n",name);
	}else if (flag==3){
		/* directory non-recursive */
		if (fp->part->type==PARTTYPE_OS9MDR){
			if (!fp->part->mdr){
				/* OS9 */
				printfiledes(fp->part,&fp->fdes.os9,1);
			}
		}
	}

	return 0;
}



/* general */
int
treecallback_delfile(char *name,struct genfile_s *fp,int flag)
{

	if (name==NULL){
		return -1;
	}

	if (flag==0){
		/* regular file */
		delfile3(name,0); /* 0: non-directory */
	}else if (flag==2){
		/* after leaving directory */
		delfile3(name,1); /* 1: directory */
	}
	/* XXX ignore errors */

	return 0;
}



/* general */
char *treecopy_buf;
u_int treecopy_bufsiz;
int treecopy_announce;
struct part_s *treecopy_dst_curpart;
struct gendir_s treecopy_dst_curdir;

int
treecallback_copy(char *name,struct genfile_s *fp,int flag)
{
	struct part_s *savecurpart;
	struct gendir_s savecurdir;
	int ret;
	struct genfile_s dstf;
	u_int byte,byteend,bytenr;
	double xold;

	if (name==NULL){
		return -1;
	}
	if (fp==NULL){
		return -1;
	}

	/* save current state */
	savecurpart=curpart;
	savecurdir=curdir;
	/* change dir to destination */
	curpart=treecopy_dst_curpart;
	curdir=treecopy_dst_curdir;

	/* Note: we are now in the destination directory */

	ret=-1; /* default: error */
	if (flag==0){
		/* regular file */
		if (treecopy_announce){
			printf("copying %s\n",name);
		}

		/* check if destination file already exists */
		ret=findfile3(name,NULL,NULL);
		err_reset(); /* clear findfile error! */
		if (ret>0){
			/* exists */
			/* delete to overwrite old file */
			if (delfile3(name,0)<0){
				printf("cannot overwrite destination file\n");
				ret=0; /* XXX ignore error */
				goto treecallback_copy_exit;
			}
		}

		/* get source file size */
		byteend=fp->size;

		/* create destination file */
		/* same name, use fp as template */
		ret=creatfile3(name,byteend,0,fp); /* 0: non-dir */
		if (ret<0){
			printf("cannot create destination file\n");
			ret=0; /* XXX ignore error */
			goto treecallback_copy_exit;
		}

		/* open destination file */
		if (readfiledes4(name,&dstf,0)<0){ /* 0: do not touch */
			printf("cannot open destination file\n");
			ret=0; /* XXX ignore error */
			goto treecallback_copy_exit;
		}
					
		bytenr=0;
		if (byteend>treecopy_bufsiz){
			xold=0.0;
			printf("0         20        40        60        80        100%%\n.");
			fflush(NULL);
		}
		for (byte=0;byte<byteend;){
			/* number of bytes */
			bytenr=byteend-byte;
			if (bytenr>treecopy_bufsiz){
				bytenr=treecopy_bufsiz;
			}
			
			/* read */
			/* note: we can read a directory file */
			if (readfile4(fp,treecopy_buf,byte,bytenr)<0){
				ret=0; /* XXX ignore error */
				goto treecallback_copy_exit;
			}
			/* XXX bzero incomplete block at the end */
			if (byteend==byte+bytenr){
				bzero(treecopy_buf+bytenr,treecopy_bufsiz-bytenr); /* XXX all */
			}
			
			/* write */
			if (writefile4(&dstf,treecopy_buf,byte,bytenr)<0){
				ret=0; /* XXX ignore error */
				goto treecallback_copy_exit;
			}
			
			byte+=bytenr;
			if (byteend>treecopy_bufsiz){
				for (;((double)byte)/(double)byteend-xold>=0.02;xold+=0.02){
					printf(".");
					fflush(NULL);
				}
			}
		}
		if (byteend>treecopy_bufsiz){
			printf("\n");
		}
		ret=0;
	}else if (flag==1){
		/* before entering directory */
		if (treecopy_announce){
			printf("entering %s\n",name);
		}

		/* check if destination directory already exists */
		ret=readfiledes4(name,&dstf,0); /* 0: do not touch */
		err_reset(); /* clear openfile error! */
		if (ret<0){
			/* does not exist yet */
			/* create new destination directory */
			/* use fp as a template */
			if (creatfile3(name,0,1,fp)<0){ /* 1:dir */
				printf("cannot create directory\n");
				ret=-1; /* fatal */
				goto treecallback_copy_exit;
			}
		}else if (!dstf.dir){
			/* not a directory */
			printf("destination directory not valid\n");
			ret=-1; /* fatal */
			goto treecallback_copy_exit;
		}

		/* enter destination directory */
		if (changedir3(name,0)<0){ /* 0: no dirlist update */
			printf("unable to change to destination directory\n");
			ret=-1; /* fatal */
			goto treecallback_copy_exit;
		}

		/* save new destination directory */
		treecopy_dst_curpart=curpart;
		treecopy_dst_curdir=curdir;
		ret=0;
	}else if (flag==2){
		/* after leaving directory */
		if (treecopy_announce){
			printf("leaving %s\n",name);
		}

#define TREECALLBACK_COPY_DIRUP ".." /* XXX for OS9,... */
		if (changedir3(TREECALLBACK_COPY_DIRUP,0)<0){ /* 0: no dirlist update */
			printf("unable to leave destination directory\n");
			ret=-1; /* fatal */
			goto treecallback_copy_exit;
		}

		/* save new destination directory */
		treecopy_dst_curpart=curpart;
		treecopy_dst_curdir=curdir;
		ret=0;
	}else if (flag==3){
		/* directory non-recursive */
		/* nothing */
		ret=0;
	}


treecallback_copy_exit:
	/* restore state */
	curpart=savecurpart;
	curdir=savecurdir;

	err_output(stderr);
	err_reset();

	return ret;
}



/* general */
char *treevc2wav_buf;
u_int treevc2wav_bufsiz;

int
treecallback_vc2wav(char *name,struct genfile_s *fp,int flag)
{
	int l;

	if (name==NULL){
		return -1;
	}

	if (flag==0){
		/* regular file */

		/* XXX check for VC- or SV-file */
		l=strlen(name);
		if ((l<3)||
				((strcasecmp(name+l-3,".VC")!=0)
				&&(strcasecmp(name+l-3,".SV")!=0))){
			return 0; /* skip */
		}
		
		/* export subvoice */
		printf("=== %s:\n",name);
		vc2wav(fp,name,-1,treevc2wav_buf,treevc2wav_bufsiz); /* -1 means all subvoices */
		/* XXX ignore errors */
	}

	return 0;
}



/* general */
char *treemt2wav_buf;
u_int treemt2wav_bufsiz;

int
treecallback_mt2wav(char *name,struct genfile_s *fp,int flag)
{
	int l;

	if (name==NULL){
		return -1;
	}

	if (flag==0){
		/* regular file */

		/* XXX check for VC- or SV-file */
		l=strlen(name);
		if ((l<3)||(strcasecmp(name+l-3,".MT")!=0)){
			return 0; /* skip */
		}
		
		/* export wave */
		printf("=== %s:\n",name);
		mt2wav(fp,name,(u_int)-1,(u_int)-1,0,0,treemt2wav_buf,treemt2wav_bufsiz); /* -1 means all waves */
		/* XXX ignore errors */
	}

	return 0;
}



/* general */
char *eximport_buf;
u_int eximport_bufsiz;

int
export_file(struct genfile_s *fp,int outdes,u_int byteend,int aconv,int verbose)
{
	u_int byte,bytenr;
	double xold;
	u_int i;
	static char spacebuf[128]; /* must be static */
	static int spaceflag=1; /* must be static */
	u_int sc;

	if (spaceflag){
		for (i=0;i<128;i++){
			spacebuf[i]=' ';
		}
		spaceflag=0; /* not again */
	}

	bytenr=0;
	xold=0.0;
	if (verbose){
		printf("0         20        40        60        80        100%%\n.");
		fflush(NULL);
	}
	for (byte=0;byte<byteend;){
		/* number of bytes */
		bytenr=byteend-byte;
		if (bytenr>eximport_bufsiz){
			bytenr=eximport_bufsiz;
		}
		
		/* read */
		/* note: we can read a directory file */
		if (readfile4(fp,eximport_buf,byte,bytenr)<0){
			return -1;
		}
		
		/* ascii conversion */
		if (aconv){
			for (i=0;i<bytenr;i++){
				/* convert non-printable character */
#if 0
				if (eximport_buf[i]=='\n'){
					eximport_buf[i]='\n';
				}else
#endif
				if (eximport_buf[i]=='\r'){
					eximport_buf[i]='\n';
				}else if (eximport_buf[i]=='\t'){
					eximport_buf[i]=' ';
				}else if ((fp->ptype==PARTTYPE_QDOS)&&(0x80&eximport_buf[i])){
					/* QDOS multi-space */
					sc=(u_int)(0x7f&eximport_buf[i]); /* space count */
					eximport_buf[i]=' '; /* XXX also for sc==0 */
					if (sc>1){
						sc--;
						/* write space */
						if (WRITE(outdes,spacebuf,sc)!=(ssize_t)sc){
							perror("write");
							return -1;
						}
					}
				}else if (!isprint(eximport_buf[i])){
					if (isprint(0x7f&eximport_buf[i])){
						eximport_buf[i]&=0x7f;
					}else{
						eximport_buf[i]=' ';
					}
				}

				if (fp->ptype==PARTTYPE_QDOS){
					/* QDOS */
					/* XXX single byte output instead of repeated readfile4 after multi-space */
					if (WRITE(outdes,eximport_buf+i,1)!=(ssize_t)1){
						perror("write");
						return -1;
					}
				}
			}
		}
		
		if ((fp->ptype!=PARTTYPE_QDOS)||(!aconv)){
			/* write */
			/* XXX size must not exceed SSIZE_MAX! */
			if (WRITE(outdes,eximport_buf,bytenr)!=(ssize_t)bytenr){
				perror("write");
				return -1;
			}
		}
		
		byte+=bytenr;
		if (verbose){
			for (;((double)byte)/(double)byteend-xold>=0.02;xold+=0.02){
				printf(".");
				fflush(NULL);
			}
		}
	}

	return 0;
}

int
import_file(struct genfile_s *fp,int indes,u_int byteend,int aconv,int verbose)
{
	u_int byte,bytenr;
	double xold;
	u_int i;

	bytenr=0;
	xold=0.0;
	if (verbose){
		printf("0         20        40        60        80        100%%\n.");
		fflush(NULL);
	}
	for (byte=0;byte<byteend;){
		/* number of bytes */
		bytenr=byteend-byte;
		if (bytenr>eximport_bufsiz){
			bytenr=eximport_bufsiz;
		}
		
		/* read */
		/* XXX size must not exceed SSIZE_MAX! */
		if (READ(indes,eximport_buf,bytenr)!=(ssize_t)bytenr){
			perror("read");
			return -1;
		}
		/* XXX bzero incomplete block at the end */
		if (byteend==byte+bytenr){
			bzero(eximport_buf+bytenr,eximport_bufsiz-bytenr); /* XXX all */
		}
		
		/* ascii conversion */
		if (aconv){
			for (i=0;i<bytenr;i++){
				/* convert non-printable character */
#if 0
				if (eximport_buf[i]=='\r'){
					eximport_buf[i]='\r';
				}else
#endif
				if (eximport_buf[i]=='\n'){
					eximport_buf[i]='\r';
				}else if (eximport_buf[i]=='\t'){
					eximport_buf[i]=' ';
				}else if ((fp->ptype==PARTTYPE_QDOS)&&(eximport_buf[i]==' ')){
					eximport_buf[i]=(char)0x81; /* QDOS space */
				}else if (!isprint(eximport_buf[i])){
					if (fp->ptype==PARTTYPE_QDOS){
						eximport_buf[i]=(char)0x81; /* QDOS space */
					}else{
						eximport_buf[i]=' ';
					}
				}
			}
		}
		
		/* write */
		if (writefile4(fp,eximport_buf,byte,bytenr)<0){
			return -1;
		}
		
		
		byte+=bytenr;
		if (verbose){
			for (;((double)byte)/(double)byteend-xold>=0.02;xold+=0.02){
				printf(".");
				fflush(NULL);
			}
		}
	}

	return 0;
}



/* general */

int eximport2_verbose;
int eximport2_aconv;
int eximport2_fdes;

int
export2_file(char *name,struct genfile_s *fp)
{
	int outdes;
	u_int byteend;
	int ret;

	if ((name==NULL)||(fp==NULL)){
		return -1;
	}

	/* get file size */
	byteend=fp->size;
	if (byteend==0){
		fprintf(stderr,"empty file\n");
		return -1;
	}

	/* open destination file */
	/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
	if ((outdes=OPEN(name,O_RDWR|O_CREAT|O_TRUNC|O_BINARY,0600))<0){
		perror("open");
		return -1;
	}

	/* export file */
	if (eximport2_verbose){
		printf("exporting file\n");
	}
	ret=export_file(fp,outdes,byteend,eximport2_aconv,eximport2_verbose);
	if (eximport2_verbose){
		printf(".\n");
	}

	CLOSE(outdes);
	if (ret<0){					
		return -1;
	}
	if (eximport2_verbose){
		printf("file export complete\n");
	}

	/* export file descriptor */
	if (eximport2_fdes){
		char fdesname[256]; /* XXX */

		/* file descriptor file-name */
		sprintf(fdesname,"%s%s",name,FDES_EXPORT_SUF); /* XXX no overflow check */

		/* open file */
		/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
		if ((outdes=OPEN(fdesname,O_RDWR|O_CREAT|O_TRUNC|O_BINARY,0600))<0){
			perror("open fdes");
			return -1;
		}

		/* write */
		if (WRITE(outdes,fp,sizeof(struct genfile_s))!=sizeof(struct genfile_s)){
			perror("write fdes");
			CLOSE(outdes);
			return -1;
		}

		CLOSE(outdes);

		if (eximport2_verbose){
			printf("file descriptor export complete\n");
		}
	}

	return 0;
}

int
treecallback_export2_file(char *name,struct genfile_s *fp,int flag)
{

	if (name==NULL){
		return -1;
	}

#if 1
	if ((flag==0)||(flag==1)||(flag==3)){
		/* regular file or directory */
#else
	if (flag==0){
		/* regular file */
#endif
		export2_file(name,fp);
	}
	/* XXX ignore errors */

	return 0;
}

int
import2_file(char *name)
{
	struct genfile_s tmplt;
	struct genfile_s *tmpltp;
	struct genfile_s file;
	u_int size;
	int indes,fdindes;
	struct stat instat;
	int ret;
	u_int byteend;

	if (name==NULL){
		return -1;
	}

	/* open source file */
	/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
	if ((indes=OPEN(name,O_RDONLY|O_BINARY,0666))<0){
		perror("open");
		return -1;
	}

	/* get file size */
	if (fstat(indes,&instat)<0){
		perror("stat");
		CLOSE(indes);
		return -1;
	}
	size=(u_int)instat.st_size;
	byteend=(u_int)size;

	/* import file descriptor (if present) */
	/* (might be useless for destination partition type) */
	tmpltp=NULL; /* nothing yet */
	if (eximport2_fdes){
		char fdesname[256]; /* XXX */

		/* see if we can load a template file descriptor */
		/* file descriptor file-name */
		sprintf(fdesname,"%s%s",name,FDES_EXPORT_SUF); /* XXX no overflow check */
		/* open file */
		/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
		if ((fdindes=OPEN(fdesname,O_RDONLY|O_BINARY,0666))>=0){
			/* get file size */
			if (fstat(fdindes,&instat)>=0){
				size=(u_int)instat.st_size;
				/* size OK? */
				if (size>=sizeof(struct genfile_s)){
					size=sizeof(struct genfile_s);
					/* read */
					/* XXX size must not exceed SSIZE_MAX! */
					if (READ(fdindes,&tmplt,(u_int)size)==(ssize_t)size){
						/* we have a valid file descriptor */
						tmpltp=&tmplt;
						if (eximport2_verbose){
							printf("imported file descriptor\n");
						}
						/* Note: imported pointer tmplt.part useless!!! */
						tmplt.part=NULL;
						/* check if directory */
						if ((tmplt.ptype==PARTTYPE_OS9MDR)&&(!tmplt.pmdr)){
							/* OS9 */
							if (tmplt.fdes.os9.fd_att&ATT_D){
								fprintf(stderr,"warning: converting directory to regular file\n");
								tmplt.fdes.os9.fd_att&=~ATT_D; /* clear flag */
							}
						}
					}
				}
			}
			CLOSE(fdindes);
		}
	}

	/* create destination file */
	if (creatfile3(name,byteend,0,tmpltp)<0){ /* 0: non-dir */
		return -1;
	}

	/* open destination file */
	if (readfiledes4(name,&file,0)<0){ /* 0: do not touch */
		fprintf(stderr,"cannot open destination file\n");
		return -1;
	}

	/* import file */
	if (eximport2_verbose){
		printf("importing file\n");
	}
	ret=import_file(&file,indes,byteend,eximport2_aconv,eximport2_verbose);
	if (eximport2_verbose){
		printf(".\n");
	}

	CLOSE(indes);
	if (ret<0){
		return -1;
	}
	if (eximport2_verbose){
		printf("file import complete\n");
	}

	return 0;
}



/* general */
int tar_des;
int tar_dirlevel; /* level index to dirlist=dirlevel-1 */
char tar_dirlist[TAR_DIRLEVELNR][FILENAMELEN+1];

u_int
tar_checksum(u_char *hp)
{
	u_int sum;
	u_int i;
	int zeroflag;

	if (hp==NULL){
		return 0;
	}

	sum=0;
	zeroflag=1; /* still all zeroes */
	for (i=0;i<sizeof(struct tar_head_s);i++){
		if (hp[i]!=0){
			zeroflag=0;
		}
		if ((i>=148)&&(i<156)){	/* treat chksum field */
			sum+=(u_int)' '; /* space */
		}else{
			sum+=(u_int)hp[i];
		}
	}
	/* Note: sum is always >0 !!! */

	if (zeroflag){
		sum=0; /* unique value, see comment above */
	}
	return sum;
}

int
treecallback_tar_c(char *name,struct genfile_s *fp,int flag)
{
	int ret;
	u_int byteend,bytenr;
	struct tar_head_s tarhd;
	int i;
	char *pathp;
	int n,namelen;
	u_int chksum;
	u_int mode;
	u_int mtime;
	u_int uid;

	if (name==NULL){
		return -1;
	}
	if (eximport_bufsiz<TAR_BLOCKSIZE){ /* XXX see below */
		return -1;
	}

	if (flag==2){
		/* after leaving directory */
		/* one up */
		if (tar_dirlevel>0){
			tar_dirlevel--;
		}
		ret=0; /* success */
		goto treecallback_tar_c_exit;
	}
	if ((flag!=0)&&(flag!=1)){
		ret=0; /* success */
		goto treecallback_tar_c_exit;
	}

	ret=-1;

	if (flag==1){
		/* before entering directory */
		/* one down */
		if (tar_dirlevel<TAR_DIRLEVELNR){ /* space left? */
			/* save name */
			strcpy(tar_dirlist[tar_dirlevel],name);
		}
		tar_dirlevel++;
	}

	/* regular file or directory */
	
	/* create tar header */
	bzero((char *)&tarhd,sizeof(struct tar_head_s));
	
	/* name */
	pathp=tarhd.name;
	namelen=1; /* incl. '\0' */
	if (tar_dirlevel>TAR_DIRLEVELNR){
		printf("max. tar recursion level, skipping file\n");
		goto treecallback_tar_c_exit;
	}
	/* path component */
	for (i=0;i<tar_dirlevel;i++){
		n=strlen(tar_dirlist[i])+1; /* incl. '/' */
		namelen+=n;
		if (namelen>TAR_NAMELEN){
			printf("max. tar name size, skipping file\n");
			goto treecallback_tar_c_exit;
		}
		sprintf(pathp,"%s/",tar_dirlist[i]);
		pathp+=n;
	}
	if (flag==0){
		/* regular file */
		n=strlen(name);
		namelen+=n;
		if (namelen>TAR_NAMELEN){
			printf("max. tar name size, skipping file\n");
			goto treecallback_tar_c_exit;
		}
		sprintf(pathp,"%s",name);
		pathp+=n;
	}
#if 1
	printf("%s\n",tarhd.name);
#endif

#if 1
	/* user name */
	strcpy(tarhd.uname,"Fairlight"); /* XXX */
	/* group name */
	strcpy(tarhd.gname,"Fairlight"); /* XXX */
#endif

	/* mode, uid, gid */
	if ((fp->part->type==PARTTYPE_OS9MDR)&&(fp->part->mdr==0)){
		/* OS9 */
		/* mode */
		mode=0;
		if (fp->fdes.os9.fd_att&ATT_R){
			mode|=0400; /* -r-------- */
		}
		if (fp->fdes.os9.fd_att&ATT_W){
			mode|=0200; /* --w------- */
		}
		if (fp->fdes.os9.fd_att&ATT_E){
			mode|=0100; /* ---x------ */
		}
		if (fp->fdes.os9.fd_att&ATT_PR){
			mode|=0044; /* ----r--r-- */
		}
		if (fp->fdes.os9.fd_att&ATT_PW){
			mode|=0022; /* -----w--w- */
		}
		if (fp->fdes.os9.fd_att&ATT_PE){
			mode|=0011; /* ------x--x */
		}
		/* uid */
		uid=(fp->fdes.os9.fd_own[0]<<8)+fp->fdes.os9.fd_own[1];
	}else{
		mode=0600; /* XXX -rw------- */
		uid=curowner; /* XXX */
	}
	sprintf(tarhd.mode,"%07o",mode);
	sprintf(tarhd.uid,"%07o",uid);
	sprintf(tarhd.gid,"%07o",0); /* XXX */

	/* size */
	if (flag==0){
		/* regular file */
		byteend=fp->size;
	}else{
		byteend=0;
	}
	sprintf(tarhd.size,"%011o",byteend);

	/* mtime */
	if (fp->part->type==PARTTYPE_OS9MDR){
		if (fp->part->mdr==0){
			/* OS9 */
			mtime=(u_int)fgetdate(&fp->fdes.os9,0);
		}else{
			/* MDR-DOS */
			mtime=(u_int)mdr_fgetdate(&fp->fdes.mdrdos,0);
		}
	}else{
		mtime=(u_int)time(NULL); /* XXX current time */
	}
	sprintf(tarhd.mtime,"%011o",mtime);

	/* space */
	tarhd.space=' ';

	/* type */
	if (flag==1){
		/* directory */
		tarhd.type=TAR_TYPE_DIR;
	}else{
		tarhd.type=TAR_TYPE_REG;
	}

	/* magic */
	bcopy(TAR_USTAR,tarhd.ustar,8);
	
	/* checksum */
	chksum=tar_checksum((u_char *)&tarhd);
	sprintf(tarhd.chksum,"%06o",chksum);
	
	/* write tar header */
	if (WRITE(tar_des,(char *)&tarhd,sizeof(struct tar_head_s))!=sizeof(struct tar_head_s)){
		perror("write fdes");
		ret=-1;
		goto treecallback_tar_c_exit;
	}
	
	/* type */
	if (flag==1){
		/* directory */
		ret=0; /* success */
		goto treecallback_tar_c_exit;
	}

	/* export file */
	ret=export_file(fp,tar_des,byteend,0,0);
	if (ret<0){
		goto treecallback_tar_c_exit;
	}
	
	/* fill rest of TAR_BLOCKSIZE bytes */
	bytenr=byteend%TAR_BLOCKSIZE;
	bytenr=(TAR_BLOCKSIZE-bytenr)%TAR_BLOCKSIZE;
	if (bytenr>0){
		bzero(eximport_buf,bytenr);
		/* write */
		if (WRITE(tar_des,eximport_buf,bytenr)!=(int)bytenr){
			perror("write");
			ret=-1;
			goto treecallback_tar_c_exit;
		}
	}

	ret=0; /* success */

treecallback_tar_c_exit:
	err_output(stderr);
	err_reset();

	return ret;
}

int
tar_tailzero(void)
{
	u_char buf[TAR_TAILZERO_BLOCKS*TAR_BLOCKSIZE];

	/* zero blocks */
	bzero(buf,TAR_TAILZERO_BLOCKS*TAR_BLOCKSIZE);

	/* write blocks */
	if (WRITE(tar_des,buf,TAR_TAILZERO_BLOCKS*TAR_BLOCKSIZE)!=TAR_TAILZERO_BLOCKS*TAR_BLOCKSIZE){
		perror("write");
		return -1;
	}

	return 0;
}

int
directcall_tar_x(void)
{
	int ret;
	struct tar_head_s tarhd;
	u_int byteend,skip;
	u_int chksum,chksum0;
	u_int mode;
	u_int uid;
	u_int mtime;
	u_int l;
	struct genfile_s file;
	struct genfile_s tmplt;
	struct genfile_s *tmpltp;
	struct genfile_s dstf;

	/* interpret tar file */
	skip=0;
	for (;;){
		ret=-1;

		/* skip if necessary */
		if (skip>0){
			/* skip file */
			if (LSEEK(tar_des,(long)skip,SEEK_CUR)<0){
				perror("lseek");
				goto directcall_tar_x_exit;
			}
		}
		skip=0;

		/* read */
		/* XXX size must not exceed SSIZE_MAX! */
		ret=(int)READ(tar_des,(char *)&tarhd,sizeof(struct tar_head_s));
		if (ret==0){
			/* end of file */
			ret=0; /* success */
			goto directcall_tar_x_exit;
		}else if (ret!=sizeof(struct tar_head_s)){
			perror("read");
			goto directcall_tar_x_exit;
		}

		/* check header */
		chksum0=tar_checksum((u_char *)&tarhd);
		if (chksum0==0){ /* all zero block? */
			/* XXX ignore */
			continue; /* next */
		}
		/* checksum */
		sscanf(tarhd.chksum,"%o",&chksum);
		if (chksum!=chksum0){
			printf("checksum error in next header\n");
			goto directcall_tar_x_exit;
		}

		/* get size */
		sscanf(tarhd.size,"%o",&byteend);
		/* skip rest of full TAR_BLOCKSIZE */
		skip=byteend%TAR_BLOCKSIZE;
		skip=(TAR_BLOCKSIZE-skip)%TAR_BLOCKSIZE;

		/* get name */
#if 1
		printf("%s\n",tarhd.name);
#endif
		l=strlen(tarhd.name);
		if (l>0){
			/* correct name */
			if (tarhd.name[l-1]=='/'){
				tarhd.name[l-1]='\0'; /* terminate */
				l--;
			}
		}
		if (l==0){
			printf("empty name, skipping file\n");
			skip+=byteend; /* skip file */
			continue; /* next */
		}

		/* create template */
		if (curpart->type==PARTTYPE_OS9MDR){
			bzero((char *)&tmplt,sizeof(struct genfile_s));
			tmplt.ptype=curpart->type;
			tmplt.pmdr=curpart->mdr;
			if (curpart->mdr==0){
				/* OS9 */

				/* get mode */
				sscanf(tarhd.mode,"%o",&mode);
				tmplt.fdes.os9.fd_att=0;
				if (mode&0444){ /* -r--r--r-- */
					tmplt.fdes.os9.fd_att|=ATT_R;
				}
				if (mode&0222){ /* --w--w--w- */
					tmplt.fdes.os9.fd_att|=ATT_W;
				}
				if (mode&0111){ /* ---x--x--x */
					tmplt.fdes.os9.fd_att|=ATT_E;
				}
				if (mode&0004){ /* -------r-- */
					tmplt.fdes.os9.fd_att|=ATT_PR;
				}
				if (mode&0002){ /* --------w- */
					tmplt.fdes.os9.fd_att|=ATT_PW;
				}
				if (mode&0001){ /* ---------x */
					tmplt.fdes.os9.fd_att|=ATT_PE;
				}

				/* get uid */
				sscanf(tarhd.uid,"%o",&uid);
				tmplt.fdes.os9.fd_own[0]=0xff&(uid>>8);
				tmplt.fdes.os9.fd_own[1]=0xff&uid;

				/* get mtime */
				sscanf(tarhd.mtime,"%o",&mtime);
				fsetdate(&tmplt.fdes.os9,(time_t)mtime,1); /* 1: also create */
			}else{
				/* MDR-DOS */

				/* get mtime */
				sscanf(tarhd.mtime,"%o",&mtime);
				mdr_fsetdate(&tmplt.fdes.mdrdos,(time_t)mtime,1); /* 1: also create */
			}
			tmpltp=&tmplt;
		}else{
			tmpltp=NULL; /* XXX */
		}
		/* XXX ignore some other tarhd elements */

		/* parse type */
		if (tarhd.type==TAR_TYPE_DIR){
			/* directory */
			if (tmpltp!=NULL){
				tmpltp->fdes.os9.fd_att|=ATT_D;
			}
#if 1
			/* check if destination directory already exists */
			ret=readfiledes4(tarhd.name,&dstf,0); /* 0: do not touch */
			err_reset(); /* clear openfile error! */
			if (ret<0){
				/* does not exist yet */
				/* create new destination directory */
				if (creatfile3(tarhd.name,0,1,tmpltp)<0){ /* 1:dir */
					printf("cannot create directory\n");
					ret=-1;
					goto directcall_tar_x_exit;
				}
			}else if (!dstf.dir){
				/* not a directory */
				printf("destination directory not valid\n");
				ret=-1;
				goto directcall_tar_x_exit;
			}
#endif
			skip+=byteend; /* skip file */
			continue; /* next */
		}
		if ((tarhd.type!=TAR_TYPE_REG)&&(tarhd.type!=TAR_TYPE_REG0)){
			skip+=byteend; /* skip file */
			continue; /* next */
		}
		/* now, regular file */

		/* import file */
#if 0
		skip+=byteend; /* skip file */
		continue; /* next */
#endif
		/* check if destination file already exists */
		ret=findfile3(tarhd.name,NULL,NULL);
		err_reset(); /* clear findfile error! */
		if (ret>0){
			/* exists */
			/* delete to overwrite old file */
			if (delfile3(tarhd.name,0)<0){
				printf("cannot overwrite destination file\n");
				ret=-1;
				goto directcall_tar_x_exit;
			}
		}

		/* create destination file */
		if (creatfile3(tarhd.name,byteend,0,tmpltp)<0){ /* 0: non-dir */
			goto directcall_tar_x_exit;
		}
		
		/* open destination file */
		if (readfiledes4(tarhd.name,&file,0)<0){ /* 0: do not touch */
			printf("cannot open destination file\n");
			goto directcall_tar_x_exit;
		}
		
		/* import file */
		if (import_file(&file,tar_des,byteend,0,0)<0){
			goto directcall_tar_x_exit;
		}
	}

directcall_tar_x_exit:
	err_output(stderr);
	err_reset();

	return ret;
}



/* EOF */
