#ifndef __CMIOS9_TREE_H
#define __CMIOS9_TREE_H
/*
* Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2010  K. M. Indlekofer
*
* m.indlekofer@gmx.de
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



#include "cmios9.h"



/* tar header */
/* see tar.h of GNU tar Archive Format description */

/* Note: numbers in octal, trailing '\0' */
/* should be 512 bytes */
struct tar_head_s{
#define TAR_NAMELEN 100 /* incl. '\0' */
	char name[TAR_NAMELEN];
	char mode[8];
	char uid[8];
	char gid[8];
	char size[12];
	char mtime[12];
	char chksum[7];
	char space;
	char type;
	char linkname[TAR_NAMELEN];
#if 0
#define TAR_USTAR "ustar  " /* XXX old GNU format */
#else
#define TAR_USTAR "ustar\00000"
#endif
	char ustar[8];
	char uname[32];
	char gname[32];
	char devmajor[8];
	char devminor[8];
	char prefix[155];
	char zero[12];
};
#define TAR_BLOCKSIZE 512 /* in bytes */

#define TAR_TYPE_REG  '0'  /* regular file */
#define TAR_TYPE_REG0 '\0' /* regular file */
#define TAR_TYPE_DIR  '5'  /* directory */

#define TAR_TAILZERO_BLOCKS	6 /* number or zero blocks at file end */



extern int listdirtree(int (*recfct)(char *name,int rec,int arg),int arg);
extern int mdr_listdir(void);
extern int qdos_listdir(void);
extern int listdir(int (*recfct)(char *name,int rec,int arg),int arg);
extern int printdir3(char *path,int recursive,int tree);

extern int treecall(int (*callback)(char *name,struct genfile_s *fp,int flag),int recursive);
extern int mdr_treecall(int (*callback)(char *name,struct genfile_s *fp,int flag),int recursive);
extern int qdos_treecall(int (*callback)(char *name,struct genfile_s *fp,int flag));
extern int treecall3(char *path,int (*callback)(char *name,struct genfile_s *fp,int flag),int recursive);

extern int treecallback_fdes(char *name,struct genfile_s *fp,int flag);

extern int treecallback_delfile(char *name,struct genfile_s *fp,int flag);

extern char *treecopy_buf;
extern u_int treecopy_bufsiz;
extern int treecopy_announce;
extern struct part_s *treecopy_dst_curpart;
extern struct gendir_s treecopy_dst_curdir;
extern int treecallback_copy(char *name,struct genfile_s *fp,int flag);

extern char *treevc2wav_buf;
extern u_int treevc2wav_bufsiz;
extern int treecallback_vc2wav(char *name,struct genfile_s *fp,int flag);

extern char *treemt2wav_buf;
extern u_int treemt2wav_bufsiz;
extern int treecallback_mt2wav(char *name,struct genfile_s *fp,int flag);

extern char *eximport_buf;
extern u_int eximport_bufsiz;
extern int export_file(struct genfile_s *fp,int outdes,u_int byteend,int aconv,int verbose);
extern int import_file(struct genfile_s *fp,int indes,u_int byteend,int aconv,int verbose);

#define FDES_EXPORT_SUF ".#FDES#"
extern int eximport2_verbose;
extern int eximport2_aconv;
extern int eximport2_fdes;
extern int export2_file(char *name,struct genfile_s *fp);
extern int treecallback_export2_file(char *name,struct genfile_s *fp,int flag);
extern int import2_file(char *name);

extern int tar_des;
#define TAR_DIRLEVELNR 8 /* XXX */
extern int tar_dirlevel; /* level index to dirlist=dirlevel-1 */
extern char tar_dirlist[TAR_DIRLEVELNR][FILENAMELEN+1];
extern u_int tar_checksum(u_char *hp);
extern int treecallback_tar_c(char *name,struct genfile_s *fp,int flag);
extern int tar_tailzero(void);
extern int directcall_tar_x(void);



#endif /* !__CMIOS9_TREE_H */
