/*
* Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2010, 2011  K. M. Indlekofer
*
* m.indlekofer@gmx.de
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



#include "cmios9_err.h"
#include "cmios9.h"
#include "cmios9_voice.h"
#include "cmios9_wav.h"



/*#define MPCMDDEBUG*/

/* cmios9-internal mpcmd type: how to interpret sid and data */
static struct cmi3_mpcmdid_s cmi3_mpcmdid_list[]={
	CMI3_MPCMDID_VCEFF,		CMI3_MPCMDTYPE_VCEFF,	"voice effect",
	CMI3_MPCMDID_WAVE,		CMI3_MPCMDTYPE_ENT,		"wave",
	CMI3_MPCMDID_CREATE,	CMI3_MPCMDTYPE_ENT,		"create",
	CMI3_MPCMDID_CURRENT,	CMI3_MPCMDTYPE_ENT,		"current",
	CMI3_MPCMDID_SVMAP,		CMI3_MPCMDTYPE_SVMAP,	"subvoice map",
	CMI3_MPCMDID_NAME,		CMI3_MPCMDTYPE_NAMENT,	"name",
	CMI3_MPCMDID_LOAD,		CMI3_MPCMDTYPE_NAMENT,	"load",
	CMI3_MPCMDID_SVEFF,		CMI3_MPCMDTYPE_SVEFF,	"subvoice effect",
	CMI3_MPCMDID_TUNE,		CMI3_MPCMDTYPE_TUNENT,	"tune",
	CMI3_MPCMDID_RESET,		CMI3_MPCMDTYPE_ENT,		"reset",
	CMI3_MPCMDID_SVLOOPA,	CMI3_MPCMDTYPE_LOOPPT,	"subvoice loop A-side",
	CMI3_MPCMDID_DELETE,	CMI3_MPCMDTYPE_ENT,		"delete",
	CMI3_MPCMDID_NPHONY,	CMI3_MPCMDTYPE_NUMBER,	"nphony",
	CMI3_MPCMDID_CHMASK,	CMI3_MPCMDTYPE_CHMASK,	"channel mask",
	CMI3_MPCMDID_SWAP,		CMI3_MPCMDTYPE_ENT,		"swap",
	CMI3_MPCMDID_SVLOOPB,	CMI3_MPCMDTYPE_LOOPPT,	"subvoice loop B-side",
	CMI3_MPCMDID_SAVE,		CMI3_MPCMDTYPE_ENT,		"save",
	CMI3_MPCMDID_UPDATE,	CMI3_MPCMDTYPE_ENT,		"update",
	CMI3_MPCMDID_SWAP2,		CMI3_MPCMDTYPE_ENT,		"swap",
	CMI3_MPCMDID_MIDIOUTM,	CMI3_MPCMDTYPE_MIDIP,	"MIDI out map",
	CMI3_MPCMDID_MIDIINM,	CMI3_MPCMDTYPE_MIDIP,	"MIDI in map",
	CMI3_MPCMDID_CCCMD,		CMI3_MPCMDTYPE_NUMBER,	"CC command",
	CMI3_MPCMDID_COPYSV,	CMI3_MPCMDTYPE_NUMBER,	"copy subvoice",
	CMI3_MPCMDID_GETNAME,	CMI3_MPCMDTYPE_ENT,		"get name",
	CMI3_MPCMDID_MATCHNAME,	CMI3_MPCMDTYPE_ENT,		"match name",
	CMI3_MPCMDID_LDCCPROG,	CMI3_MPCMDTYPE_NUMBER,	"load CC prog",
	CMI3_MPCMDID_CHMASK2,	CMI3_MPCMDTYPE_NUMBER,	"channel mask",
	CMI3_MPCMDID_PATH,		CMI3_MPCMDTYPE_NUMBER,	"path",
	CMI3_MPCMDID_DRCCMASK,	CMI3_MPCMDTYPE_NUMBER,	"disk recorder CC mask",
	CMI3_MPCMDID_SYNC,		CMI3_MPCMDTYPE_NUMBER,	"sync",
	CMI3_MPCMDID_MIDICTRLM,	CMI3_MPCMDTYPE_CTRLMAP,	"MIDI control map",
	CMI3_MPCMDID_CHLEVELS,	CMI3_MPCMDTYPE_NUMBER,	"channel levels",
	-1,						-1,						NULL /* end of list */
};

/* mpcmd SIDs: entities for those IDs with types CMI3_MPCMDTYPE_MASK_ENT */
static struct cmi3_mpcmdentity_s cmi3_mpcmdentity_list[]={
	CMI3_MPCMDENT_IN,		"instrument",
	CMI3_MPCMDENT_VC,		"voice",
	CMI3_MPCMDENT_SV,		"subvoice",
	CMI3_MPCMDENT_SY,		"system",
	CMI3_MPCMDENT_SYWOVC,	"system w/o voices",
	CMI3_MPCMDENT_FNCT,		"function",
	CMI3_MPCMDENT_STEREOVC,	"stereo voice",
	CMI3_MPCMDENT_SVNAMES,	"subvoice names",
	CMI3_MPCMDENT_DEFVC,	"default voice",
	CMI3_MPCMDENT_DEFSV,	"default subvoice",
	CMI3_MPCMDENT_CCPAGE,	"CC page",
	CMI3_MPCMDENT_FNCTIDX,	"function index",
	CMI3_MPCMDENT_EXVC,		"external voice",
	CMI3_MPCMDENT_KBDSPLIT,	"keyboard split",
	CMI3_MPCMDENT_CURRENT,	"current",
	CMI3_MPCMDENT_SYANYVC,	"system with any voices",
	-1,						NULL /* end of list */
};

/* voice effects: for those IDs with types CMI3_MPCMDTYPE_MASK_EFF */
static struct cmi3_mpcmdeffect_s cmi3_mpcmdvceffect_list[]={
	CMI3_VCEFF_VIBRATE,			"vibrato rate         ",
	CMI3_VCEFF_VIBDEPTH,		"vibrato depth        ",
	CMI3_VCEFF_VIBDELAY,		"vibrato delay        ",
	CMI3_VCEFF_VIBATTACK,		"vibrato attack       ",
	CMI3_VCEFF_ATTACK,			"attack               ",
	CMI3_VCEFF_LEVEL,			"level                ",
	CMI3_VCEFF_RELEASE,			"release              ",
	CMI3_VCEFF_RESONANCE,		"resonance            ",
	CMI3_VCEFF_FILTER0,			"filter0              ", /* 3.00 */
	CMI3_VCEFF_TREMRATE,		"tremolo rate         ",
	CMI3_VCEFF_TREMDEPTH,		"tremolo depth        ",
	CMI3_VCEFF_TREMDELAY,		"tremolo delay        ",
	CMI3_VCEFF_TREMATTACK,		"tremolo attack       ",
	CMI3_VCEFF_PITCHBEND,		"pitch bend           ",
	CMI3_VCEFF_TUNOCTSTFINE,	"tuning(oct,st,fine)  ",
	CMI3_VCEFF_FILTER,			"filter               ",
	CMI3_VCEFF_KBDMAPSW,		"kbd map switch       ",
	CMI3_VCEFF_VIBSW,			"vibrato switch       ",
	CMI3_VCEFF_TREMSW,			"tremolo switch       ",
	CMI3_VCEFF_LEVEL2,			"level2               ",
	CMI3_VCEFF_TUNPU,			"tuning(pu)           ",
	CMI3_VCEFF_FILTER2,			"filter2              ",
	CMI3_VCEFF_PORTA,			"portamento           ",
	CMI3_VCEFF_BALANCE,			"balance              ",
	CMI3_VCEFF_PORTASW,			"portamento switch    ",
	CMI3_VCEFF_GLISSSW,			"glissando switch     ",
	CMI3_VCEFF_RATETIMESW,		"rate/time switch     ",
	-1,							NULL /* end of list */
};

/* subvoice effects: for those IDs with types CMI3_MPCMDTYPE_MASK_EFF */
static struct cmi3_mpcmdeffect_s cmi3_mpcmdsveffect_list[]={
	CMI3_SVEFF_FASTATT,			"fast attack          ",
	CMI3_SVEFF_HOLDTIME,		"hold time            ",
	CMI3_SVEFF_DECAY,			"decay                ",
	CMI3_SVEFF_SUSLEVEL,		"sustain level        ",
	CMI3_SVEFF_LEVEL,			"level                ",
	CMI3_SVEFF_FASTREL,			"fast release         ",
	CMI3_SVEFF_KEYLEVEL,		"key level            ", /* 3.00 */
	CMI3_SVEFF_KEYATT,			"key attack           ", /* 3.00 */
	CMI3_SVEFF_KEYREL,			"key release          ", /* 3.00 */
	CMI3_SVEFF_RESONANCE,		"resonance            ",
	CMI3_SVEFF_FILTER0,			"filter0              ", /* 3.00 */
	CMI3_SVEFF_SLOWATT,			"slow attack          ",
	CMI3_SVEFF_SLOWREL,			"slow release         ",
	CMI3_SVEFF_LOOPDECAY,		"loop decay           ",
	CMI3_SVEFF_TUNOCTSTFINE,	"tuning(oct,st,fine)  ",
	CMI3_SVEFF_FILTER,			"filter               ",
	CMI3_SVEFF_LINATTSW,		"linear attack switch ", /* XXX hidden??? */
	CMI3_SVEFF_ATTSW,			"attack switch        ",
	CMI3_SVEFF_RELSW,			"release switch       ",
	CMI3_SVEFF_LOOPSW,			"loop switch          ",
	CMI3_SVEFF_SLURSW,			"slur switch          ",
	CMI3_SVEFF_RELLOOPSW,		"release loop switch  ",
	CMI3_SVEFF_ALTLOOPSW,		"alternate loop switch",
	-1,							NULL /* end of list */
};

/* patch types */
static struct cmi3_mpcmdpatch_s cmi3_mpcmdpatch_list[]={
	CMI3_PATCH_CONST,		1,
	CMI3_PATCH_CONTROL,		3,
	CMI3_PATCH_SWITCH,		3,
	CMI3_PATCH_FUNC,		2,
	-1,						-1 /* end of list */
};



static void
cmi3_printcontroller(u_int c)
{

	/* check if special */
	switch (c){
	case 0x70:
		printf("KNUM");
		break;
	case 0x71:
		printf("KVEL");
		break;
	case 0x72:
		printf("KPRES");
		break;
	case 0x73:
		printf("CPRES");
		break;
	case 0x75:
		printf("KREL");
		break;
	case 0x7d:
		printf("TIME");
		break;
	default:
		printf("CT#%02u",c);
	}
}

static int
cmi3_mpcmd_interpret(u_char *p,int remain)
{
	struct cmi3_mpcmdhead_s *pp;
	u_int size;
	u_int i;

	u_int dataoff;
	u_int datasize;
	int datatype; /* cmios9-internal mpcmd datatype */
	int foundidi;
	int mpcmdtype; /* cmios9-internal mpcmd type */
	int foundentityi;
	int effecttype;
	int patchtype;
	int foundpatchi;
	int foundeffecti;
	struct cmi3_mpcmdeffect_s *effectlp;
	int patchsize;
	u_int patchrev,patchw0,patchw1,patchw2;

	if (p==NULL){
		return -1;
	}

	/* check if end */
	if (remain==0){
		/* XXX no error if last packet was not an "end-packet" */
		return 0; /* end */
	}

	/* min. size */
	if (remain<2){ /* at least size entry */
		printf("no space for mpcmd header\n");
		return -1; /* error */
	}
	
	pp=(struct cmi3_mpcmdhead_s *)p;

	size=(pp->size[0]<<8)+pp->size[1];
	/* check if end */
	if (size==0){
		/* XXX no error if last packet was not an "end-packet" */
		return 0; /* end */
	}

	/* check size */
	if (size+2>(u_int)remain){ /* packet too long? +2: size of size entry */
		printf("mpcmd packet too large (%06x)\n",size);
		return -1; /* error */
	}

	/* interpret */
	/* check if end */
	if (pp->id==CMI3_MPCMDID_END){
		return 0; /* end */
	}
	dataoff=CMI3_MPCMDHEAD_SIZE; /* behind header */
	datasize=(int)size-2;
	datatype=CMI3_MPCMDDATA_HEXDUMP; /* default: hexdump */

	/* find mpcmd id */
	foundidi=-1;
	for (i=0;;i++){
		if (cmi3_mpcmdid_list[i].id<0){ /* end of list */
			/* XXX ignore error */
			break; /* done */
		}
		if (pp->id==(u_char)cmi3_mpcmdid_list[i].id){ /* match? */
			foundidi=i;
			break; /* done */
		}
	}
	if (foundidi<0){
		printf("(%02x)",pp->id);
		goto cmi3_mpcmd_interpret_data;
	}
	if (cmi3_mpcmdid_list[foundidi].name!=NULL){
		printf("%s",cmi3_mpcmdid_list[foundidi].name);
	}else{
		printf("(%02x)",pp->id);
	}
	mpcmdtype=cmi3_mpcmdid_list[foundidi].type; /* cmios9-internal mpcmd type */

	/* parse mpcmdtype */
	if (mpcmdtype&CMI3_MPCMDTYPE_MASK_ENT){
		/* entity */
		if (mpcmdtype==CMI3_MPCMDTYPE_NAMENT){
			/* named entity */
			datatype=CMI3_MPCMDDATA_STRING; /* string */
		}else if (mpcmdtype==CMI3_MPCMDTYPE_TUNENT){
			/* tune entity */
			datatype=CMI3_MPCMDDATA_TUNING; /* tuning */
		}
		/* find entity sid */
		foundentityi=-1;
		for (i=0;;i++){
			if (cmi3_mpcmdentity_list[i].sid<0){ /* end of list */
				/* XXX ignore error */
				break; /* done */
			}
			if (pp->sid==(u_char)cmi3_mpcmdentity_list[i].sid){ /* match? */
				foundentityi=i;
				break; /* done */
			}
		}
		if (foundentityi<0){
			printf(": (%02x)",pp->sid);
			goto cmi3_mpcmd_interpret_data;
		}
		if (cmi3_mpcmdentity_list[foundentityi].name!=NULL){
			printf(": %s",cmi3_mpcmdentity_list[foundentityi].name);
		}else{
			printf(": (%02x)",pp->sid);
		}
	}else if (mpcmdtype==CMI3_MPCMDTYPE_MIDIP){
		/* MIDI port */
		if ((pp->sid==0)&&(datasize>=2)){ /* need at least id,sid,0,port */
			/* master port */
			if (isprint(p[dataoff+1])){
				printf(": master port %c",p[dataoff+1]);
			}else{
				printf(": master port (%02x)",p[dataoff+1]);
			}
		}else if (isprint(pp->sid)){
			printf(": %c",pp->sid);
		}else{
			printf(": (%02x)",pp->sid);
		}
	}else if (mpcmdtype==CMI3_MPCMDTYPE_LOOPPT){
		/* loop points */
		datatype=CMI3_MPCMDDATA_LOOPPT; /* loop points */
		printf(": %u",(u_int)pp->sid);
	}else if (mpcmdtype&CMI3_MPCMDTYPE_MASK_EFF){
		/* effect */
		/* data type */
		switch (pp->sid){
		case CMI3_MPCMDEFF_PATCH:
			datatype=CMI3_MPCMDDATA_PATCH;
			break;
		case CMI3_MPCMDEFF_CONTROL:
			datatype=CMI3_MPCMDDATA_CONTROL;
			break;
		default:
			printf(": (%02x)",pp->sid);
			/* Note: datatype still hexdump */
		}
		/* effect type */
		effecttype=-1;
		patchtype=-1;
		if (datasize>=1){ /* need at least effect type */
			/* effect type */
			effecttype=(int)(u_int)p[dataoff];
			/* data follows effect type */
			dataoff++;
			datasize--;
			/* patch type (if patch data) */
			if (datatype==CMI3_MPCMDDATA_PATCH){ /* patch data? */
				if (datasize>=1){ /* need at least patch type */
					/* effect type */
					patchtype=(int)(u_int)p[dataoff];
					/* data follows patch type */
					dataoff++;
					datasize--;
				}
			}
		}
		if (effecttype>=0){ /* effect? */
			/* find effect */
			foundeffecti=-1;
			if (mpcmdtype==CMI3_MPCMDTYPE_VCEFF){
				/* voice effect */
				effectlp=cmi3_mpcmdvceffect_list;
			}else{
				/* subvoice effect */
				effectlp=cmi3_mpcmdsveffect_list;
			}
			for (i=0;;i++){
				if (effectlp[i].type<0){ /* end of list */
					/* XXX ignore error */
					break; /* done */
				}
				if (effecttype==(u_char)effectlp[i].type){ /* match? */
					foundeffecti=i;
					break; /* done */
				}
			}
			if (foundeffecti<0){
				printf(": (%02x)",effecttype);
				goto cmi3_mpcmd_interpret_data;
			}
			if (effectlp[foundeffecti].name!=NULL){
				printf(": %s",effectlp[foundeffecti].name);
			}else{
				printf(": (%02x)",effecttype);
			}
		}
	}else if (mpcmdtype==CMI3_MPCMDTYPE_NUMBER){
		/* number */
		printf(": %u",(u_int)pp->sid);
	}else if (mpcmdtype==CMI3_MPCMDTYPE_SVMAP){
		/* subvoice map */
		datatype=CMI3_MPCMDDATA_SVMAP; /* subvoice map */
		printf(": %u",(u_int)pp->sid);
	}else if (mpcmdtype==CMI3_MPCMDTYPE_CHMASK){
		/* channel mask */
		datatype=CMI3_MPCMDDATA_CHMASK; /* channel mask */
		printf(": %u",(u_int)pp->sid); /* XXX */
	}else if (mpcmdtype==CMI3_MPCMDTYPE_CTRLMAP){
		/* control map */
		datatype=CMI3_MPCMDDATA_CTRLMAP; /* control map */
		printf(": %u",(u_int)pp->sid); /* XXX */
	}

cmi3_mpcmd_interpret_data:
	if (datasize<=0){
		goto cmi3_mpcmd_interpret_exit;
	}
	/* print data */
	printf(": ");
	if (datatype==CMI3_MPCMDDATA_PATCH){
		/* effect: patch data */
		foundpatchi=-1;
		if (patchtype>=0){
			/* find patch */
			for (i=0;;i++){
				if (cmi3_mpcmdpatch_list[i].type<0){ /* end of list */
					/* XXX ignore error */
					break; /* done */
				}
				if (patchtype==(u_char)cmi3_mpcmdpatch_list[i].type){ /* match? */
					foundpatchi=i;
					break; /* done */
				}
			}
		}
		patchsize=-1;
		if (foundpatchi>=0){
			/* known */
			patchsize=cmi3_mpcmdpatch_list[i].size;
			/* patchsize in words (2 bytes) */
			if ((patchsize<<1)!=(int)datasize){ /* incompatible? */
				patchsize=-1; /* invalid */
			}
		}
		if (patchsize<0){
			/* invalid or unknown patch */
			printf("(%02x) ",patchtype);
			for (i=0;i<datasize;i++){
				if ((i>0)&&(i%2==0)){
					printf(" ");
				}
				printf("%02x",p[dataoff+i]);
			}
		}else{
			/* valid patch */
			/* get words */
			patchw0=(p[dataoff]<<8)+p[dataoff+1];
			if ((patchtype!=CMI3_PATCH_CONST)
				&&(patchw0>=CMI3_PATCH_REVOFF)){
				/* reversed controller/switch */
				patchrev=1;
				patchw0-=CMI3_PATCH_REVOFF;
			}else{
				patchrev=0;
			}
			if (patchsize>1){
				patchw1=(p[dataoff+2]<<8)+p[dataoff+3];
			}
			if (patchsize>2){
				patchw2=(p[dataoff+4]<<8)+p[dataoff+5];
			}
			/* print */
			if (patchtype==CMI3_PATCH_CONST){
				/* constant */
				printf("%04x",patchw0);
			}else if (patchtype==CMI3_PATCH_CONTROL){
				if (patchrev){
					printf("R");
				}
				cmi3_printcontroller(patchw0);
				/* patchw2 is start, patchw1 is span */
				printf(" (range %04x-%04x)",patchw2,0xffff&(patchw2+patchw1));
			}else if (patchtype==CMI3_PATCH_SWITCH){
				if (patchrev){
					printf("R");
				}
				printf("SW#%02u",0xff&patchw0);
				/* patchw2 is start, patchw1 is span */
				printf(" (range %04x-%04x)",patchw2,0xffff&(patchw2+patchw1));
			}else if (patchtype==CMI3_PATCH_FUNC){
				printf("F%02u(",0xff&patchw1);
				if (patchrev){
					printf("R");
				}
				/* Note: only controller for function argument, no range */
				cmi3_printcontroller(patchw0);
				printf(")");
			}
		}
	}else if (datatype==CMI3_MPCMDDATA_CONTROL){
		/* effect: control data */
		if (datasize==1){
			switch (p[dataoff]){
			case 0x7f:
				printf("OFF");
				break;
			case 0xff:
				printf("ON");
				break;
			default:
				printf("MIDI%02u",(u_int)p[dataoff]);
			}
		}else{
			for (i=0;i<datasize;i++){
				printf("%02x",p[dataoff+i]);
			}
		}
	}else if ((datatype==CMI3_MPCMDDATA_TUNING)&&(datasize==4)){
		/* tuning */
		printf("%02x%02x %02x%02x",
			p[dataoff+0],p[dataoff+1],
			p[dataoff+2],p[dataoff+3]);
	}else if ((datatype==CMI3_MPCMDDATA_STRING)&&(datasize>=1)){ /* at least '\0' */
		/* string */
		printf("\"%s\"",&p[dataoff]);
	}else if ((datatype==CMI3_MPCMDDATA_LOOPPT)&&(datasize==18)){
		/* loop points */
		printf("\n         word size:  %02x%02x\n",
			p[dataoff+0],p[dataoff+1]);
		printf("         play start: %02x%02x%02x%02x\n",
			p[dataoff+2],p[dataoff+3],
			p[dataoff+4],p[dataoff+5]);
		printf("         play end+1: %02x%02x%02x%02x\n",
			p[dataoff+6],p[dataoff+7],
			p[dataoff+8],p[dataoff+9]);
		printf("         loop start: %02x%02x%02x%02x\n",
			p[dataoff+10],p[dataoff+11],
			p[dataoff+12],p[dataoff+13]);
		printf("         loop end+1: %02x%02x%02x%02x",
			p[dataoff+14],p[dataoff+15],
			p[dataoff+16],p[dataoff+17]);
	}else if ((datatype==CMI3_MPCMDDATA_SVMAP)&&(datasize==128)){
		/* subvoice map */
		printf("\n  ----");
		for (i=0;i<64;i++){
			printf("-");
		}
		printf("\n  key:");
		for (i=0;i<64;i++){
			printf("%x",0xf&(i>>4));
		}
		printf("\n      ");
		for (i=0;i<64;i++){
			printf("%x",0xf&i);
		}
		printf("\n      ");
		for (i=0;i<64;i++){
			printf(".");
		}
		printf("\n   sv:");
		for (i=0;i<64;i++){
			printf("%x",0xf&(p[dataoff+i]>>4));
		}
		printf("\n      ");
		for (i=0;i<64;i++){
			printf("%x",0xf&p[dataoff+i]);
		}
		printf("\n  ----");
		for (i=0;i<64;i++){
			printf("-");
		}
		printf("\n  key:");
		for (i=64;i<128;i++){
			printf("%x",0xf&(i>>4));
		}
		printf("\n      ");
		for (i=64;i<128;i++){
			printf("%x",0xf&i);
		}
		printf("\n      ");
		for (i=0;i<64;i++){
			printf(".");
		}
		printf("\n   sv:");
		for (i=64;i<128;i++){
			printf("%x",0xf&(p[dataoff+i]>>4));
		}
		printf("\n      ");
		for (i=64;i<128;i++){
			printf("%x",0xf&p[dataoff+i]);
		}
		printf("\n  ----");
		for (i=0;i<64;i++){
			printf("-");
		}
	}else if ((datatype==CMI3_MPCMDDATA_CHMASK)&&(datasize==4)){
		/* channel mask */
		printf("(%02x%02x %02x%02x) ",
			p[dataoff+0],p[dataoff+1],
			p[dataoff+2],p[dataoff+3]);
		/* XXX if stereo: A marked only */
		for (i=0;i<8;i++){ /* channel cards */
			printf("%c%c",
				((p[dataoff+2]>>i)&1)?'*':'-',  /* A */
				((p[dataoff+3]>>i)&1)?'*':'-'); /* B */
		}
	}else if ((datatype==CMI3_MPCMDDATA_CTRLMAP)&&(datasize==128)){
		/* MIDI control map */
		for (i=0;i<128;i++){
			if ((i&0x0f)==0){
				printf("\n                ");
			}
			printf("%02x ",p[dataoff+i]);
		}
	}else{
		/* hexdump */
		if (datasize>0){
			for (i=0;i<datasize;i++){
				printf("%02x ",p[dataoff+i]);
			}
		}
	}

cmi3_mpcmd_interpret_exit:
	printf("\n");

	/* XXX no error if last packet was not an "end-packet" */
	if ((size+2)==(u_int)remain){ /* last packet? +2: size of size entry */
		return 0; /* end */
	}

	/* +2: size of size entry, see struct mpcmd_packet_s */
	return size+2; /* advance */
}

static int
cmi3_mpcmds_interpret(u_char *p,int size)
{
	int i,inc;
	int count;
	
	if (p==NULL){
		return -1;
	}
	
	/* interpret list */
	count=0;
	for (i=0;i<size;i+=inc){
#ifdef MPCMDDEBUG
		printf("mpcmd %3i: ",count);
#endif
		inc=cmi3_mpcmd_interpret(p+i,size-i);
		if (inc==0){ /* end? */
#ifdef MPCMDDEBUG
			printf("end\n");
#endif
			break; /* done */
		}
		if (inc<0){ /* error? */
			fprintf(stderr,"error in mpcmd packet\n");
			return -1;
		}
		/* sucess */
		count++;
	}
	/* XXX no error if last packet was not an "end-packet" */
	
#if 0
	printf("count: %i\n",count);
#endif
	return count;
}



static int
cmi3_find_subvoice(struct genfile_s *fp,struct cmi3_sv_index_s *svindex,u_char sv,
				   u_int *svstartp,struct cmi3_sv_id_s *svidp)
{
	int i;
	u_int svstart;
	u_char svnumber;
	u_int foff;

	if (svindex!=NULL){
		/* user-supplied svindex and sv */

		if ((sv<CMI3_SV_MIN)||(sv>CMI3_SV_MAX)){
			return -1;
		}
		/* scan subvoices */
		for (i=0;i<CMI3_SV_INDEX_NR;i++){
			/* subvoice number */
			svnumber=svindex[i].number;
			if (svnumber==CMI3_SV_NULL){ /* end? */
				return 1; /* 1: not found */
			}
			if (svnumber==sv){
				break; /* found */
			}
		}
		if (i==CMI3_SV_INDEX_NR){
			return 1; /* 1: not found */
		}

		/* start of subvoice */
		svstart=(svindex[i].start[0]<<16)
			+(svindex[i].start[1]<<8)
			+svindex[i].start[2];
		svstart*=OS9L2_BLOCKSIZE; /* fits in u_int!!! */
		/* XXX check */
		if (svstartp!=NULL){
			*svstartp=svstart;
		}
	}else{
		/* user-supplied svstart */
		if (svstartp==NULL){
			return -1;
		}
		svstart=*svstartp;
	}

	if ((fp==NULL)||(svidp==NULL)){
		return 0; /* done */
	}
	
	/* read subvoice ID */
	foff=svstart;
	if (readfile4(fp,(char *)svidp,foff,CMI3_SV_ID_SIZE)<0){
		printf("cannot read subvoice ID for %u\n",(u_int)svnumber);
		return -1;
	}
	foff+=CMI3_SV_ID_SIZE;

	return 0;
}



static int
cmi3_find_subvoice_samples(struct genfile_s *fp,struct cmi3_sv_index_s *svindex,
						   struct cmi3_sv_id_s *svidp,u_char sv,u_int svstart,int stereo,
						   u_int *svsamplesp,u_int *bsvsamplesp,
						   u_int *svsizep,u_int *bsvsizep)
{
	u_char svforwave,bsvforwave;
	u_char svforwave_b,bsvforwave_b;
	u_int svsamples,bsvsamples;
	u_int svsize,svsize0,bsvsize;
	u_int tmpsvstart;
	u_int tmpsvsize;
	u_char tmpsvforwave;
	u_char tmpsvforwave_b;
	struct cmi3_sv_id_s tmpsvid;

	if (fp==NULL){
		return -1;
	}
	if (svidp==NULL){
		return -1;
	}

	/* A-side */
	svforwave=CMI3_SVFORWAVE_MASK&svidp->svforwave; /* sv number */
	svforwave_b=CMI3_SVFORWAVE_B&svidp->svforwave; /* in B-side? */
	/* sample size */
	svsize0=(svidp->size[0]<<24)
		+(svidp->size[1]<<16)
		+(svidp->size[2]<<8)
		+svidp->size[3];

	/* B-side */
	bsvforwave=CMI3_SVFORWAVE_MASK&svidp->bsvforwave; /* sv number */
	bsvforwave_b=CMI3_SVFORWAVE_B&svidp->bsvforwave; /* in B-side? */
	/* sample size */
	bsvsize=(svidp->bsize[0]<<24)
		+(svidp->bsize[1]<<16)
		+(svidp->bsize[2]<<8)
		+svidp->bsize[3];

	if (svsize0==0){
		/* XXX fake size of original A-side */
		svsize0=bsvsize;
	}

	if (svforwave==sv){ /* samples in this subvoice? */
		if (svforwave_b==0){
			/* in A-side */
			/* sample size */
			svsize=svsize0;
			/* sample start */
			svsamples=svstart+CMI3_SV_SAMPLE_OFF;
		}else{
			/* in B-side */
			/* sample size */
			svsize=bsvsize;
			/* sample start */
			/* add size of original A-side!!!: */
			svsamples=svstart+svsize0+CMI3_SV_SAMPLE_OFF;
		}
	}else{
		/* samples in other subvoice */
		if (svindex==NULL){
			printf("invalid subvoice referral\n");
			return -1; /* fatal */
		}
		if (cmi3_find_subvoice(fp,svindex,svforwave,&tmpsvstart,&tmpsvid)!=0){
			return -1;
		}
		if (!svforwave_b){
			/* in A-side */
			tmpsvforwave=CMI3_SVFORWAVE_MASK&tmpsvid.svforwave; /* sv number */
			tmpsvforwave_b=CMI3_SVFORWAVE_B&tmpsvid.svforwave; /* in B-side? */
		}else{
			/* in B-side */
			tmpsvforwave=CMI3_SVFORWAVE_MASK&tmpsvid.bsvforwave; /* sv number */
			tmpsvforwave_b=CMI3_SVFORWAVE_B&tmpsvid.bsvforwave; /* in B-side? */
		}
		/* XXX referral must end here */
		if ((svforwave!=tmpsvforwave)||(svforwave_b!=tmpsvforwave_b)){
			printf("invalid subvoice referral %u%c->%u%c\n",
				svforwave,svforwave_b?'B':'A',
				tmpsvforwave,tmpsvforwave_b?'B':'A');
			return -1; /* fatal */
		}
		/* get new subvoice size from svforwave */
		/* sample size A-side */
		tmpsvsize=(tmpsvid.size[0]<<24)
			+(tmpsvid.size[1]<<16)
			+(tmpsvid.size[2]<<8)
			+tmpsvid.size[3];
		if (!svforwave_b){
			/* in A-side */
			/* sample size */
			svsize=tmpsvsize;
			/* sample start */
			svsamples=tmpsvstart+CMI3_SV_SAMPLE_OFF;
		}else{
			/* in B-side */
			/* sample size */
			svsize=(tmpsvid.bsize[0]<<24)
				+(tmpsvid.bsize[1]<<16)
				+(tmpsvid.bsize[2]<<8)
				+tmpsvid.bsize[3];
			/* sample start */
			/* add size of new A-side!!!: */
			svsamples=tmpsvstart+tmpsvsize+CMI3_SV_SAMPLE_OFF;
		}
	}

	if (stereo){
		if (bsvforwave==sv){ /* samples in this subvoice? */
			if (bsvforwave_b==0){
				/* in A-side */
				/* sample size (original A-side!!!) */
				bsvsize=svsize0;
				/* sample start */
				bsvsamples=svstart+CMI3_SV_SAMPLE_OFF;
			}else{
				/* in B-side */
				/* sample size */
				/*bsvsize=bsvsize;*/
				/* sample start */
				/* add size of original A-side!!!: */
				bsvsamples=svstart+svsize0+CMI3_SV_SAMPLE_OFF;
			}
		}else{
			/* samples in other subvoice */
			if (svindex==NULL){
				printf("invalid subvoice referral\n");
				return -1; /* fatal */
			}
			if (cmi3_find_subvoice(fp,svindex,bsvforwave,&tmpsvstart,&tmpsvid)!=0){
				return -1;
			}
			if (!bsvforwave_b){
				/* in A-side */
				tmpsvforwave=CMI3_SVFORWAVE_MASK&tmpsvid.svforwave; /* sv number */
				tmpsvforwave_b=CMI3_SVFORWAVE_B&tmpsvid.svforwave; /* in B-side? */
			}else{
				/* in B-side */
				tmpsvforwave=CMI3_SVFORWAVE_MASK&tmpsvid.bsvforwave; /* sv number */
				tmpsvforwave_b=CMI3_SVFORWAVE_B&tmpsvid.bsvforwave; /* in B-side? */
			}
			/* XXX referral must end here */
			if ((bsvforwave!=tmpsvforwave)||(bsvforwave_b!=tmpsvforwave_b)){
				printf("invalid subvoice referral %u%c->%u%c\n",
					bsvforwave,bsvforwave_b?'B':'A',
					tmpsvforwave,tmpsvforwave_b?'B':'A');
				return -1; /* fatal */
			}
			/* get new subvoice size from svforwave */
			/* sample size A-side */
			tmpsvsize=(tmpsvid.size[0]<<24)
				+(tmpsvid.size[1]<<16)
				+(tmpsvid.size[2]<<8)
				+tmpsvid.size[3];
			if (!bsvforwave_b){
				/* in A-side */
				/* sample size */
				bsvsize=tmpsvsize;
				/* sample start */
				bsvsamples=tmpsvstart+CMI3_SV_SAMPLE_OFF;
			}else{
				/* in B-side */
				/* sample size */
				bsvsize=(tmpsvid.bsize[0]<<24)
					+(tmpsvid.bsize[1]<<16)
					+(tmpsvid.bsize[2]<<8)
					+tmpsvid.bsize[3];
				/* sample start */
				/* add size of new A-side!!!: */
				bsvsamples=tmpsvstart+tmpsvsize+CMI3_SV_SAMPLE_OFF;
			}
		}

#if 0
		if (svsize==0){
			/* XXX fake size A-side */
			svsize=bsvsize;
		}
		if (bsvsize==0){
			/* XXX fake size B-side */
			bsvsize=svsize;
		}

		if (svsize!=bsvsize){
			printf("inconsistent sample size A/B\n");
			return -1;
		}
#endif
	}

	if (svsamplesp!=NULL){
		*svsamplesp=svsamples;
	}
	if (bsvsamplesp!=NULL){
		*bsvsamplesp=bsvsamples;
	}
	if (svsizep!=NULL){
#if 1
		/* only svsizep given? => provide smallest if stereo */
		if (stereo&&(bsvsizep==NULL)&&(bsvsize<svsize)){
			svsize=bsvsize;
		}
#endif
		*svsizep=svsize;
	}
	if (bsvsizep!=NULL){
		*bsvsizep=bsvsize;
	}

	return 0;
}



static int
cmi3_wav2sv(int indes,
			u_int wavsize,u_int bitnr,int stereo,
			struct genfile_s *fp,u_int foffa,u_int foffb,
			char *buf,u_int bufsiz)
{
	u_int foff;

	u_int chnr;

	u_int svsize;
	u_int remain,chunk,wdone,wtotsize,wchunk;
	u_int i,j,k;
	u_char b0,b1;

	int ret;

	if (fp==NULL){
		return -1;
	}

	if (buf==NULL){
		return -1;
	}
	/* round down */
	bufsiz/=OS9L2_BLOCKSIZE;
	bufsiz*=OS9L2_BLOCKSIZE;
	if (bufsiz==0){
		return -1;
	}

	if (wavsize==0){
		printf("no samples\n");
		return -1;
	}

	if ((bitnr!=8)&&(bitnr!=16)){
		printf("invalid bitnr\n");
		return -1;
	}

	if (stereo){
		chnr=2;
	}else{
		chnr=1;
	}

	/* XXX assume that overflow check for wavsize has been done by caller */

	/* samples to be written */
	/* Note: wavsize: total bytes per channel in WAV-file */
	if (foffb<=foffa+wavsize){
		foffb=foffa+wavsize;
		svsize=wavsize;
	}else{
		svsize=foffb-foffa;
	}
	/* Note: now svsize>=wavsize */

	ret=-1;
	
	/* write samples */
	/* Note: WAV-file is little-endian, subvoice if big-endian!!! */
	remain=svsize*chnr; /* total bytes */
	wtotsize=wavsize*chnr; /* total bytes */
	wdone=0; /* bytes read from WAV-file */
	while (remain>0){
#ifdef DEBUG
		printf("remain=%08x foffa=%08x foffb=%08x\n",remain,foffa,foffb);
#endif
		if (stereo){
			/* XXX use 1/4 buffer per side, not very efficient */
			chunk=2*(bufsiz>>2); /* use 2*(1/4) buffer for output */
			if (chunk>remain){
				chunk=remain;
			}
		}else{
			chunk=bufsiz;
			if (chunk>remain){
				chunk=remain;
			}
		}
		/* read */
		if (wdone+chunk>wtotsize){
			wchunk=wtotsize-wdone;
		}else{
			wchunk=chunk;
		}
		if (wchunk>0){
			if (READ(indes,buf,wchunk)!=(int)wchunk){
				printf("cannot read samples\n");
				goto cmi3_wav2sv_exit;
			}
			if (bitnr==8){
				/* 8 bits */
				for (i=0;i<wchunk;i++){ /* wchunk!!! */
					buf[i]^=0x80; /* toggle sign-bit */
				}
			}
			wdone+=wchunk;
		}
		if (wchunk<chunk){
			/* fill rest */
			bzero(buf+wchunk,chunk-wchunk);
			/* Note: must shuffle full chunk below if stereo!!! */
		}
		if (stereo){
			j=2*(bufsiz>>2); /* A-side start */
			k=3*(bufsiz>>2); /* B-side start */
			/* shuffle sides and byte order */
			for (i=0;i<chunk;){ /* chunk!!! */
				if (bitnr==16){
					/* 16 bits */
					/* get */
					b1=buf[i];
					b0=buf[i+1];
					if (((i>>1)&1)==0){ /* side toggle */
						/* save A-side */
						buf[j]=b0;
						buf[j+1]=b1;
						j+=2;
					}else{
						/* save B-side */
						buf[k]=b0;
						buf[k+1]=b1;
						k+=2;
					}
					/* next sample */
					i+=2;
				}else{
					/* 8 bits */
					/* get */
					b0=buf[i];
					if ((i&1)==0){ /* side toggle */
						/* save A-side */
						buf[j]=b0;
						j++;
					}else{
						/* save B-side */
						buf[k]=b0;
						k++;
					}
					/* next sample */
					i++;
				}
			}
			j=2*(bufsiz>>2); /* A-side start */
			k=3*(bufsiz>>2); /* B-side start */
			/* write A-side */
			foff=foffa;
			/* use chunk/2 */
			if (writefile4(fp,buf+j,foff,chunk>>1)<0){
				printf("cannot write file A-side (at %08x %08x)\n",foff,chunk>>1);
				goto cmi3_wav2sv_exit;
			}
			foffa+=(chunk>>1);
			foff=foffa;
			/* write B-side */
			foff=foffb;
			/* use chunk/2 */
			if (writefile4(fp,buf+k,foff,chunk>>1)<0){
				printf("cannot write file B-side (at %08x %08x)\n",foff,chunk>>1);
				goto cmi3_wav2sv_exit;
			}
			foffb+=(chunk>>1);
			foff=foffb;
		}else{
			if (bitnr==16){
				/* 16 bits */
				/* shuffle byte order */
				for (i=0;i<wchunk;i+=2){ /* wchunk suffices!!! */
					b0=buf[i];
					buf[i]=buf[i+1];
					buf[i+1]=b0;
				}
			}
			/* write A-side */
			foff=foffa;
			/* use full chunk */
			if (writefile4(fp,buf,foff,chunk)<0){
				printf("cannot write file A-side (at %08x %08x)\n",foff,chunk);
				goto cmi3_wav2sv_exit;
			}
			foffa+=chunk;
			foff=foffa;
		}
		remain-=chunk;
	}
	ret=0;

cmi3_wav2sv_exit:
	return ret;
}

int
cmi3_wav2vc(char *vcname,char *wname,char *buf,u_int bufsiz)
{
	struct genfile_s vcfile;
	u_int foff;
	int ret;
	int indes;

	u_int wavsize;
	u_int chnr;
	int stereo;
	u_int samprate;
	u_int bitnr;
	char *errstr;

	u_int sv1head;
	u_int svstart;
	u_int svsize;
	u_int totsize;

	struct cmi3_vc_id_s *vcid;
	struct cmi3_sv_index_s *svindex;
	struct cmi3_sv_id_s *svid;
	struct cmi3_mpcmdhead_s *mpcmd;
	u_int tmp;

	if ((vcname==NULL)||(wname==NULL)){
		return -1;
	}

	/* voice header with no function + 1 subvoice header */
	/* start of first subvoice: */
	sv1head=CMI3_FUNC_TABLE_OFF+0;
	/* subvoice sample start: */
	svstart=sv1head+CMI3_SV_SAMPLE_OFF;
	/* XXX use buf, see below */
	if (bufsiz<svstart){
		printf("internal error: bufsiz too small\n");
		return -1;
	}

	/* open input file */
	/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
	if ((indes=OPEN(wname,O_RDONLY|O_BINARY,0666))<0){
		perror("open");
		return -1;
	}

	ret=-1;

	/* read and parse WAV-header */
	if (wav_read_head(indes,&wavsize,&chnr,&samprate,&bitnr,&errstr)<0){
		if (errstr!=NULL){
			printf("invalid or unsupported WAV-format (%s)\n",errstr);
		}
		goto cmi3_wav2vc_exit;
	}
	/* Note: now file-pointer for indes at start of samples */
	if (chnr==1){
		stereo=0;
	}else if (chnr==2){
		stereo=1;
	}else{
		printf("invalid chnr\n");
		goto cmi3_wav2vc_exit;
	}
	if ((bitnr!=8)&&(bitnr!=16)){
		printf("invalid bitnr\n");
		goto cmi3_wav2vc_exit;
	}
	/* XXX overflow check for wavsize */
	if ((wavsize>CMI3_WRAM_MAX)
		||(stereo&&(2*wavsize>CMI3_WRAM_MAX))){
			printf("too large for CMI Series III voice\n");
			return -1;
	}
	/* round up to CMI3_WARM_BLOCKSIZE */
	svsize=CMI3_WRAM_BLOCKSIZE*((wavsize+CMI3_WRAM_BLOCKSIZE-1)/CMI3_WRAM_BLOCKSIZE);

	/* create voice file */

	/* total size of voice file: */
	totsize=svstart+chnr*svsize;
	/* XXX round up to OS9L2_BLOCKSIZE */
	totsize=OS9L2_BLOCKSIZE*((totsize+OS9L2_BLOCKSIZE-1)/OS9L2_BLOCKSIZE);
	/* XXX last incomplete block might contain garbage in unused area */

	/* create voice-file */
	if (creatfile3(vcname,totsize,0,NULL)<0){ /* 0: non-dir */
		printf("cannot create voice-file\n");
		goto cmi3_wav2vc_exit;
	}
	/* open voice-file */
	if (readfiledes4(vcname,&vcfile,0)<0){ /* 0: do not touch */
		printf("cannot access voice-file\n");
		goto cmi3_wav2vc_exit;
	}

	/* create voice+subvoice header */
	/* XXX use buf, svstart bytes */

	/* clear */
	bzero(buf,svstart);

	/* voice ID */
	vcid=(struct cmi3_vc_id_s *)(buf+0);
	vcid->version[0]=CMI3_VC_ID_VER3;
	vcid->version[1]=CMI3_VC_ID_SVER1; /* XXX */
	if (stereo){
		vcid->stereo=CMI3_VC_ID_STEREO;
	}else{
		vcid->stereo=CMI3_VC_ID_MONO;
	}

	/* subvoice index, Note: just one subvoice */
	svindex=(struct cmi3_sv_index_s *)(buf+CMI3_SV_INDEX_OFF);
	/* subvoice 1 */
	tmp=sv1head/OS9L2_BLOCKSIZE;
	svindex[0].start[0]=0xff&(tmp>>16);
	svindex[0].start[1]=0xff&(tmp>>8);
	svindex[0].start[2]=0xff&tmp;
	svindex[0].number=CMI3_SV_MIN;
#if 1
	/* subvoice 2 */
	/* first block behind samples of subvoice 1: */
	/* Note: totsize already rounded up!!! */
	tmp=totsize/OS9L2_BLOCKSIZE;
	svindex[1].start[0]=0xff&(tmp>>16);
	svindex[1].start[1]=0xff&(tmp>>8);
	svindex[1].start[2]=0xff&tmp;
	svindex[1].number=CMI3_SV_NULL;
#endif

	/* voice mpcmds at CMI3_VC_MPCMDS_OFF */
	foff=CMI3_VC_MPCMDS_OFF;
	/* end: */
	mpcmd=(struct cmi3_mpcmdhead_s *)(buf+foff);
	mpcmd->size[0]=0;
	mpcmd->size[1]=2;
	mpcmd->id=CMI3_MPCMDID_END;
	foff+=CMI3_MPCMDHEAD_SIZE;

	/* no function table */

	/* subvoice ID, Note: just one subvoice */
	svid=(struct cmi3_sv_id_s *)(buf+sv1head);
	svid->svforwave=CMI3_SV_MIN; /* here! */
	svid->sasize=bitnr/8;
	/* svsize!!! */
	svid->size[0]=0xff&(svsize>>24);
	svid->size[1]=0xff&(svsize>>16);
	svid->size[2]=0xff&(svsize>>8);
	svid->size[3]=0xff&svsize;
	svid->sarate[0]=0xff&(samprate>>24);
	svid->sarate[1]=0xff&(samprate>>16);
	svid->sarate[2]=0xff&(samprate>>8);
	svid->sarate[3]=0xff&samprate;
	/* XXX must generate unique CRC */
	tmp=(u_int)rand();
	tmp=(tmp<<16)+(u_int)rand();
	if (tmp==0){
		/* must not be CRC of null waveform */
		tmp=0x01234567; /* XXX */
	}
	svid->crc[0]=0xff&(tmp>>24);
	svid->crc[1]=0xff&(tmp>>16);
	svid->crc[2]=0xff&(tmp>>8);
	svid->crc[3]=0xff&tmp;
	if (stereo){
		svid->bsvforwave=CMI3_SVFORWAVE_B|CMI3_SV_MIN; /* here! */
		svid->bsize[0]=svid->size[0];
		svid->bsize[1]=svid->size[1];
		svid->bsize[2]=svid->size[2];
		svid->bsize[3]=svid->size[3];
		/* XXX must generate unique CRC */
		tmp=(u_int)rand();
		tmp=(tmp<<16)+(u_int)rand();
		if (tmp==0){
			/* must not be CRC of null waveform */
			tmp=0x89abcdef; /* XXX */
		}
		svid->bcrc[0]=0xff&(tmp>>24);
		svid->bcrc[1]=0xff&(tmp>>16);
		svid->bcrc[2]=0xff&(tmp>>8);
		svid->bcrc[3]=0xff&tmp;
	}
	/* XXX use WAV-file name */
	strncpy(svid->name,wname,CMI3_SV_NAME_LEN-1);
	svid->name[CMI3_SV_NAME_LEN-1]='\0';

	/* subvoice mpcmds at sv1head+CMI3_SV_MPCMDS_OFF */
	foff=sv1head+CMI3_SV_MPCMDS_OFF;
	/* loop off: */
	mpcmd=(struct cmi3_mpcmdhead_s *)(buf+foff);
	mpcmd->size[0]=0;
	mpcmd->size[1]=4;
	mpcmd->id=CMI3_MPCMDID_SVEFF;
	mpcmd->sid=CMI3_MPCMDEFF_CONTROL;
	foff+=CMI3_MPCMDHEAD_SIZE;
	(buf+foff)[0]=CMI3_SVEFF_LOOPSW;
	foff++;
	(buf+foff)[0]=CMI3_CONTROL_OFF;
	foff++;
	/* loop A-side: */
	mpcmd=(struct cmi3_mpcmdhead_s *)(buf+foff);
	mpcmd->size[0]=0;
	mpcmd->size[1]=0x14;
	mpcmd->id=CMI3_MPCMDID_SVLOOPA; /* loop A-side */
	mpcmd->sid=0x01; /* XXX */
	foff+=CMI3_MPCMDHEAD_SIZE;
	tmp=svid->sasize; /* XXX */
	(buf+foff)[0]=0xff&(tmp>>8);
	(buf+foff)[1]=0xff&tmp;
	foff+=2;
	tmp=0; /* sample start */
	(buf+foff)[0]=0xff&(tmp>>24);
	(buf+foff)[1]=0xff&(tmp>>16);
	(buf+foff)[2]=0xff&(tmp>>8);
	(buf+foff)[3]=0xff&tmp;
	foff+=4;
	/* wavsize!!! */
	tmp=wavsize/svid->sasize; /* sample end+1 */
	(buf+foff)[0]=0xff&(tmp>>24);
	(buf+foff)[1]=0xff&(tmp>>16);
	(buf+foff)[2]=0xff&(tmp>>8);
	(buf+foff)[3]=0xff&tmp;
	foff+=4;
	tmp=0; /* loop start */
	(buf+foff)[0]=0xff&(tmp>>24);
	(buf+foff)[1]=0xff&(tmp>>16);
	(buf+foff)[2]=0xff&(tmp>>8);
	(buf+foff)[3]=0xff&tmp;
	foff+=4;
	/* wavsize!!! */
	tmp=wavsize/svid->sasize; /* loop end+1 */
	(buf+foff)[0]=0xff&(tmp>>24);
	(buf+foff)[1]=0xff&(tmp>>16);
	(buf+foff)[2]=0xff&(tmp>>8);
	(buf+foff)[3]=0xff&tmp;
	foff+=4;
	if (stereo){
		/* loop B-side: */
		mpcmd=(struct cmi3_mpcmdhead_s *)(buf+foff);
		mpcmd->size[0]=0;
		mpcmd->size[1]=0x14;
		mpcmd->id=CMI3_MPCMDID_SVLOOPB; /* loop B-side */
		mpcmd->sid=0x01; /* XXX */
		foff+=CMI3_MPCMDHEAD_SIZE;
		tmp=svid->sasize; /* XXX */
		(buf+foff)[0]=0xff&(tmp>>8);
		(buf+foff)[1]=0xff&tmp;
		foff+=2;
		tmp=0; /* sample start */
		(buf+foff)[0]=0xff&(tmp>>24);
		(buf+foff)[1]=0xff&(tmp>>16);
		(buf+foff)[2]=0xff&(tmp>>8);
		(buf+foff)[3]=0xff&tmp;
		foff+=4;
		/* wavsize!!! */
		tmp=wavsize/svid->sasize; /* sample end+1 */
		(buf+foff)[0]=0xff&(tmp>>24);
		(buf+foff)[1]=0xff&(tmp>>16);
		(buf+foff)[2]=0xff&(tmp>>8);
		(buf+foff)[3]=0xff&tmp;
		foff+=4;
		tmp=0; /* loop start */
		(buf+foff)[0]=0xff&(tmp>>24);
		(buf+foff)[1]=0xff&(tmp>>16);
		(buf+foff)[2]=0xff&(tmp>>8);
		(buf+foff)[3]=0xff&tmp;
		foff+=4;
		/* wavsize!!! */
		tmp=wavsize/svid->sasize; /* loop end+1 */
		(buf+foff)[0]=0xff&(tmp>>24);
		(buf+foff)[1]=0xff&(tmp>>16);
		(buf+foff)[2]=0xff&(tmp>>8);
		(buf+foff)[3]=0xff&tmp;
		foff+=4;
	}
	/* end: */
	mpcmd=(struct cmi3_mpcmdhead_s *)(buf+foff);
	mpcmd->size[0]=0;
	mpcmd->size[1]=2;
	mpcmd->id=CMI3_MPCMDID_END;
	foff+=CMI3_MPCMDHEAD_SIZE;

	/* write voice+subvoice header */
	foff=0;
	if (writefile4(&vcfile,buf,0,svstart)<0){
		printf("cannot write voice-file\n");
		goto cmi3_wav2vc_exit;
	}
	foff+=svstart;

	/* convert samples */
	printf("converting WAV to subvoice (%i)\n",CMI3_SV_MIN);
	ret=cmi3_wav2sv(indes,wavsize,bitnr,stereo,&vcfile,svstart,svstart+svsize,buf,bufsiz);

cmi3_wav2vc_exit:
	/* XXX if unsuccessful: must remove voice file */
	if (indes>=0){
		CLOSE(indes);
	}
	return ret;
}



static int
cmi3_sv2wav(struct genfile_s *fp,struct cmi3_sv_index_s *svindex,u_char sv,
				   u_int svstart,struct cmi3_sv_id_s *svidp,int stereo,
				   char *svwname,char *buf,u_int bufsiz)
{
	u_int svsamples,bsvsamples;
	u_int svsize;

	u_int foff;

	u_int totsize;
	u_int bitnr;
	u_int chnr;
	u_int samprate;
	int outdes;

	u_int remain,chunk;
	u_int foffa,foffb;
	u_int i,j,k;
	u_char b0,b1;

	int ret;

	if ((fp==NULL)||(svindex==NULL)||(svidp==NULL)||(svwname==NULL)){
		return -1;
	}

	if (buf==NULL){
		return -1;
	}
	/* round down */
	bufsiz/=OS9L2_BLOCKSIZE;
	bufsiz*=OS9L2_BLOCKSIZE;
	if (bufsiz==0){
		return -1;
	}

	/* samples for subvoice */
	if (cmi3_find_subvoice_samples(fp,svindex,svidp,sv,
		/* NULL: svsize=smallest of A/B-side */
		svstart,stereo,&svsamples,&bsvsamples,&svsize,NULL)<0){
		printf("cannot find samples\n");
		return -1;
	}
	if (svsize==0){
		printf("no samples\n");
		return -1;
	}

	/* check for file size */
	/* XXX OFF_T to prevent overflow */
	if ((((OFF_T)svsamples)+((OFF_T)svsize))>((OFF_T)fp->size)){
		svsize=fp->size-svsamples;
		printf("warning: truncated sample size (A-side overflow)\n");
	}
	if (stereo&&((((OFF_T)bsvsamples)+((OFF_T)svsize))>((OFF_T)fp->size))){
		svsize=fp->size-bsvsamples;
		printf("warning: truncated sample size (B-side overflow)\n");
	}

	/* open output file */
	/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
	if ((outdes=OPEN(svwname,O_RDWR|O_CREAT|O_TRUNC|O_BINARY,0666))<0){
		perror("open");
		return -1;
	}

	ret=-1;
	
	/* WAV parameters */
	bitnr=(svidp->sasize==CMI3_SV_ID_16BIT)?16:8;
	totsize=svsize; /* bytes per channel */
	/* sample rate */
	samprate=(svidp->sarate[0]<<24)
		+(svidp->sarate[1]<<16)
		+(svidp->sarate[2]<<8)
		+svidp->sarate[3];
	if (samprate==0){
		samprate=CMI3_SV_SAMPRATE_DEF; /* XXX */
	}
	chnr=stereo?2:1;
	/* create and write WAV-header */
	if (wav_write_head(outdes,totsize,chnr,samprate,bitnr)<0){
		printf("cannot write WAV-header\n");
		goto cmi3_sv2wav_exit;
	}

	/* write samples */
	/* Note: WAV-file is little-endian, subvoice is big-endian!!! */
	remain=totsize*chnr; /* total bytes */
	foffa=svsamples; /* start A-side */
	foffb=bsvsamples; /* start B-side */
	while (remain>0){
#ifdef DEBUG
		printf("remain=%08x foffa=%08x foffb=%08x\n",remain,foffa,foffb);
#endif
		if (stereo){
			/* XXX use 1/4 buffer per side, not very efficient */
			chunk=2*(bufsiz>>2); /* use 2*(1/4) buffer for output */
			if (chunk>remain){
				chunk=remain;
			}
			j=2*(bufsiz>>2); /* A-side start */
			k=3*(bufsiz>>2); /* B-side start */
			/* read A-side */
			foff=foffa;
			/* use chunk/2 */
			if (readfile4(fp,buf+j,foff,chunk>>1)<0){
				printf("cannot read file A-side (at %08x %08x)\n",foff,chunk>>1);
				goto cmi3_sv2wav_exit;
			}
			foffa+=(chunk>>1);
			foff=foffa;
			/* read B-side */
			foff=foffb;
			/* use chunk/2 */
			if (readfile4(fp,buf+k,foff,chunk>>1)<0){
				printf("cannot read file B-side (at %08x %08x)\n",foff,chunk>>1);
				goto cmi3_sv2wav_exit;
			}
			foffb+=(chunk>>1);
			foff=foffb;
			/* shuffle sides and byte order */
			for (i=0;i<chunk;){
				if (bitnr==16){
					/* 16 bits */
					if (((i>>1)&1)==0){ /* side toggle */
						/* get A-side */
						b0=buf[j];
						b1=buf[j+1];
						j+=2;
					}else{
						/* get B-side */
						b0=buf[k];
						b1=buf[k+1];
						k+=2;
					}
					/* save */
					buf[i]=b1;
					buf[i+1]=b0;
					/* next sample */
					i+=2;
				}else{
					/* 8 bits */
					if ((i&1)==0){ /* side toggle */
						/* get A-side */
						b0=buf[j];
						j++;
					}else{
						/* get B-side */
						b0=buf[k];
						k++;
					}
					/* save */
					buf[i]=b0;
					/* next sample */
					i++;
				}
			}
		}else{
			chunk=bufsiz;
			if (chunk>remain){
				chunk=remain;
			}
			/* read A-side */
			foff=foffa;
			/* use full chunk */
			if (readfile4(fp,buf,foff,chunk)<0){
				printf("cannot read file A-side (at %08x %08x)\n",foff,chunk);
				goto cmi3_sv2wav_exit;
			}
			foffa+=chunk;
			foff=foffa;
			if (bitnr==16){
				/* 16 bits */
				/* shuffle byte order */
				for (i=0;i<chunk;i+=2){
					b0=buf[i];
					buf[i]=buf[i+1];
					buf[i+1]=b0;
				}
			}
		}
		/* write */
		if (bitnr==8){
			/* 8 bits */
			for (i=0;i<chunk;i++){
				buf[i]^=0x80; /* toggle sign-bit */
			}
		}
		if (WRITE(outdes,buf,chunk)!=(int)chunk){
			printf("cannot write samples\n");
			goto cmi3_sv2wav_exit;
		}
		remain-=chunk;
	}
	ret=0;

cmi3_sv2wav_exit:
	if (outdes>=0){
		CLOSE(outdes);
	}
	return ret;
}

static int
cmi3_vc2wav(struct genfile_s *fp,char *wname,int sv,char *buf,u_int bufsiz)
{
	int ret;
	int allflag;
	int svmin,svmax;
	char *svwname;
	u_int wnamelen;

	u_int svstart;
	struct cmi3_vc_id_s vcid;
	struct cmi3_sv_index_s svindex[CMI3_SV_INDEX_NR];
	struct cmi3_sv_id_s svid;
	int stereo;

	u_int foff;

	if ((fp==NULL)||(wname==NULL)||(buf==NULL)){
		return -1;
	}

	/* XXX use buf for name as well */
	wnamelen=strlen(wname);
	if (bufsiz<(wnamelen+11)){ /* need space for "_SVxxxS.wav" */
		fprintf(stderr,"cmi3_vc2wav: bufsize too small\n");
		return -1;
	}

	allflag=0; /* just one subvoice */
	if (sv!=CMI3_SV_NULL){
		/* CMI Series III voice file */
		/* read voice ID */
		foff=0; /* begin of file */
		if (readfile4(fp,(char *)&vcid,foff,CMI3_VC_ID_SIZE)<0){
			printf("cannot read file\n");
			return -1;
		}
		foff+=CMI3_VC_ID_SIZE;
		
		if (vcid.version[0]!=CMI3_VC_ID_VER3){
			printf("not CMI Series III\n");
			return -1;
		}
		
		if ((vcid.version[1]!=CMI3_VC_ID_SVER0)&&(vcid.version[1]!=CMI3_VC_ID_SVER1)){
			printf("unsupported version (%x.%02x)\n",vcid.version[0],vcid.version[1]);
			return -1;
		}
		
		/* CMI Series III voice version 3.xx */
		
		stereo=vcid.stereo?1:0;

		/* read subvoice index */
		if (readfile4(fp,(char *)svindex,foff,CMI3_SV_INDEX_SIZE)<0){
			printf("cannot read subvoice index\n");
			return -1;
		}
		foff+=CMI3_SV_INDEX_SIZE;

		if ((sv>=CMI3_SV_MIN)&&(sv<=CMI3_SV_MAX)){
			/* find specific subvoice */
			if (cmi3_find_subvoice(fp,svindex,(u_char)sv,&svstart,&svid)!=0){
				printf("cannot find subvoice\n");
				return -1;
			}
		}else{
			allflag=1; /* all subvoices */
		}
	}else{
		/* CMI Series III subvoice file */
		u_int bsvsize;

		/* get subvoice id */
		svstart=0;
		if (cmi3_find_subvoice(fp,NULL,0,&svstart,&svid)!=0){
			printf("cannot get subvoice id\n");
			return -1;
		}
		
		/* XXX fake subvoice number */
		sv=CMI3_SVFORWAVE_MASK&svid.svforwave; /* sv number A-side */
		
		/* check if stereo */
		/* XXX check sample size B-side */
		bsvsize=(svid.bsize[0]<<24)
			+(svid.bsize[1]<<16)
			+(svid.bsize[2]<<8)
			+svid.bsize[3];
		if (bsvsize!=0){
			stereo=1;
		}else{
			stereo=0;
		}
	}

	/* scan subvoices */
	if (allflag){
		svmin=CMI3_SV_MIN;
		svmax=CMI3_SV_MAX;
		svwname=buf;
	}else{
		svmin=sv;
		svmax=sv;
		svwname=wname;
	}
	ret=0; /* no error yet */
	for (sv=svmin;sv<=svmax;sv++){
		if (allflag){
			/* find subvoice */
			ret=cmi3_find_subvoice(fp,svindex,(u_char)sv,&svstart,&svid);
			if (ret>0){
				/* sv not found */
				continue; /* next */
			}else if (ret<0){
				ret=-1;
				continue; /* XXX ignore error in loop */
			}
		}

		/* name of output file */
		strcpy(svwname,wname);
		if ((wnamelen>=4)&&(strcasecmp(wname+wnamelen-4,".wav")==0)){
			if (stereo){
				sprintf(svwname+wnamelen-4,"_SV%03iS%s",sv,wname+wnamelen-4);
			}else{
				sprintf(svwname+wnamelen-4,"_SV%03i%s",sv,wname+wnamelen-4);
			}
		}else{
			if (stereo){
				sprintf(svwname+wnamelen,"_SV%03iS.wav",sv);
			}else{
				sprintf(svwname+wnamelen,"_SV%03i.wav",sv);
			}
		}

		/* convert subvoice samples */
		printf("converting subvoice (%i)\n",sv);
		ret=cmi3_sv2wav(fp,(sv==CMI3_SV_NULL)?NULL:svindex,
			(u_char)sv,svstart,&svid,stereo,svwname,buf,bufsiz);
		/* XXX ignore error in loop */
	}
	if (allflag){
		ret=0; /* XXX ignore error */
	}

	return ret;
}



int
vc2wav(struct genfile_s *fp,char *wname,int sv,char *buf,u_int bufsiz)
{
	u_char c[2];

	u_int foff;

	if (wname==NULL){
		return -1;
	}
	
	/* read voice ID */
	foff=0; /* begin of file */
	if (readfile4(fp,(char *)c,foff,2)<0){
		printf("cannot read file\n");
		return -1;
	}
	foff+=2;

	if ((sv==CMI3_SV_NULL)
		||((c[0]==0x00)&&(c[1]==0x00))){ /* XXX typical */
		/* CMI Series III subvoice */
		return cmi3_vc2wav(fp,wname,CMI3_SV_NULL,buf,bufsiz);
	}

	if (c[0]==CMI3_VC_ID_VER3){
		/* CMI Series III */
		return cmi3_vc2wav(fp,wname,sv,buf,bufsiz);
	}

	if ((c[0]==CMI_VC_MODE1)||(c[0]==CMI_VC_MODE4)){
		/* CMI Series I/II/IIx */

		if (((c[1]==CMI_VC_VER3)||(c[0]==CMI_VC_VER4))
			&&(fp->size>=CMI_VC_FILE_SIZE*QDOS_BLOCKSIZE)){ /* XXX */
			/* CMI Series I/II/IIx voice version 3 and 4 */
			int outdes;
			u_int totsize;
			
			if (c[0]==CMI_VC_MODE4){
				totsize=CMI_VC_SAMPLE_SIZE4*QDOS_BLOCKSIZE;
			}else{
				totsize=CMI_VC_SAMPLE_SIZE1*QDOS_BLOCKSIZE;
			}
			
			if ((buf==NULL)||(bufsiz<totsize)){
				printf("buffer too small\n");
				return -1;
			}
			
			/* XXX ignore sv */
			
			/* read samples */
			foff=CMI_VC_SAMPLE_START*QDOS_BLOCKSIZE;
			if (readfile4(fp,buf,foff,totsize)<0){
				printf("cannot read samples\n");
				return -1;
			}
			foff+=totsize;
			
			/* open output file */
			/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
			if ((outdes=OPEN(wname,O_RDWR|O_CREAT|O_TRUNC|O_BINARY,0666))<0){
				perror("open");
				return -1;
			}
			
			/* create and write WAV-header */
			if (wav_write_head(outdes,totsize,1,CMI_VC_SAMPRATE_MAX,8)<0){ /* 1==mono */
				printf("cannot write WAV-header\n");
				CLOSE(outdes);
				return -1;
			}
			
			/* write samples */
			if (WRITE(outdes,buf,totsize)!=(int)totsize){
				printf("cannot write samples\n");
				CLOSE(outdes);
				return -1;
			}
			
			CLOSE(outdes);
			return 0;
		}
	}

	printf("voice file format (%u.%02u) not supported\n",(u_int)c[0],(u_int)c[1]);

	return -1;
}



static int
subvoice3_info(struct genfile_s *fp,struct cmi3_sv_index_s *svindex,
			   struct cmi3_sv_id_s *svidp,u_char sv,u_int svstart,int stereo,int verbose)
{
	u_int svsamples,bsvsamples;
	u_int svsize,bsvsize;
	u_int sarate;
	u_int sacrc;
	u_char svforwave,bsvforwave;
	u_char svforwave_b,bsvforwave_b;
	u_int foff;

	if (svidp==NULL){
		return -1;
	}

	if (verbose){
		/* subvoice mpcmds */
		u_int svmpcmds;
		u_char buf[CMI3_SV_MPCMDS_SIZE];

		svmpcmds=svstart+CMI3_SV_MPCMDS_OFF;
		
		/* read */
		foff=svmpcmds;
		if (readfile4(fp,(char *)buf,foff,CMI3_SV_MPCMDS_SIZE)<0){
			printf("cannot read subvoice mpcmds\n");
			return -1;
		}
		foff+=CMI3_SV_MPCMDS_SIZE;
		
		/* interpret */
		if (cmi3_mpcmds_interpret(buf,CMI3_SV_MPCMDS_SIZE)<0){ /* 1: subvoice */
			return -1;
		}
	}

	/* samples for subvoice */
	if (cmi3_find_subvoice_samples(fp,svindex,svidp,sv,
		svstart,stereo,&svsamples,&bsvsamples,&svsize,&bsvsize)<0){
		printf("cannot find samples\n");
		return -1;
	}
	svforwave=CMI3_SVFORWAVE_MASK&svidp->svforwave; /* sv number */
	svforwave_b=CMI3_SVFORWAVE_B&svidp->svforwave; /* in B-side? */
	bsvforwave=CMI3_SVFORWAVE_MASK&svidp->bsvforwave; /* sv number */
	bsvforwave_b=CMI3_SVFORWAVE_B&svidp->bsvforwave; /* in B-side? */
	if (verbose){
		if (svforwave!=sv){
			printf("sample A in subvoice: %3i%c\n",
				svforwave,
				svforwave_b?'B':'A');
		}
		printf("sample A start: %08x\n",svsamples);
		if (stereo){
			if (bsvforwave!=sv){
				printf("sample B in subvoice: %3i%c\n",
					bsvforwave,
					bsvforwave_b?'B':'A');
			}
			printf("sample B start: %08x\n",bsvsamples);
		}
	}else{
#if 1
		printf("  %08x",svsamples);
		if (stereo){
			printf(" %08x",bsvsamples);
		}
#else
		if (svforwave!=sv){
			printf("  ->  %3i%c",
				svforwave,
				svforwave_b?'B':'A');
		}else{
			printf("  %08x",svsamples);
		}
		if (stereo){
			if (bsvforwave!=sv){
				printf("  ->  %3i%c         ",
					bsvforwave,
					bsvforwave_b?'B':'A');
			}else{
				printf(" %08x",bsvsamples);
			}
		}
#endif
	}
	if (verbose){
		printf("sample A size:  %08x\n",svsize);
		if (stereo&&(bsvsize!=svsize)){
			printf("sample B size:  %08x\n",bsvsize);
		}
	}else{
		/* XXX */
		if (stereo&&(bsvsize<svsize)){
			svsize=bsvsize;
		}
		printf(" %08x",svsize);
	}
	
	/* Bit number */
	if (verbose){
		if (svidp->sasize==CMI3_SV_ID_16BIT){
			printf("bits: 16\n");
		}else{
			printf("bits: 8\n");
		}
	}else{
		if (svidp->sasize==CMI3_SV_ID_16BIT){
			printf("  16");
		}else{
			printf("   8");
		}
	}
	
	/* sample rate */
	sarate=(svidp->sarate[0]<<24)
		+(svidp->sarate[1]<<16)
		+(svidp->sarate[2]<<8)
		+svidp->sarate[3];
	if (verbose){
		printf("sample rate: %i\n",sarate);
	}else{
		printf(" %6i",sarate);
	}
	
	/* ID */
	sacrc=(svidp->crc[0]<<24)
		+(svidp->crc[1]<<16)
		+(svidp->crc[2]<<8)
		+svidp->crc[3];
	if (verbose){
		printf("sample A CRC: %08x\n",sacrc);
	}
	if (stereo){
		sacrc=(svidp->bcrc[0]<<24)
			+(svidp->bcrc[1]<<16)
			+(svidp->bcrc[2]<<8)
			+svidp->bcrc[3];
		if (verbose){
			printf("sample B CRC: %08x\n",sacrc);
		}
	}
	
	/* name */
	if (verbose){
		printf("name: \"%s\"\n",svidp->name);
	}else{
		printf("  %s\n",svidp->name);
	}

	return 0;
}



static int
voice3_info(struct genfile_s *fp,int verbose)
{
	u_int psize;
	u_int fsize;
	
	u_char sv;
	u_int svstart;
	struct cmi3_vc_id_s vcid;
	struct cmi3_sv_index_s svindex[CMI3_SV_INDEX_NR];
	struct cmi3_sv_id_s svid;

	int i;

	u_int foff;

	/* read voice ID */
	foff=0; /* begin of file */
	if (readfile4(fp,(char *)&vcid,foff,CMI3_VC_ID_SIZE)<0){
		printf("cannot read file\n");
		return -1;
	}
	foff+=CMI3_VC_ID_SIZE;

	if (vcid.version[0]!=CMI3_VC_ID_VER3){
		printf("not CMI Series III\n");
		return -1;
	}

	printf("CMI Series III voice file\n\n");
	printf("version:  %x.%02x\n",vcid.version[0],vcid.version[1]);

	if ((vcid.version[1]!=CMI3_VC_ID_SVER0)&&(vcid.version[1]!=CMI3_VC_ID_SVER1)){
		return -1;
	}

	/* CMI Series III voice version 3.xx */

	printf("mode:     %s\n",vcid.stereo?"STEREO":"MONO");
	
	/* mpcmds size */
	psize=(vcid.psize[0]<<8)+vcid.psize[1];
	if ((psize!=0)&&(psize!=CMI3_VC_MPCMDS_SIZE)){
		printf("incompatible voice mpcmds size\n");
		return -1;
	}
	
	/* functions */
	fsize=(vcid.fsize[0]<<8)+vcid.fsize[1];
	printf("funcs:    %i\n",fsize/CMI3_FUNC_TABLE_SIZE);
	printf("funcsmax: %i\n",vcid.funcsmax);
	if (fsize>0){
		printf("funcoff:  %08x (%i Bytes per table)\n",
			CMI3_FUNC_TABLE_OFF,CMI3_FUNC_TABLE_SIZE);
	}
	
	/* read subvoice index */
	if (readfile4(fp,(char *)svindex,foff,CMI3_SV_INDEX_SIZE)<0){
		printf("cannot read subvoice index\n");
		return -1;
	}
	foff+=CMI3_SV_INDEX_SIZE;
	
	if (verbose){
		/* voice mpcmds */
		u_int vcmpcmds;
		u_char buf[CMI3_VC_MPCMDS_SIZE];
		
		vcmpcmds=CMI3_VC_MPCMDS_OFF;

		/* read */
		foff=vcmpcmds;
		if (readfile4(fp,(char *)buf,foff,CMI3_VC_MPCMDS_SIZE)<0){
			printf("cannot read voice mpcmds\n");
			return -1;
		}
		foff+=CMI3_VC_MPCMDS_SIZE;
		
		/* interpret */
		if (cmi3_mpcmds_interpret(buf,CMI3_VC_MPCMDS_SIZE)<0){ /* 0: voice */
			return -1;
		}
	}
	
	if (!verbose){
		if (vcid.stereo){
			printf("\nsubvoices:\n sv start     sampleA  sampleB  size     bit  rate   name\n");
		}else{
			printf("\nsubvoices:\n sv start     sample   size     bit  rate   name\n");
		}
		printf("--------------------------------------------------------------------------\n");
	}
	
	/* scan subvoices */
	for (sv=CMI3_SV_MIN;sv<=CMI3_SV_MAX;sv++){
		/* find subvoice */
		i=cmi3_find_subvoice(fp,svindex,sv,&svstart,&svid);
		if (i>0){
			/* sv not found */
			continue; /* next */
		}else if (i<0){
			return -1; /* fatal */
		}
		/* XXX check svstart */
		
		if (verbose){
			printf("\nsubvoice: %3i\n",sv);
			printf("start: %08x\n",svstart);
		}else{
			printf("%3i",sv);
			printf(" %08x",svstart);
		}
		
		if (subvoice3_info(fp,svindex,&svid,sv,svstart,vcid.stereo,verbose)<0){
			continue; /* XXX ignore error */
		}
	}
	printf("\n");

	return 0;
}



int
printinfo_voice(struct genfile_s *fp,int verbose)
{
	u_char c[2];

	u_int foff;
	
	/* read voice ID */
	foff=0; /* begin of file */
	if (readfile4(fp,(char *)c,foff,2)<0){
		printf("cannot read file\n");
		return -1;
	}
	foff+=2;

	if (c[0]==CMI3_VC_ID_VER3){
		/* CMI Series III */
		return voice3_info(fp,verbose);
	}

	if ((c[0]==CMI_VC_MODE1)||(c[0]==CMI_VC_MODE4)){ /* XXX */
		/* CMI Series I/II/IIx */
		printf("CMI Series I/II/IIx voice file\n\n");
		printf("version:  %i\n",(int)c[1]); /* XXX */
		printf("mode:     %i\n",(int)c[0]); /* XXX */
		printf("size:     %04x\n",fp->size);
		/* XXX */
		if ((c[1]==CMI_VC_VER3)||(c[1]==CMI_VC_VER4)){
			/* CMI Series I/II/IIx voice version 3 and 4 */
			struct cmi_vc_des_s vcdes;

			/* read voice descriptor */
			foff=0; /* begin of file */
			if (readfile4(fp,(char *)&vcdes,foff,CMI_VC_DES_SIZE*QDOS_BLOCKSIZE)<0){
				printf("cannot read voice descriptor\n");
				return -1;
			}
			foff+=CMI_VC_DES_SIZE*QDOS_BLOCKSIZE;

			if (vcdes.mode!=vcdes.mode2){
				printf("corrupted mode\n");
			}

			if (vcdes.mode==CMI_VC_MODE4){
				printf("\n");
				printf("swav: %02x\n",(u_int)vcdes.swav);
				printf("ewav: %02x\n",(u_int)vcdes.ewav);
				printf("lflg: %02x\n",(u_int)vcdes.lflg);
			}

			if (vcdes.mode==CMI_VC_MODE1){
				int i;

				printf("\nsegments:\n");
				for (i=0;i<CMI_LOOP1SEG_NR;i++){
					printf("  %02x: slp:%02x\n",
						vcdes.loop1seg[i].nr,
						vcdes.loop1seg[i].slp);
				}
			}
		}
		printf("\n");
		return 0;
	}

	printf("unknown voice file format (%u.%02u)\n",(u_int)c[0],(u_int)c[1]);

	return -1;
}



int
printinfo_subvoice(struct genfile_s *fp,int verbose)
{
	int ret;
	int stereo;
	u_int svstart;
	struct cmi3_sv_id_s svid;
	u_char sv;
	u_int bsvsize;

	printf("CMI Series III subvoice file\n\n");

	/* get subvoice id */
	svstart=0;
	ret=cmi3_find_subvoice(fp,NULL,0,&svstart,&svid);
	if (ret!=0){
		printf("cannot get subvoice id\n");
		return -1;
	}

	/* XXX fake subvoice number */
	sv=CMI3_SVFORWAVE_MASK&svid.svforwave; /* sv number A-side */

	/* check if stereo */
	/* XXX check sample size B-side */
	bsvsize=(svid.bsize[0]<<24)
		+(svid.bsize[1]<<16)
		+(svid.bsize[2]<<8)
		+svid.bsize[3];
	if (bsvsize!=0){
		stereo=1;
	}else{
		stereo=0;
	}

	ret=subvoice3_info(fp,NULL,&svid,sv,svstart,stereo,verbose);

	printf("\n");
	return ret;
}



int
printinfo_system(struct genfile_s *fp,int verbose)
{
	u_int size;
	u_int foff;
	char *buf;
	int ret;

	size=fp->size;

	buf=(char *)malloc(size);
	if (buf==NULL){
		printf("cannot allocate buffer\n");
		return -1;
	}

	ret=0;
	
	/* read file */
	foff=0; /* begin of file */
	if (readfile4(fp,buf,foff,size)<0){
		printf("cannot read file\n");
		ret=-1;
		goto printinfo_system_exit;
	}
	foff+=size;

	/* check if valid CMI Series III system file */
	/* XXX */

	printf("CMI Series III system file\n\n");

	if (verbose){
		/* interpret mpcmd packets */
		ret=cmi3_mpcmds_interpret((u_char *)buf,size);
		printf("\n");
	}

printinfo_system_exit:
	free(buf);
	return ret;
}



int
printinfo_instrument(struct genfile_s *fp,int verbose)
{
	u_int size;
	u_int foff;
	char *buf;
	int ret;

	size=fp->size;

	buf=(char *)malloc(size);
	if (buf==NULL){
		printf("cannot allocate buffer\n");
		return -1;
	}

	ret=0;
	
	/* read file */
	foff=0; /* begin of file */
	if (readfile4(fp,buf,foff,size)<0){
		printf("cannot read file\n");
		ret=-1;
		goto printinfo_instrument_exit;
	}
	foff+=size;

	/* XXX */
	if (fp->part->type==PARTTYPE_QDOS){
		/* assume CMI Series I/II/IIx instrument file */
		printf("CMI Series I/II/IIx instrument file (page 3)\n");
		goto printinfo_instrument_exit;
	}

	/* assume CMI Series III instrument file */
	printf("CMI Series III instrument file\n\n");

	if (verbose){
		/* interpret mpcmd packets */
		ret=cmi3_mpcmds_interpret((u_char *)buf,size);
		printf("\n");
	}

printinfo_instrument_exit:
	free(buf);
	return ret;
}



/* EOF */
