#ifndef __CMIOS9_VOICE_H
#define __CMIOS9_VOICE_H
/*
* Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2011  K. M. Indlekofer
*
* m.indlekofer@gmx.de
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



/* CMI Series I/II/IIx voice file */



#define CMI_VC_FILE_SIZE 0x00ab /* filesize in QDOS_BLOCKSIZE */



/* mode 1 loop segment parameters */
struct cmi_loop1seg_s{
	u_char dummy0;
	u_char dummy1;
	u_char dummy2;
	u_char nr; /* segment number: 0,...,32 */
	u_char dummy4;
	u_char dummy5;
	u_char dummy6;
	u_char dummy7;
	u_char dummy8;
	u_char dummy9;
	u_char dummya;
	u_char slp;
	u_char dummyc;
	u_char dummyd;
	u_char dummye;
	u_char dummyf;
}; /* should be 0x0010 bytes */



/* abs. sector 0 */
/* voice descriptor version 3 and 4 */
struct cmi_vc_des_s{
#define CMI_VC_MODE1	1
#define CMI_VC_MODE4	4
	u_char mode; /* voice mode */
#define CMI_VC_VER3		3
#define CMI_VC_VER4		4
	u_char version; /* file version */
	u_char dummy1[0x007e];

	/* at byte 0x0080 */
	u_char dummy2;
	u_char dummy3;
	u_char dummy4;
	u_char mode2; /* voice mode, must be the same as above! */
	u_char dummy5[0x001c];
	char name[8]; /* name of linked control file (w/o .CO suffix) */
	u_char dummy6[0x0058];

	/* at byte 0x0100 */
	u_char dummy7[0x1000];

	/* at byte 0x1100: mode 1 segment loops parameters */
#define CMI_LOOP1SEG_NR 33 /* number of segments (33!!!) */
	struct cmi_loop1seg_s loop1seg[CMI_LOOP1SEG_NR];
	/* at byte 0x1310: mode 1 loop parameters */
	u_char dummy8[0x0020];

	/* at byte 0x1330: mode 4 loop parameters */
	u_char dummy11;
	u_char dummy12;
	u_char swav;
	u_char ewav;
	u_char dummy13;
	u_char dummy14;
	u_char dummy15;
	u_char dummy16;
	u_char dummy17;
	u_char dummy18;
	u_char dummy19;
	u_char lflg;
	u_char dummy20;
	u_char dummy21;
	u_char dummy22;
	u_char dummy23;

	/* at byte 0x1340 */
	u_char dummy9[0x00c0];

	/* at byte 0x1400 */
	u_char dummy10[0x00d0];

	/* at byte 0x14d0 */
	u_char dummy24[0x0030];
}; /* should be 0x002a QDOS blocks (0x1500 bytes) */

#define CMI_VC_DES_SIZE 0x002a /* in QDOS_BLOCKSIZE */



/* sample data (8 bits, 1 channel) */
#define CMI_VC_SAMPLE_START		0x002a /* in QDOS_BLOCKSIZE */
#define CMI_VC_SAMPLE_SIZE1		0x0020 /* mode 1, in QDOS_BLOCKSIZE => 4KB */
#define CMI_VC_SAMPLE_SIZE4		0x0080 /* mode 4, in QDOS_BLOCKSIZE => 16KB */

/* samplerate in Hz: */
#define CMI_VC_SAMPRATE_MIN		2100  /* min. */
#define CMI_VC_SAMPRATE_DEF		14080 /* default */
#define CMI_VC_SAMPRATE_MAX		30200 /* max. */



/* CMI Series III */



/* CMI Series III mpcmd packets (for mptask) */

/* mpcmd packet header */
struct cmi3_mpcmdhead_s{
	u_char size[2]; /* number of following Bytes, >=2 */
	u_char id;   /* ID */
	u_char sid;  /* sub-ID */
};
#define CMI3_MPCMDHEAD_SIZE 4 /* in bytes */

/* mpcmd IDs */
#define CMI3_MPCMDID_END		0x00
#define CMI3_MPCMDID_VCEFF		0x01          /* EFF */
#define CMI3_MPCMDID_WAVE		0x03 /* ENT */
#define CMI3_MPCMDID_CREATE		0x04 /* ENT */
#define CMI3_MPCMDID_CURRENT	0x05 /* ENT */
#define CMI3_MPCMDID_SVMAP		0x06
#define CMI3_MPCMDID_NAME		0x07 /* ENT */
#define CMI3_MPCMDID_LOAD		0x08 /* ENT */
#define CMI3_MPCMDID_SVEFF		0x09          /* EFF */
#define CMI3_MPCMDID_TUNE		0x0a /* ENT */
#define CMI3_MPCMDID_RESET		0x0b /* ENT */
#define CMI3_MPCMDID_SVLOOPA	0x0d
#define CMI3_MPCMDID_DELETE		0x0e /* ENT */
#define CMI3_MPCMDID_NPHONY		0x0f
#define CMI3_MPCMDID_CHMASK		0x10
#define CMI3_MPCMDID_SWAP		0x11 /* ENT */
#define CMI3_MPCMDID_SVLOOPB	0x12
#define CMI3_MPCMDID_SAVE		0x13 /* ENT */
#define CMI3_MPCMDID_UPDATE		0x14 /* ENT */
#define CMI3_MPCMDID_SWAP2		0x16 /* ENT */
#define CMI3_MPCMDID_MIDIOUTM	0x21
#define CMI3_MPCMDID_MIDIINM	0x22
#define CMI3_MPCMDID_CCCMD		0x23
#define CMI3_MPCMDID_COPYSV		0x24
#define CMI3_MPCMDID_GETNAME	0x25 /* ENT */
#define CMI3_MPCMDID_MATCHNAME	0x26 /* ENT */
#define CMI3_MPCMDID_LDCCPROG	0x27
#define CMI3_MPCMDID_CHMASK2	0x28
#define CMI3_MPCMDID_PATH		0x29
#define CMI3_MPCMDID_DRCCMASK	0x2a
#define CMI3_MPCMDID_SYNC		0x2b
#define CMI3_MPCMDID_MIDICTRLM	0x2c
#define CMI3_MPCMDID_CHLEVELS	0x2d

/* mpcmd SIDs: entities for those IDs marked "ENT" above! */
#define CMI3_MPCMDENT_IN		0x01
#define CMI3_MPCMDENT_VC		0x02
#define CMI3_MPCMDENT_SV		0x03
#define CMI3_MPCMDENT_SY		0x04
#define CMI3_MPCMDENT_SYWOVC	0x05
#define CMI3_MPCMDENT_FNCT		0x06
#define CMI3_MPCMDENT_STEREOVC	0x07
#define CMI3_MPCMDENT_SVNAMES	0x08
#define CMI3_MPCMDENT_DEFVC		0x09
#define CMI3_MPCMDENT_DEFSV		0x0a
#define CMI3_MPCMDENT_CCPAGE	0x0b
#define CMI3_MPCMDENT_FNCTIDX	0x0c
#define CMI3_MPCMDENT_EXVC		0x0e
#define CMI3_MPCMDENT_KBDSPLIT	0x0f
#define CMI3_MPCMDENT_CURRENT	0x10
#define CMI3_MPCMDENT_SYANYVC	0x11

/* mpcmd SIDs: effect data for those IDs marked "EFF" above! */
#define CMI3_MPCMDEFF_PATCH		1
#define CMI3_MPCMDEFF_CONTROL	2

/* voice effect types */
#define CMI3_VCEFF_VIBRATE		0x01
#define CMI3_VCEFF_VIBDEPTH		0x02
#define CMI3_VCEFF_VIBDELAY		0x03
#define CMI3_VCEFF_VIBATTACK	0x04
#define CMI3_VCEFF_ATTACK		0x05
#define CMI3_VCEFF_LEVEL		0x09
#define CMI3_VCEFF_RELEASE		0x0a
#define CMI3_VCEFF_RESONANCE	0x0e
#define CMI3_VCEFF_FILTER0		0x0f /* 3.00 */
#define CMI3_VCEFF_TREMRATE		0x12
#define CMI3_VCEFF_TREMDEPTH	0x13
#define CMI3_VCEFF_TREMDELAY	0x14
#define CMI3_VCEFF_TREMATTACK	0x15
#define CMI3_VCEFF_PITCHBEND	0x16
#define CMI3_VCEFF_TUNOCTSTFINE	0x18
#define CMI3_VCEFF_FILTER		0x19
#define CMI3_VCEFF_KBDMAPSW		0x1e
#define CMI3_VCEFF_VIBSW		0x1f
#define CMI3_VCEFF_TREMSW		0x20
#define CMI3_VCEFF_LEVEL2		0x22
#define CMI3_VCEFF_TUNPU		0x23
#define CMI3_VCEFF_FILTER2		0x24
#define CMI3_VCEFF_PORTA		0x25
#define CMI3_VCEFF_BALANCE		0x26
#define CMI3_VCEFF_PORTASW		0x27
#define CMI3_VCEFF_GLISSSW		0x28
#define CMI3_VCEFF_RATETIMESW	0x29

/* subvoice effect types */
#define CMI3_SVEFF_FASTATT		0x05
#define CMI3_SVEFF_HOLDTIME		0x06
#define CMI3_SVEFF_DECAY		0x07
#define CMI3_SVEFF_SUSLEVEL		0x08
#define CMI3_SVEFF_LEVEL		0x09
#define CMI3_SVEFF_FASTREL		0x0a
#define CMI3_SVEFF_KEYLEVEL		0x0b /* 3.00 */
#define CMI3_SVEFF_KEYATT		0x0c /* 3.00 */
#define CMI3_SVEFF_KEYREL		0x0d /* 3.00 */
#define CMI3_SVEFF_RESONANCE	0x0e
#define CMI3_SVEFF_FILTER0		0x0f /* 3.00 */
#define CMI3_SVEFF_SLOWATT		0x10
#define CMI3_SVEFF_SLOWREL		0x11
#define CMI3_SVEFF_LOOPDECAY	0x17
#define CMI3_SVEFF_TUNOCTSTFINE	0x18
#define CMI3_SVEFF_FILTER		0x19
#define CMI3_SVEFF_LINATTSW		0x1a /* XXX hidden??? */
#define CMI3_SVEFF_ATTSW		0x1b
#define CMI3_SVEFF_RELSW		0x1c
#define CMI3_SVEFF_LOOPSW		0x1d
#define CMI3_SVEFF_SLURSW		0x21
#define CMI3_SVEFF_RELLOOPSW	0x2a
#define CMI3_SVEFF_ALTLOOPSW	0x2b

/* control data */
#define CMI3_CONTROL_ON			0xff
#define CMI3_CONTROL_OFF		0x7f
#define CMI3_CONTROL_MIDI(x)	(x)

/* patch types */
#define CMI3_PATCH_CONST		1
#define CMI3_PATCH_CONTROL		2
#define CMI3_PATCH_SWITCH		3
#define CMI3_PATCH_FUNC			4
#define CMI3_PATCH_REVOFF		0xff80 /* offset for reversed controller/switch */



/* CMI Series III voice file */



#define CMI3_WRAM_MAX			0x02000000 /* 32MB, max. WRAM size in bytes */
#define CMI3_WRAM_BLOCKSIZE		0x00001000 /* 4KB, WRAM blocksize in bytes */
/* Note: "null waveform" consists of 16bit mono data samples: */
/*       sine     32, 64, 128, 256   */
/*       zero     32                 */
/*       square   32, 64, 128, 256   */
/*       zero     32                 */
/*       tringle  32, 64, 128, 256   */
/*       zero     32                 */
/*       sawtooth 32, 64, 128, 256   */
/*       zero     32                 */
/*                total: 4KB         */
/* Note: "null waveform" has CRC==0x00000000 */
/* Note: wsexec uses CRC as a (hopefully) unique ID of a subvoice sample */
/*       for checking if a subvoice sample is already loaded */



/* abs. sector 0 */
/* voice ID version 3.01 */
struct cmi3_vc_id_s{
	u_char version[2]; /* version number */
	u_char dummy1[14];
	u_char stereo;     /* mode: stereo/mono */
	u_char dummy2;
	u_char psize[2];   /* size of mpcmds (or 0 if default) */
	u_char fsize[2];   /* size of space for function tables */
	u_char funcsmax;   /* max. number of functions */
	u_char dummy4[233];
}; /* should be OS9L2_BLOCKSIZE */
#define CMI3_VC_ID_MONO		0x00
#define CMI3_VC_ID_STEREO	0xff

#define CMI3_VC_ID_SIZE		OS9L2_BLOCKSIZE /* in bytes */
#define CMI3_VC_ID_VER3		3 /* CMI Series III */
#define CMI3_VC_ID_SVER0	0 /* sub-version */
#define CMI3_VC_ID_SVER1	1 /* sub-version */



/* abs. sector 1-2 */
/* subvoice index */
struct cmi3_sv_index_s{
	u_char start[3]; /* subvoice start address */
	u_char number;   /* subvoice number */
};

#define CMI3_SV_NULL		0        /* stands for: not used */
#define CMI3_SV_MIN			1        /* min. subvoice number */
#define CMI3_SV_MAX			64       /* max. subvoice number */
#define CMI3_SV_NR			(SV_MAX-SV_MIN+1)
#define CMI3_SV_INDEX_OFF	(1*OS9L2_BLOCKSIZE) /* in bytes */
#define CMI3_SV_INDEX_NR	128 /* number of subvoice indices */
#define CMI3_SV_INDEX_SIZE	(OS9L2_BLOCKSIZE<<1) /* in bytes */



/* abs. sector 3-10 */
/* voice mpcmds */
#define CMI3_VC_MPCMDS_OFF		(3*OS9L2_BLOCKSIZE) /* in bytes */
#define CMI3_VC_MPCMDS_SIZE		(OS9L2_BLOCKSIZE<<3) /* default 8*OS9L2_BLOCKSIZE, in bytes */



/* function */
struct cmi3_func_entry_s{
	u_char f[2]; /* 16-Bit function value */
};

#define CMI3_FUNC_TABLE_OFF		(11*OS9L2_BLOCKSIZE) /* in bytes */
#define CMI3_FUNC_TABLE_NR		128 /* number of entries per table */
#define CMI3_FUNC_TABLE_SIZE	OS9L2_BLOCKSIZE /* in bytes */



/* subvoice */

/* rel. sector 0 */
/* subvoice ID */
struct cmi3_sv_id_s{
	u_char dummy1[16];
#define CMI3_SVFORWAVE_B		0x80 /* B-side Bit */
#define CMI3_SVFORWAVE_MASK		0x7f /* subvoice number mask */
	u_char svforwave;  /* subvoice where samples are stored */
#define CMI3_SV_ID_16BIT 0x02
	u_char sasize;     /* size of one sample value in Bytes */
	u_char size[4];    /* size of samples (per A-side if stereo) */
	u_char sarate[4];  /* sample rate */
	u_char crc[4];     /* sample CRC */
	u_char segment[2]; /* segment length */
	u_char dummy2;
	/* Stereo */
	u_char bsvforwave; /* subvoice where samples of B-side are stored */
	u_char bsize[4];   /* B-side size of samples */
	u_char bcrc[4];    /* B-side sample CRC */
	/* Stereo end */
#define CMI3_SV_NAME_LEN 16
	char name[CMI3_SV_NAME_LEN];  /* name */
	char sname[CMI3_SV_NAME_LEN]; /* source voice name (or 0 if this voice) */
	u_char dummy3[182];
}; /* should be OS9L2_BLOCKSIZE */
#define CMI3_SV_ID_SIZE OS9L2_BLOCKSIZE /* in bytes */

/* rel. sector 1-8 */
/* subvoice mpcmds */
#define CMI3_SV_MPCMDS_OFF		CMI3_SV_ID_SIZE /* in bytes */
#define CMI3_SV_MPCMDS_SIZE		(OS9L2_BLOCKSIZE<<3) /* default 8*OS9L2_BLOCKSIZE, in bytes */

/* sample start */
#define CMI3_SV_SAMPLE_OFF		(CMI3_SV_ID_SIZE+CMI3_SV_MPCMDS_SIZE)

#define CMI3_SV_SAMPRATE_DEF	44100 /* default samplerate in Hz */



/* cmios9-internal definitions for CMI Series III mpcmd packets */

/* cmios9-internal mpcmd type: how to interpret sid and data */
#define CMI3_MPCMDTYPE_ENT			0x0100 /* entity */
#define CMI3_MPCMDTYPE_NAMENT		0x0101 /* named entity */
#define CMI3_MPCMDTYPE_TUNENT		0x0104 /* tune entity */
#define CMI3_MPCMDTYPE_VCEFF		0x0206 /* voice effect */
#define CMI3_MPCMDTYPE_SVEFF		0x0207 /* subvoice effect */
#define CMI3_MPCMDTYPE_MIDIP		0x0002 /* MIDI port */
#define CMI3_MPCMDTYPE_LOOPPT		0x0003 /* loop points */
#define CMI3_MPCMDTYPE_NUMBER		0x0005 /* number */
#define CMI3_MPCMDTYPE_SVMAP		0x0008 /* subvoice map */
#define CMI3_MPCMDTYPE_CHMASK		0x0009 /* channel mask */
#define CMI3_MPCMDTYPE_CTRLMAP		0x000a /* MIDI control map */
#define CMI3_MPCMDTYPE_MASK_ENT		0x0100 /* mask for entity type */
#define CMI3_MPCMDTYPE_MASK_EFF		0x0200 /* mask for effect type */

struct cmi3_mpcmdid_s{
	int id;
	int type; /* cmios9-internal mpcmd type */
	char *name;
};

/* mpcmd SIDs: entities for those IDs with types CMI3_MPCMDTYPE_MASK_ENT */
struct cmi3_mpcmdentity_s{
	int sid;
	char *name;
};

/* (sub)voice effects: for those IDs with types CMI3_MPCMDTYPE_MASK_EFF */
struct cmi3_mpcmdeffect_s{
	int type;
	char *name;
};

/* patch types */
struct cmi3_mpcmdpatch_s{
	int type;
	int size; /* in words (2 bytes) */
};

/* cmios9-internal mpcmd datatype: */
#define CMI3_MPCMDDATA_HEXDUMP		0	/* hexdump */
#define CMI3_MPCMDDATA_PATCH		1	/* patch */
#define CMI3_MPCMDDATA_CONTROL		2	/* control */
#define CMI3_MPCMDDATA_TUNING		4	/* tuning */
#define CMI3_MPCMDDATA_STRING		5	/* string */
#define CMI3_MPCMDDATA_LOOPPT		6	/* loop points */
#define CMI3_MPCMDDATA_SVMAP		8	/* subvoice map */
#define CMI3_MPCMDDATA_CHMASK		9	/* channel mask */
#define CMI3_MPCMDDATA_CTRLMAP		10	/* MIDI control map */



extern int cmi3_wav2vc(char *vcname,char *wname,char *buf,u_int bufsiz);

extern int vc2wav(struct genfile_s *fp,char *wname,int sv,char *buf,u_int bufsiz);

extern int printinfo_voice(struct genfile_s *fp,int verbose);
extern int printinfo_subvoice(struct genfile_s *fp,int verbose);
extern int printinfo_system(struct genfile_s *fp,int verbose);
extern int printinfo_instrument(struct genfile_s *fp,int verbose);



#endif /* !__CMIOS9_VOICE_H */
