/*
* Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2010  K. M. Indlekofer
*
* m.indlekofer@gmx.de
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



#include "cmios9_err.h"
#include "cmios9.h"
#include "cmios9_wav.h"



int
wav_write_head(int outdes,
			   u_int totsize,u_int chnr,u_int samprate,u_int bitnr)
{
	struct wav_riffhead_s wavriffhead;
	struct wav_chunkhead_s wavchunkhead;
	struct wav_fmthead_s wavfmthead;

	u_int drate;



	/* totsize: bytes per channel */
	totsize*=chnr;



	/* create WAVE RIFF-header */
	bzero(&wavriffhead,sizeof(struct wav_riffhead_s));
	bcopy(WAV_RIFFHEAD_RIFFSTR,wavriffhead.riffstr,strlen(WAV_RIFFHEAD_RIFFSTR));
	{
		u_int fsize;

		fsize=WAV_HEAD_SIZE-8+totsize;
		wavriffhead.fsize[0]=0xff&fsize;
		wavriffhead.fsize[1]=0xff&(fsize>>8);
		wavriffhead.fsize[2]=0xff&(fsize>>16);
		wavriffhead.fsize[3]=0xff&(fsize>>24);
	}
	bcopy(WAV_RIFFHEAD_WAVESTR,wavriffhead.wavestr,strlen(WAV_RIFFHEAD_WAVESTR));

	/* write WAVE RIFF-header */
	if (WRITE(outdes,&wavriffhead,sizeof(struct wav_riffhead_s))!=sizeof(struct wav_riffhead_s)){
		perror("write WAVE RIFF-header");
		return -1;
	}



	/* create FMT chunk-header */
	bcopy(WAV_CHUNKHEAD_FMTSTR,wavchunkhead.typestr,strlen(WAV_CHUNKHEAD_FMTSTR));
	{
		u_int csize;

		csize=WAV_CHUNKHEAD_FMTCSIZE;
		wavchunkhead.csize[0]=0xff&csize;
		wavchunkhead.csize[1]=0xff&(csize>>8);
		wavchunkhead.csize[2]=0xff&(csize>>16);
		wavchunkhead.csize[3]=0xff&(csize>>24);
	}
	/* write chunk-header */
	if (WRITE(outdes,&wavchunkhead,sizeof(struct wav_chunkhead_s))!=sizeof(struct wav_chunkhead_s)){
		perror("write WAVE FMT-header");
		return -1;
	}

	/* set FMT-header */
	wavfmthead.ftag[0]=0xff&WAV_HEAD_FTAG;
	wavfmthead.ftag[1]=0xff&(WAV_HEAD_FTAG>>8);
	wavfmthead.chnr[0]=0xff&chnr;
	wavfmthead.chnr[1]=0xff&(chnr>>8);
	wavfmthead.srate[0]=0xff&samprate;
	wavfmthead.srate[1]=0xff&(samprate>>8);
	wavfmthead.srate[2]=0xff&(samprate>>16);
	wavfmthead.srate[3]=0xff&(samprate>>24);
	drate=chnr*samprate*(bitnr/8); /* chnr!!! */
	wavfmthead.drate[0]=0xff&drate;
	wavfmthead.drate[1]=0xff&(drate>>8);
	wavfmthead.drate[2]=0xff&(drate>>16);
	wavfmthead.drate[3]=0xff&(drate>>24);
	wavfmthead.balign[0]=0xff&((bitnr/8)*chnr);      /* chnr!!! */
	wavfmthead.balign[1]=0xff&(((bitnr/8)*chnr)>>8); /* chnr!!! */
	wavfmthead.bitnr[0]=0xff&bitnr;
	wavfmthead.bitnr[1]=0xff&(bitnr>>8);

	/* write WAVE FMT-header */
	if (WRITE(outdes,&wavfmthead,sizeof(struct wav_fmthead_s))!=sizeof(struct wav_fmthead_s)){
		perror("write WAVE FMT-header");
		return -1;
	}



	/* create WAVE DATA chunk-header */
	bcopy(WAV_CHUNKHEAD_DATASTR,wavchunkhead.typestr,strlen(WAV_CHUNKHEAD_DATASTR));
	wavchunkhead.csize[0]=0xff&totsize;
	wavchunkhead.csize[1]=0xff&(totsize>>8);
	wavchunkhead.csize[2]=0xff&(totsize>>16);
	wavchunkhead.csize[3]=0xff&(totsize>>24);

	/* write chunk-header */
	if (WRITE(outdes,&wavchunkhead,sizeof(struct wav_chunkhead_s))!=sizeof(struct wav_chunkhead_s)){
		perror("write WAVE DATA-header");
		return -1;
	}

	return 0;
}



int
wav_read_head(int indes,
			  u_int *sizep,u_int *chnrp,u_int *sampratep,u_int *bitnrp,
			  char **errstrp)
{
	struct wav_riffhead_s wavriffhead;
	struct wav_chunkhead_s wavchunkhead;
	struct wav_fmthead_s wavfmthead;

	u_int i;

	u_int fsize;
	u_int csize;
	u_int totsize;

	u_int size;
	u_int chnr;
	u_int samprate;
	u_int bitnr;



	/* read WAVE RIFF-header */
	if (READ(indes,(char *)&wavriffhead,sizeof(struct wav_riffhead_s))!=sizeof(struct wav_riffhead_s)){
		if (errstrp!=NULL){
			*errstrp="read WAVE RIFF-header";
		}
		return -1;
	}

	/* check riffstr */
	for (i=0;i<strlen(WAV_RIFFHEAD_RIFFSTR);i++){
		if (WAV_RIFFHEAD_RIFFSTR[i]!=wavriffhead.riffstr[i]){
			if (errstrp!=NULL){
				*errstrp="no RIFF";
			}
			return -1;
		}
	}

	/* get fsize */
	fsize=wavriffhead.fsize[0]
		+(wavriffhead.fsize[1]<<8)
		+(wavriffhead.fsize[2]<<16)
		+(wavriffhead.fsize[3]<<24);
	if (fsize<sizeof(struct wav_riffhead_s)-8){
		if (errstrp!=NULL){
			*errstrp="invalid fsize";
		}
		return -1;
	}

	/* check wavestr */
	for (i=0;i<strlen(WAV_RIFFHEAD_WAVESTR);i++){
		if (WAV_RIFFHEAD_WAVESTR[i]!=wavriffhead.wavestr[i]){
			if (errstrp!=NULL){
				*errstrp="no WAVE";
			}
			return -1;
		}
	}



	/* search for fmt chunk */
	for(;;){
		/* read next chunk */
		if (READ(indes,(char *)&wavchunkhead,sizeof(struct wav_chunkhead_s))!=sizeof(struct wav_chunkhead_s)){
			if (errstrp!=NULL){
				*errstrp="read WAVE FMT-header";
			}
			return -1;
		}

		/* get csize */
		csize=wavchunkhead.csize[0]
			+(wavchunkhead.csize[1]<<8)
			+(wavchunkhead.csize[2]<<16)
			+(wavchunkhead.csize[3]<<24);

		for (i=0;i<strlen(WAV_CHUNKHEAD_FMTSTR);i++){
			if (WAV_CHUNKHEAD_FMTSTR[i]!=wavchunkhead.typestr[i]){
				break; /* not fmt chunk */
			}
		}
		if (i==strlen(WAV_CHUNKHEAD_FMTSTR)){
			break; /* found fmt */
		}

		/* skip rest of chunk */
		if (LSEEK(indes,(OFF_T)csize,SEEK_CUR)<0){
			if (errstrp!=NULL){
				*errstrp="lseek WAVE FMT-header";
			}
		}
	}
	if (csize!=WAV_CHUNKHEAD_FMTCSIZE){
		if (errstrp!=NULL){
			*errstrp="invalid FMT csize";
		}
		return -1;
	}

	/* read rest of fmt chunk */
	if (READ(indes,(char *)&wavfmthead,sizeof(struct wav_fmthead_s))!=sizeof(struct wav_fmthead_s)){
		if (errstrp!=NULL){
			*errstrp="read WAVE FMT-header";
		}
		return -1;
	}

	/* check ftag */
	/* XXX allow only particular format */
	if ((wavfmthead.ftag[0]!=(0xff&WAV_HEAD_FTAG))
		||(wavfmthead.ftag[1]!=(0xff&(WAV_HEAD_FTAG>>8)))){
		if (errstrp!=NULL){
			*errstrp="invalid ftag";
		}
		return -1;
	}

	/* get chnr */
	chnr=wavfmthead.chnr[0]
		+(wavfmthead.chnr[1]<<8);
	if (chnr==0){
		if (errstrp!=NULL){
			*errstrp="invalid chnr";
		}
		return -1;
	}

	/* set samprate */
	samprate=wavfmthead.srate[0]
		+(wavfmthead.srate[1]<<8)
		+(wavfmthead.srate[2]<<16)
		+(wavfmthead.srate[3]<<24);

	/* XXX check drate: must be chnr*samprate*(bitnr/8) */

	/* get bitnr */
	bitnr=wavfmthead.bitnr[0]
		+(wavfmthead.bitnr[1]<<8);
	if (bitnr==0){
		if (errstrp!=NULL){
			*errstrp="invalid bitnr";
		}
		return -1;
	}

	/* XXX check balign: must be (bitnr/8)*chnr */



	/* search for data chunk */
	for(;;){
		/* read next chunk */
		if (READ(indes,(char *)&wavchunkhead,sizeof(struct wav_chunkhead_s))!=sizeof(struct wav_chunkhead_s)){
			if (errstrp!=NULL){
				*errstrp="read WAVE DATA-header";
			}
			return -1;
		}

		/* get csize */
		csize=wavchunkhead.csize[0]
			+(wavchunkhead.csize[1]<<8)
			+(wavchunkhead.csize[2]<<16)
			+(wavchunkhead.csize[3]<<24);

		for (i=0;i<strlen(WAV_CHUNKHEAD_DATASTR);i++){
			if (WAV_CHUNKHEAD_DATASTR[i]!=wavchunkhead.typestr[i]){
				break; /* not data chunk */
			}
		}
		if (i==strlen(WAV_CHUNKHEAD_DATASTR)){
			break; /* found fmt */
		}

		/* skip rest of chunk */
		if (LSEEK(indes,(OFF_T)csize,SEEK_CUR)<0){
			if (errstrp!=NULL){
				*errstrp="lseek WAVE DATA-header";
			}
		}
	}

	/* totsize */
	totsize=csize;
	if (totsize%chnr!=0){
		if (errstrp!=NULL){
			*errstrp="invalid totsize";
		}
		return -1;
	}



	/* size: bytes per channel */
	size=totsize/chnr;

	if (sizep!=NULL){
		*sizep=size;
	}
	if (chnrp!=NULL){
		*chnrp=chnr;
	}
	if (sampratep!=NULL){
		*sampratep=samprate;
	}
	if (bitnrp!=NULL){
		*bitnrp=bitnr;
	}
	
	return 0;
}



/* EOF */

